//,     ,  
class CPoint3D
{
public:
	double X, Y, Z;
	void operator = (double* dPoint);
	double operator - (CPoint3D Other);	//
	bool operator == (CPoint3D Other);
	bool operator != (CPoint3D Other);

	void Fill(double *dPoint);
	bool Equal(CPoint3D *Other);		//     (  dG_DistTol)
};

class CEdgeParameters
{
public:
	// 
	enum {Line, Arc, Other} Type;
	
	// 
	double dLength;

	//   (  0)
	double dAngle;

	//  (  0)
	double dRadius;

	void ScanEdge(tag_t Edge);
	void WriteParameters(CFile *File);
};

class CHole;
class CPunch;
class CPressIn;

class CSection
{
public:
	CPoint3D PointBase;				//,    
	CPoint3D PointRealBase;			//,      (    )
	CPoint3D PointRounding;			//,     
	CPoint3D PointBevel;			//   
	CPoint3D PointRealBevel;		//,      (        )
	CPoint3D PointUndercutRounding;	//,      
	CPoint3D PointUndercut;			//,    
//	CPoint3D VectorBaseToRealBase;	//,    Base   RealBase (   bPoint=true)

	double r;						//    
	double R;						//    
	bool bNullCurva;				// ,      , . .   
	double H;						//    
	double Alpha;					// 
	double rUndercut;				//  
//	double AlphaUndercut;			//  

//    -
	struct
	{
		CPoint3D PointRounding;			//,     
		CPoint3D PointBevel;			//   
		double Alpha;					//
		double r;						//    
		double H;						//    
	} CalcPoints;

//	bool bPoint;					//      
//	bool bPointUndercut;			//       

	bool bIsRoundingPresent;		//     
	bool bIsBevelPresent;			//     
	bool bIsUndercutRoundingPresent;//      
	bool bIsUndercutPresent;		//     

	tag_t PlaneSection;				//    
	CPoint3D PlaneSectionNormal;	//   (   )

/*	tag_t FaceRounding;				//    
	tag_t FaceBevel;				//    
	tag_t FaceUndercutRounding;		//     
	tag_t FaceUndercut;				//    */

	CSection(void);

	//      
	//   :
	//{0x},{0y},{0z} -   PointRealBase
	//{1x},{1y},{1z} -   PointRealBevel
	//{2x},{2y},{2z} -   PointUndecut
	//{3x},{3y},{3z} -   PointBase
	//{4x},{4y},{4z} -   PointRounding
	//{c4x},{c4y},{c4z} -   PointRounding  
	//{5x},{5y},{5z} -   PointBevel
	//{5x},{5y},{5z} -   PointBevel  
	//{6x},{6y},{6z} -   PointUndercutRounding
	//{7} -   r   
	//{7} -   r     
	//{8} -   R   
	//{9} -   H   
	//{9} -   H      
	//{10g},{10r}  -  Alpha  , 
	//{10g},{10r}  -  Alpha  ,   
	//{11} -   rUndercut 
	//{12} - 
	//{13} -    bIsRoundingPresent
	//{14} -    bIsBevelPresent
	//{15} -     bIsUndercutRoundingPresent
	//{16} -    bIsUndercutPresent
	//{i} -   (  )  
	void WriteSection(CFile *File, CString sFormat, int iIndex);
	//       
	void PostCalc(double t, double E, double K, double n, double *dNormal, tag_t BaseFace);
	//      - (    )
//	void CalcFirstStep(double E, double K, double n);

/*	//     
#ifndef _DEBUG
	void WritePart(ofstream *File, bool bNumber, bool bBase, bool bRealBase, bool bRounding, bool bBevel, bool bFlag, bool br, bool bR, bool bH, bool bAlpha);
#endif	
	//   RealBasePoint
	void CalcRealBasePoint();
	//   PointRealBevel
	void CalcRealBevelPoint();
	//        .      ,  .
	void ScanOuterEdgeRounding(tag_t Base, tag_t Base2, tag_t Plane, uf_list_p_t *RoundingFaces);*/
};

class CEdge
{
public:
	CSection *Sections;		//    
	int iSectionsNum;

//	tag_t Edge;				//

	bool bPS;				//      
	bool bNothing;			//     
	bool bIgnore;			//  

	double dOffset;			//      
	int iBridgeType;		//       (  )
							//0 - 
							//1 - 
							//2 - 

	CEdge(void);
	~CEdge(void);
/*	//     
#ifndef _DEBUG
	void WritePart(ofstream *File, bool bNumber, bool bBase, bool bRealBase, bool bRounding, bool bBevel, bool bFlag, bool br, bool bR, bool bH, bool bAlpha);
#endif
	//   
	void CorrectRadiuses(void);
	//    (        ,   ,   )
	void CorrectAnglesBase(int iFirst, int iLast);
	//    (        ,   ,   )
	void CorrectAnglesUndercut(int iFirst, int iLast);
	//    (      )
	void CorrectAngles(void);
	void ScanOuterEdge(tag_t Base, tag_t Base2, tag_t Curve, int iIntervalsNum, uf_list_p_t *RoundingFaces);
	//        .      , ,     ,   .
	//        .
	void ScanOuterEdgeRounding(tag_t Base, tag_t Base2, tag_t Curve, int iIntervalsNum, uf_list_p_t *RoundingFaces, uf_list_p_t *SectionPlanes);
	//        .      ,
	//     ,   ,   .
	//     .
	void ScanOuterEdgeBevel(tag_t Base, tag_t Base2, int iPlaneNum, uf_list_p_t SectionPlanes, uf_list_p_t RoundingFaces, uf_list_p_t *BevelFaces);
	//         .      ,
	//     ,   ,   ,    (    )
	//      .
	void ScanOuterEdgeUndercutRounding(tag_t Base, tag_t Base2, int iPlaneNum, uf_list_p_t SectionPlanes, uf_list_p_t BevelFaces, uf_list_p_t RoundingFaces, uf_list_p_t *UndercutRoundingFaces);
	//        .      ,
	//     ,   ,    ,    (    )
	//     .
	void ScanOuterEdgeUndercut(tag_t Base, tag_t Base2, int iPlaneNum, uf_list_p_t SectionPlanes, uf_list_p_t UndercutRoundingFaces, uf_list_p_t BevelFaces, uf_list_p_t *UndercutFaces);
	//,  ,           Base      .
	//  ,      ,   
	void ScanOuterEdgeVector(tag_t Base, int iPlaneNum, uf_list_p_t SectionPlanes);
	//    
	void ScanOuterEdge(tag_t Base, tag_t Base2, tag_t Curve, uf_list_p_t Adjacent, int iIntervalsNum);*/

//##########################################################################

	//      
	int CreateOuterPlanes(tag_t Base, tag_t Edge, int iIntervalsNum);
	//     :
	//PointRealBase
	//PointRealBevel
	//PointUndercut
	int ScanRealPoints(tag_t FaceBase, uf_list_p_t FacesBevel, uf_list_p_t FacesUndercut);
	//     :
	//PointBase
	//PointRounding
	//PointBevel
	//PointUndercutRounding
	//    ScanRealPoints
	int ScanPoints(uf_list_p_t FacesBevel, uf_list_p_t FacesUndercut, uf_list_p_t FacesRounding, uf_list_p_t FacesUndercutRounding);
	//     
	//   :
	//{0} -   iSectionsNum   
	//{1} -  bPS ,          
	//{2} -  bNothing ,       
	void WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections, int iIndex);
	//       
	//    ,  
	void PostCalcA(tag_t FaceBase);
	//       
	//    ,  
	void PostCalcB(void);
	//        
	void Reverse(void);
	//      
	double CalcMeanR(void);
};

class CPart
{
public:
	double t;				// 
	tag_t m_BaseFace;		//  

	CEdge *Edges;			//и
	int iEdgesNum;

	CHole *Holes;			//
	int iHolesNum;

	CPunch *Punches;		//
	int iPunchesNum;

	CPressIn *PressIns;		//
	int iPressInsNum;

	int iPressEdgesNum;		//   
	int *iPressPointsNum;	//     
	CPoint3D **PointsPress;	//  

//    
	struct
	{
		double dBaseNormal[3];		//  
		double dBaseRoundingRadius;	//,       
	} Service;

	CPart(void);
	void Delete(void);
/*	//      ,      OXY,
	//             X  Y
	//        OZ
	//    
	int Rotate(CPoint3D Normal);
	//    
	int CreateBase(void);
	//     
	void WritePart(CString File, bool bNumber=true, bool bBase=true, bool bRealBase=true, bool bRounding=true, bool bBevel=true, bool bFlag=true, bool br=true, bool bR=true, bool bH=true, bool bAlpha=true);
	//         
	void CorrectScannedResultsOutline(void);
	//  
	void ScanPart(tag_t Part, int iIntervalsNum);
	//    
	void ScanOuterOutline(tag_t Base, tag_t Base2, uf_list_p_t Adjacent, int iIntervalsNum);
	void ScanOuterOutline(tag_t Base, tag_t Base2, int iIntervalsNum, uf_list_p_t *RoundingFaces);*/

//##########################################################################

	//  
	int ScanPart(tag_t Part, int iIntervalsNum);
	//        
	int CreateOuterPlanes(tag_t Base, int iIntervalsNum);
	//        
	int CreateInnerPlanes(tag_t Base, uf_list_p_t FacesPunchBevel, uf_list_p_t FacesPressInBevel, int iIntervalsNum);
	//    CreateOuterPlanes  
	void DeleteOuterPlanes(void);
	//    CreateInnerPlanes  
	void DeleteInnerPlanes(void);
	//      
	void LineUp(void);
	// 
	void Correct(double E, double K, double n);
	//        
	//   , . .    
	//    
	void RotateAndMoveToZero(CPoint3D Normal);
	//      
	//   false   
	bool AskParametersInteractive(void);
	//    
	//   false   
	bool AskDistancesInteractive(void);
	//     
	//   false   
	bool AskCoordsInteractive(void);
	//    
	//   false   
	bool AskBridgesInteractive(void);
	//     
	void WriteParameters(CFile *File);
	//   
	tag_t CreateBase(void);
	tag_t CreateBase1(void);
	//     ( )
	double CalcMinBaseHeight(void);
	//   
	void ScanPress(tag_t Base, int iIntervalsNum);
	//  
	tag_t CreatePress(void);
	tag_t CreatePress1(void);
	//  
	tag_t CreateInvolute(void);
	//      
	void ProcessIgnoredEdges(CEdge *NewEdges);
};

//,   
class CHoleSection
{
public:
	CPoint3D PointRealBase;			//,    

	//      
	//   :
	//{0x},{0y},{0z} -   PointRealBase
	void WriteSection(CFile *File, CString sFormatSections);
};

class CHoleEdge
{
public:
	CHoleSection *Sections;			//    
	int iSectionsNum;
	CEdgeParameters Parameters;			//  

	//   
	int ScanEdge(tag_t Edge, int iIntervalsNum);
	//      
	//   :
	//{0} -   iSectionsNum   
	void WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections);
	//        
	void Reverse(void);
};

class CHole
{
public:
	CHoleEdge *Edges;				//и
	int iEdgesNum;

	//  
	int ScanHole(uf_list_p_t Edges, int iIntervalsNum);
	//     
	//   :
	//{0} -   iEdgesNum   
	void WriteHole(CFile *File, CString sFormat, CString sFormatEdge, CString sFormatSections);
	//     
	void WriteParameters(CFile *File);
};

//,   
class CPunchSection
{
public:
	CPoint3D PointBase;				//,    
	CPoint3D PointRealBase;			//,      (    )
	CPoint3D PointRounding;			//,     
	CPoint3D PointBevel;			//   
	double r;						// 

	tag_t PlaneSection;				//    

	//      
	//   :
	//{0x},{0y},{0z} -   PointRealBase
	//{1x},{1y},{1z} -   PointBevel
	//{2x},{2y},{2z} -   PointBase
	//{3x},{3y},{3z} -   PointRounding
	void WriteSection(CFile *File, CString sFormatSections);
};

class CPunchEdge
{
public:
	CPunchSection *Sections;			//    
	int iSectionsNum;
	CEdgeParameters Parameters;			//  

	//        
	int CreateInnerPlanes(tag_t Edge, int iIntervalsNum);
	//    :
	//PointBevel
	int ScanRealPoints(uf_list_p_t FacesPunchBevel);
	//    :
	//PointRounding
	//PointBase
	int ScanPoints(uf_list_p_t FacesPunchBevel, uf_list_p_t FacesPunchRounding);
	//      
	//   :
	//{0} -   iSectionsNum   
	void WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections);
	//        
	void Reverse(void);
};

class CPunch
{
public:
	CPunchEdge *Edges;				//и
	int iEdgesNum;

	//       
	int CreateInnerPlanes(uf_list_p_t Edges, int iIntervalsNum);
	//     
	//   :
	//{0} -   iEdgesNum   
	void WritePunch(CFile *File, CString sFormat, CString sFormatEdge, CString sFormatSections);
	//     
	void WriteParameters(CFile *File);
	//  ""     
	void CreatePunchBase(tag_t *Base, double dBaseHeight, double dBaseNormal[3]);
	//  ""     
	void CreatePunchPress(tag_t *Press, double dBaseNormal[3]);
};

//,   
class CPressInSection
{
public:
	CPoint3D PointBase;				//,    
	CPoint3D PointRealBase;			//,      (    )
	CPoint3D PointRounding1;		//,     
	CPoint3D PointBevel;			//   
	CPoint3D PointBottom;			//,      
	CPoint3D PointRealBottom;		//,      (         )
	double r1;						//  
	double r2;						//  

	tag_t PlaneSection;				//    

	//      
	//   :
	//{0x},{0y},{0z} -   PointRealBase
	//{1x},{1y},{1z} -   PointRealBottom
	//{2x},{2y},{2z} -   PointBase
	//{3x},{3y},{3z} -   PointRounding1
	//{4x},{4y},{4z} -   PointBevel
	//{5x},{5y},{5z} -   PointBottom
	void WriteSection(CFile *File, CString sFormatSections);
};

class CPressInEdge
{
public:
	CPressInSection *Sections;			//    
	int iSectionsNum;
	CEdgeParameters Parameters;			//  

	//        
	int CreateInnerPlanes(tag_t Edge, int iIntervalsNum);
	//    :
	//PointRealBottom
	int ScanRealPoints(uf_list_p_t FacesPressInBevel);
	//    :
	//PointBase
	//PointRounding1
	//PointBevel
	//PointBottom
	// :
	//1 -       
	//2 -     ,  ,   
	int ScanPoints(uf_list_p_t FacesPressInBevel, uf_list_p_t FacesPressInRounding1, uf_list_p_t FacesPressInRounding2, uf_list_p_t FacesPressInBottom);
	//      
	//   :
	//{0} -   iSectionsNum   
	void WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections);
	//        
	void Reverse(void);
};

class CPressIn
{
public:
	CPressInEdge *Edges;				//и
	int iEdgesNum;

	//       
	int CreateInnerPlanes(uf_list_p_t Edges, int iIntervalsNum);
	//     
	//   :
	//{0} -   iEdgesNum   
	void WritePressIn(CFile *File, CString sFormat, CString sFormatEdge, CString sFormatSections);
	//     
	void WriteParameters(CFile *File);
	//  ""     
	void CreatePressInBase(tag_t *Base, double dBaseNormal[3]);
	//       
	void CreatePressInPress(tag_t *Press, double dBaseNormal[3], double dPartThickness);
	//     
	tag_t CreatePressInBush(double dBaseNormal[3], double dPartThickness);
};
