using System.Collections;
using System.IO;
using System;

namespace KnowledgeBase
{
	public enum EdgeType
	{
		Line,
		Arc,
		Other
	}

	public struct _Lex
	{
		/*  0 - 
					  1 - 
					  2 - 
					  3 - -
					  4 - */
		public int type;
		public string L;	// 
		public int p1,p2;	//    
	};

	public class _Op
	{
		public double priority=0;
		public string L;

		public _Op()
		{
		}

		public _Op(int iP, string sL)
		{
			priority = iP;
			L = sL;
		}
	};

	public class EdgeParameters
	{
		public EdgeType		Type;		// 
		public double		dLength;	// 
		public double		dAngle;		//  ( )
		public double		dRadius;	// ( )

		public int ReadParameters(StreamReader SR)
		{
			string s;

			s = SR.ReadLine();
			if(s.StartsWith("#EdgeType = "))
				switch(s.Substring("#EdgeType = ".Length))
				{
					case "Arc":
						Type = EdgeType.Arc;
						break;
					case "Line":
						Type = EdgeType.Line;
						break;
					case "Other":
						Type = EdgeType.Other;
						break;
					default:
						return 1;
				}
			else
				return 1;

			s = SR.ReadLine();
			if(s.StartsWith("#EdgeLength = "))
				dLength = Convert.ToDouble(s.Substring("#EdgeLength = ".Length));
			else
				return 1;

			s = SR.ReadLine();
			if(s.StartsWith("#EdgeAngle = "))
				dAngle = Convert.ToDouble(s.Substring("#EdgeAngle = ".Length));
			else
				return 1;

			s = SR.ReadLine();
			if(s.StartsWith("#EdgeRadius = "))
				dRadius = Convert.ToDouble(s.Substring("#EdgeRadius = ".Length));
			else
				return 1;

			return 0;
		}
	}

	public enum ObjectType
	{
		Base,
		Hole,
		Punch,
		PressIn
	}

	public class ObjectParameters
	{
		ObjectType			Type;		// 
		EdgeParameters[]	Edges;		//и 

		public int ReadParameters(StreamReader SR)
		{
			string s;

			s = SR.ReadLine();
			if(s.StartsWith("#TypeOfObject = "))
				switch(s.Substring("#TypeOfObject = ".Length))
				{
					case "Base":
						Type = ObjectType.Base;
						break;
					case "Hole":
						Type = ObjectType.Hole;
						break;
					case "Punch":
						Type = ObjectType.Punch;
						break;
					case "PressIn":
						Type = ObjectType.PressIn;
						break;
					default:
						return 1;
				}
			else
				return 1;

			s = SR.ReadLine();
			if(s.StartsWith("#NumOfEdges = "))
				Edges = new EdgeParameters[Convert.ToInt32(s.Substring("#NumOfEdges = ".Length))];
			else
				return 1;

			for(int i=0;i<Edges.Length;i++)
				Edges[i] = new EdgeParameters();

			foreach(EdgeParameters Edge in Edges)
				if(Edge.ReadParameters(SR)!=0)
					return 1;

			return 0;
		}

		public ArrayList FillLex(string sExpression)
		{
			int iPos = 0;
			ArrayList Lexs = new ArrayList();
			_Lex Lex;
			
			// 
			sExpression += "  ";
			sExpression = sExpression.Replace('[', '(');
			sExpression = sExpression.Replace(']', ')');

			while(iPos<sExpression.Length)
			{
				switch(sExpression[iPos])
				{
					//  
					case ' ':
						iPos++;
						break;
					case '	':
						iPos++;
						break;
					case '\r':
						iPos++;
						break;
					case '\n':
						iPos++;
						break;
					//-
					case ',':
						iPos++;
						break;
					//
					case '*':
						Lex.type = 1;
						Lex.L = "*";
						Lex.p1 = iPos;
						Lex.p2 = iPos;
						Lexs.Add(Lex);
						iPos++;
						break;
					case '/':
						Lex.type = 1;
						Lex.L = "/";
						Lex.p1 = iPos;
						Lex.p2 = iPos;
						Lexs.Add(Lex);
						iPos++;
						break;
					case '+':
						Lex.type = 1;
						Lex.L = "+";
						Lex.p1 = iPos;
						Lex.p2 = iPos;
						Lexs.Add(Lex);
						iPos++;
						break;
					//   
					case '-':
						// -  
						if(
							Lexs.Count==0 || 
							((_Lex)Lexs[Lexs.Count-1]).L=="(" ||
							((_Lex)Lexs[Lexs.Count-1]).L=="==" ||
							((_Lex)Lexs[Lexs.Count-1]).L=="!=" ||
							((_Lex)Lexs[Lexs.Count-1]).L==">=" ||
							((_Lex)Lexs[Lexs.Count-1]).L=="<=" ||
							((_Lex)Lexs[Lexs.Count-1]).L==">" ||
							((_Lex)Lexs[Lexs.Count-1]).L=="<" ||
							((_Lex)Lexs[Lexs.Count-1]).L=="&&" ||
							((_Lex)Lexs[Lexs.Count-1]).L=="||" ||
							((_Lex)Lexs[Lexs.Count-1]).L=="!")
						{
							Lex.type = 0;
							Lex.L = "0";
							Lex.p1 = iPos;
							Lex.p2 = iPos;
							Lexs.Add(Lex);
						}
						Lex.type = 1;
						Lex.L = "-";
						Lex.p1 = iPos;
						Lex.p2 = iPos;
						Lexs.Add(Lex);
						iPos++;
						break;
					//
					case '(':
						Lex.type = 2;
						Lex.L = "(";
						Lex.p1 = iPos;
						Lex.p2 = iPos;
						Lexs.Add(Lex);
						iPos++;
						break;
					case ')':
						Lex.type = 2;
						Lex.L = ")";
						Lex.p1 = iPos;
						Lex.p2 = iPos;
						Lexs.Add(Lex);
						iPos++;
						break;
					//
					//
					default:
						if(char.IsDigit(sExpression[iPos]) || (sExpression[iPos]=='.' && char.IsDigit(sExpression[iPos+1])))
						{
							Lex.type = 0;
							Lex.L = "";
							Lex.p1 = iPos;
							while((char.IsDigit(sExpression[iPos]) || sExpression[iPos]=='.') && iPos<sExpression.Length)
								Lex.L += sExpression[iPos++];
							Lex.p2 = iPos-1;
							Lexs.Add(Lex);
						}
						//-
						if(sExpression[iPos]=='.' && char.IsLetter(sExpression[iPos+1]))
						{
							Lex.type = 3;
							Lex.L = "";
							Lex.p1 = iPos;
							while((char.IsLetter(sExpression[iPos]) || sExpression[iPos]=='.') && iPos<sExpression.Length)
								Lex.L += sExpression[iPos++];
							Lex.p2 = iPos-1;
							Lexs.Add(Lex);
						}
						//
						if(char.IsLetter(sExpression[iPos]))
						{
							Lex.type = 4;
							Lex.L = "";
							Lex.p1 = iPos;
							while(char.IsLetter(sExpression[iPos]) && iPos<sExpression.Length)
								Lex.L += sExpression[iPos++];
							Lex.p2 = iPos-1;
							Lexs.Add(Lex);
						}
						// 
						if(sExpression.Substring(iPos, 2) == "==")
						{
							Lex.type = 1;
							Lex.L = "==";
							Lex.p1 = iPos;
							Lex.p2 = iPos+1;
							Lexs.Add(Lex);
							iPos += 2;
						}
						if(sExpression.Substring(iPos, 2) == "<=")
						{
							Lex.type = 1;
							Lex.L = "<=";
							Lex.p1 = iPos;
							Lex.p2 = iPos+1;
							Lexs.Add(Lex);
							iPos += 2;
						}
						if(sExpression.Substring(iPos, 2) == ">=")
						{
							Lex.type = 1;
							Lex.L = ">=";
							Lex.p1 = iPos;
							Lex.p2 = iPos+1;
							Lexs.Add(Lex);
							iPos += 2;
						}
						if(sExpression.Substring(iPos, 2) == "!=")
						{
							Lex.type = 1;
							Lex.L = "!=";
							Lex.p1 = iPos;
							Lex.p2 = iPos+1;
							Lexs.Add(Lex);
							iPos += 2;
						}
						if(sExpression.Substring(iPos, 1) == "<")
						{
							Lex.type = 1;
							Lex.L = "<";
							Lex.p1 = iPos;
							Lex.p2 = iPos;
							Lexs.Add(Lex);
							iPos++;
						}
						if(sExpression.Substring(iPos, 1) == ">")
						{
							Lex.type = 1;
							Lex.L = ">";
							Lex.p1 = iPos;
							Lex.p2 = iPos;
							Lexs.Add(Lex);
							iPos++;
						}
						// 
						if(sExpression.Substring(iPos, 2) == "&&")
						{
							Lex.type = 1;
							Lex.L = "&&";
							Lex.p1 = iPos;
							Lex.p2 = iPos+1;
							Lexs.Add(Lex);
							iPos += 2;
						}
						if(sExpression.Substring(iPos, 2) == "||")
						{
							Lex.type = 1;
							Lex.L = "||";
							Lex.p1 = iPos;
							Lex.p2 = iPos+1;
							Lexs.Add(Lex);
							iPos += 2;
						}
						if(sExpression.Substring(iPos, 1) == "!")
						{
							Lex.type = 1;
							Lex.L = "!";
							Lex.p1 = iPos;
							Lex.p2 = iPos;
							Lexs.Add(Lex);
							iPos++;
						}
						break;
				}
			}
			
			return Lexs;
		}

		public double Test(string sExpression)
		{
			int iPos = 0;
			Stack OpStack=new Stack(), NumStack=new Stack();
			ArrayList Lexs = FillLex(sExpression);

			OpStack.Push(new _Op(-1, "+"));
			NumStack.Push(.0);

			while(iPos<Lexs.Count)
			{
				//
				if(((_Lex)Lexs[iPos]).type == 0)
					NumStack.Push(Convert.ToDouble(((_Lex)Lexs[iPos++]).L));

				//
				else if(((_Lex)Lexs[iPos]).type == 4)
				{
					_Op Op=new _Op();

					Op.L = ((_Lex)Lexs[iPos++]).L;
					Op.priority = 5;

					if(Op.L == "NumOfEdges")
						NumStack.Push((double)Edges.Length);
					else if(Op.L == "TypeOfObject")
						NumStack.Push((double)Type);
					else if(Op.L == "pi")
						NumStack.Push(Math.PI);
					else if(Op.L == "e")
						NumStack.Push(Math.E);
					else if(Op.L == "Line")
						NumStack.Push((double)EdgeType.Line);
					else if(Op.L == "Arc")
						NumStack.Push((double)EdgeType.Arc);
					else if(Op.L == "Other")
						NumStack.Push((double)EdgeType.Other);
					else if(Op.L == "Base")
						NumStack.Push((double)ObjectType.Base);
					else if(Op.L == "Hole")
						NumStack.Push((double)ObjectType.Hole);
					else if(Op.L == "PressIn")
						NumStack.Push((double)ObjectType.PressIn);
					else if(Op.L == "Punch")
						NumStack.Push((double)ObjectType.Punch);
					else
						OpStack.Push(Op);
				}

				// 
				else if(((_Lex)Lexs[iPos]).L == "(")
				{
					_Op Op=new _Op();

					Op.L = ((_Lex)Lexs[iPos++]).L;
					Op.priority = 0;
					OpStack.Push(Op);
				}

				// 
				else if(((_Lex)Lexs[iPos]).L == ")")
				{
					_Op Op=new _Op();
					double dOp1, dOp2;

					iPos++;
					while(OpStack.Count!=0)
					{
						Op = (_Op)OpStack.Pop();
						if(Op.L=="(")
						{
							//   
							Op = (_Op)OpStack.Pop();
							if(Op.L!="!")
								OpStack.Push(Op);
							else
							{
								if((double)NumStack.Pop()==0)
                                    NumStack.Push(1.0);
								else
									NumStack.Push(.0);
							}
							break;
						}

						switch(Op.L)
						{
							// 
							case "+":
								NumStack.Push((double)NumStack.Pop()+(double)NumStack.Pop());
								break;
							case "-":
								NumStack.Push((-(double)NumStack.Pop())+(double)NumStack.Pop());
								break;
							case "*":
								NumStack.Push((double)NumStack.Pop()*(double)NumStack.Pop());
								break;
							case "/":
								NumStack.Push(1.0/(double)NumStack.Pop()*(double)NumStack.Pop());
								break;
							// 
							case "==":
								if((double)NumStack.Pop()==(double)NumStack.Pop())
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case "!=":
								if((double)NumStack.Pop()==(double)NumStack.Pop())
									NumStack.Push(.0);
								else
									NumStack.Push(1.0);
								break;
							case "<":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1<dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case ">":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1>dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case "<=":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1<=dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case ">=":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1>=dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							// 
							case "&&":
								if((double)NumStack.Pop()*(double)NumStack.Pop()==0)
									NumStack.Push(.0);
								else
									NumStack.Push(1.0);
								break;
							case "||":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1!=0 || dOp2!=0)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							//
							case "sin":
								NumStack.Push(Math.Sin((double)NumStack.Pop()*Math.PI/180));
								break;
							case "cos":
								NumStack.Push(Math.Cos((double)NumStack.Pop()*Math.PI/180));
								break;
						}
					}
				}

				//-
				else if(((_Lex)Lexs[iPos]).type == 3)
				{
					int i = Convert.ToInt32(NumStack.Pop());
					if(i<0 || i> Edges.Length-1)
						NumStack.Push(.0);
					else
						switch(((_Lex)Lexs[iPos++]).L)
						{
							case ".Length":
								OpStack.Pop();
								NumStack.Push(Edges[i].dLength);
								break;
							case ".Angle":
								OpStack.Pop();
								NumStack.Push(Edges[i].dAngle*180.0/Math.PI);
								break;
							case ".Radius":
								OpStack.Pop();
								NumStack.Push(Edges[i].dRadius);
								break;
							case ".Type":
								OpStack.Pop();
								NumStack.Push((double)Edges[i].Type);
								break;
						}
				}

				//  '!'
				else if(((_Lex)Lexs[iPos]).L=="!")
				{
					_Op Op=new _Op();

					Op.L = ((_Lex)Lexs[iPos++]).L;
					OpStack.Push(Op);
				}

				//
				else if(((_Lex)Lexs[iPos]).type == 1)
				{
					_Op Op=new _Op(), Op1=new _Op();
					double dOp1, dOp2;

					Op.L = ((_Lex)Lexs[iPos]).L;
					if(Op.L=="*" || Op.L=="/")
						Op.priority = 2;
					else if(Op.L=="+" || Op.L=="-")
						Op.priority = 1;
					else if(Op.L=="==" || Op.L=="!=" || Op.L==">=" || Op.L=="<=" || Op.L==">" || Op.L=="<")
						Op.priority = .5;
					else if(Op.L=="&&" || Op.L=="||")
						Op.priority = .3;
					Op1 = (_Op)OpStack.Pop();
					//   
					if(Op1.priority>Op.priority)
					{
						switch(Op1.L)
						{
							// 
							case "+":
								NumStack.Push((double)NumStack.Pop()+(double)NumStack.Pop());
								break;
							case "-":
								NumStack.Push((-(double)NumStack.Pop())+(double)NumStack.Pop());
								break;
							case "*":
								NumStack.Push((double)NumStack.Pop()*(double)NumStack.Pop());
								break;
							case "/":
								NumStack.Push(1.0/(double)NumStack.Pop()*(double)NumStack.Pop());
								break;
							// 
							case "==":
								if((double)NumStack.Pop()==(double)NumStack.Pop())
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case "!=":
								if((double)NumStack.Pop()==(double)NumStack.Pop())
									NumStack.Push(.0);
								else
									NumStack.Push(1.0);
								break;
							case "<":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1<dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case ">":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1>dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case "<=":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1<=dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							case ">=":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1>=dOp2)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							// 
							case "&&":
								if((double)NumStack.Pop()*(double)NumStack.Pop()==0)
									NumStack.Push(.0);
								else
									NumStack.Push(1.0);
								break;
							case "||":
								dOp2 = (double)NumStack.Pop();
								dOp1 = (double)NumStack.Pop();
								if(dOp1!=0 || dOp2!=0)
									NumStack.Push(1.0);
								else
									NumStack.Push(.0);
								break;
							//
							case "sin":
								NumStack.Push(Math.Sin((double)NumStack.Pop()*Math.PI/180));
								break;
							case "cos":
								NumStack.Push(Math.Cos((double)NumStack.Pop()*Math.PI/180));
								break;
						}
						//OpStack.Push(Op);
						continue;
					}
						//   
					else
					{
						OpStack.Push(Op1);
						OpStack.Push(Op);
					}
					iPos++;
				}
			}

			while(OpStack.Count>0)
			{
				_Op Op=new _Op();
				double dOp1, dOp2;

				Op = (_Op)OpStack.Pop();

				switch(Op.L)
				{
					// 
					case "+":
						NumStack.Push((double)NumStack.Pop()+(double)NumStack.Pop());
						break;
					case "-":
						NumStack.Push((-(double)NumStack.Pop())+(double)NumStack.Pop());
						break;
					case "*":
						NumStack.Push((double)NumStack.Pop()*(double)NumStack.Pop());
						break;
					case "/":
						NumStack.Push(1.0/(double)NumStack.Pop()*(double)NumStack.Pop());
						break;
					// 
					case "==":
						if((double)NumStack.Pop()==(double)NumStack.Pop())
							NumStack.Push(1.0);
						else
							NumStack.Push(.0);
						break;
					case "!=":
						if((double)NumStack.Pop()==(double)NumStack.Pop())
							NumStack.Push(.0);
						else
							NumStack.Push(1.0);
						break;
					case "<":
						dOp2 = (double)NumStack.Pop();
						dOp1 = (double)NumStack.Pop();
						if(dOp1<dOp2)
							NumStack.Push(1.0);
						else
							NumStack.Push(.0);
						break;
					case ">":
						dOp2 = (double)NumStack.Pop();
						dOp1 = (double)NumStack.Pop();
						if(dOp1>dOp2)
							NumStack.Push(1.0);
						else
							NumStack.Push(.0);
						break;
					case "<=":
						dOp2 = (double)NumStack.Pop();
						dOp1 = (double)NumStack.Pop();
						if(dOp1<=dOp2)
							NumStack.Push(1.0);
						else
							NumStack.Push(.0);
						break;
					case ">=":
						dOp2 = (double)NumStack.Pop();
						dOp1 = (double)NumStack.Pop();
						if(dOp1>=dOp2)
							NumStack.Push(1.0);
						else
							NumStack.Push(.0);
						break;
					// 
					case "&&":
						if((double)NumStack.Pop()*(double)NumStack.Pop()==0)
							NumStack.Push(.0);
						else
							NumStack.Push(1.0);
						break;
					case "||":
						dOp2 = (double)NumStack.Pop();
						dOp1 = (double)NumStack.Pop();
						if(dOp1!=0 || dOp2!=0)
							NumStack.Push(1.0);
						else
							NumStack.Push(.0);
						break;
					//
					case "sin":
						NumStack.Push(Math.Sin((double)NumStack.Pop()));
						break;
					case "cos":
						NumStack.Push(Math.Cos((double)NumStack.Pop()));
						break;
				}
			}

			return (double)NumStack.Pop();
		}
	}

	public class PartParameters
	{
		public ObjectParameters[]	Objects;

		public int ReadParameters(string sFileName)
		{
			string s;
			StreamReader SR = File.OpenText(sFileName);

			s = SR.ReadLine();
			if(s.StartsWith("#NumOfObjects = "))
				Objects = new ObjectParameters[Convert.ToInt32(s.Substring("#NumOfObjects = ".Length))];
			else
				return 1;

			for(int i=0;i<Objects.Length;i++)
				Objects[i] = new ObjectParameters();

			foreach(ObjectParameters Object in Objects)
				if(Object.ReadParameters(SR)!=0)
					return 1;

			SR.Close();
			
			return 0;
		}

	}
}