//,     ,  

#include "StdAfx.h"
#include "Classes.h"
#include "PowerDesigner2.h"
#include "Utilities.h"
#include "DlgOffsets.h"
#include "DlgCoords.h"
#include "DlgBridges.h"

void CPoint3D::operator =(double *dPoint)
{
	X = dPoint[0];
	Y = dPoint[1];
	Z = dPoint[2];
}

double CPoint3D::operator -(CPoint3D Other)
{
	return sqrt((Other.X-X)*(Other.X-X) + (Other.Y-Y)*(Other.Y-Y) + (Other.Z-Z)*(Other.Z-Z));
}

bool CPoint3D::operator == (CPoint3D Other)
{
	if(X==Other.X && Y==Other.Y && Z==Other.Z)
		return true;
	else
		return false;
}

bool CPoint3D::operator != (CPoint3D Other)
{
	if(X!=Other.X || Y!=Other.Y || Z!=Other.Z)
		return true;
	else
		return false;
}

void CPoint3D::Fill(double *dPoint)
{
	dPoint[0] = X;
	dPoint[1] = Y;
	dPoint[2] = Z;
}

//     (  dG_DistTol)
bool CPoint3D::Equal(CPoint3D *Other)
{
	return *this - *Other < dG_DistTol;
}

//####################################################################################################################################################
//####################################################################################################################################################
//####################################################################################################################################################

CPart::CPart(void)
{
	iEdgesNum = 0;
	Edges = NULL;

	iHolesNum = 0;
	Holes = NULL;

	iPunchesNum = 0;
	Punches = NULL;

	iPressInsNum = 0;
	PressIns = NULL;
}

int CPart::ScanPart(tag_t Part, int iIntervalsNum)
{
uf_list_p_t PartFeature, FacesRounding, FacesBevel, FacesUndercutRounding, FacesUndercut, FacesPunchRounding, FacesPunchBevel;
uf_list_p_t FacesPressInRounding1, FacesPressInBevel, FacesPressInRounding2, FacesPressInBottom;
uf_list_p_t RoundingFeatures, ListTemp;
tag_t SheetBody, FaceBase, Face;
char *cOffset1, *cOffset2;
int iNumber1, iNumber2;
UF_UNDO_mark_id_t Undo;
double dOffset1, dOffset2;	//  ,    
CPoint3D Normal;	//  

	//  
	//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	UF_CALL(UF_UNDO_set_mark(UF_UNDO_invisible, NULL, &Undo));
	//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	UF_CALL(UF_MODL_create_list(&FacesRounding));
	UF_CALL(UF_MODL_create_list(&FacesBevel));
	UF_CALL(UF_MODL_create_list(&FacesUndercutRounding));
	UF_CALL(UF_MODL_create_list(&FacesUndercut));
	UF_CALL(UF_MODL_create_list(&FacesPunchRounding));
	UF_CALL(UF_MODL_create_list(&FacesPunchBevel));
	UF_CALL(UF_MODL_create_list(&FacesPressInRounding1));
	UF_CALL(UF_MODL_create_list(&FacesPressInBevel));
	UF_CALL(UF_MODL_create_list(&FacesPressInRounding2));
	UF_CALL(UF_MODL_create_list(&FacesPressInBottom));

	UF_CALL(UF_MODL_create_list(&RoundingFeatures));

	//        
	{
		UF_CALL(UF_MODL_ask_thicken_sheet_parms(Part, &SheetBody, &cOffset1, &cOffset2));
		t = fabs(atof(cOffset2)-atof(cOffset1));
		dOffset1 = atof(cOffset1);
		dOffset2 = atof(cOffset2);
		UF_free(cOffset1);
		UF_free(cOffset2);

		UF_CALL(UF_MODL_create_list(&PartFeature));
		UF_CALL(UF_MODL_put_list_item(PartFeature, Part));

		UF_CALL(UF_MODL_suppress_feature(PartFeature));
	}

	//   
	FindSignificantFaces(
		SheetBody,
		&FaceBase,
		&FacesRounding,
		&FacesBevel,
		&FacesUndercutRounding,
		&FacesUndercut,
		&FacesPunchRounding,
		&FacesPunchBevel,
		&FacesPressInRounding1,
		&FacesPressInBevel,
		&FacesPressInRounding2,
		&FacesPressInBottom);

	{//    
		// 
		UF_CALL(UF_OBJ_set_color(FaceBase, UF_OBJ_RED));
		// 
		UF_CALL(UF_MODL_ask_list_count(FacesRounding, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesRounding, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_BLUE));
		}
		// 
		UF_CALL(UF_MODL_ask_list_count(FacesBevel, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesBevel, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_MAGENTA));
		}
		//  
		UF_CALL(UF_MODL_ask_list_count(FacesUndercutRounding, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesUndercutRounding, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_OLIVE));
		}
		// 
		UF_CALL(UF_MODL_ask_list_count(FacesUndercut, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesUndercut, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_CYAN));
		}
		//  
		UF_CALL(UF_MODL_ask_list_count(FacesPunchBevel, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPunchBevel, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_ORANGE));
		}
		//   
		UF_CALL(UF_MODL_ask_list_count(FacesPressInRounding1, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPressInRounding1, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_WHITE));
		}
		//  
		UF_CALL(UF_MODL_ask_list_count(FacesPressInBevel, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPressInBevel, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_AQUAMARINE));
		}
		//   
		UF_CALL(UF_MODL_ask_list_count(FacesPressInRounding2, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPressInRounding2, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_BROWN));
		}
		//  
		UF_CALL(UF_MODL_ask_list_count(FacesPressInBottom, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPressInBottom, i, &Face));
			UF_CALL(UF_OBJ_set_color(Face, UF_OBJ_PINK));
		}
	}

	//     
//	ScanPress(FaceBase, iIntervalsNum);

	//  
	{
		UF_CALL(UF_MODL_ask_list_count(FacesRounding, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesRounding, i, &Face));
			UF_CALL(UF_MODL_ask_face_feats(Face, &ListTemp));
			UF_CALL(UF_MODL_ask_list_count(ListTemp, &iNumber2));
			for(int j=0;j<iNumber2;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(ListTemp, j, &Face));
				UF_CALL(UF_MODL_put_list_item(RoundingFeatures, Face));
			}
		}
		UF_CALL(UF_MODL_ask_list_count(FacesUndercutRounding, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesUndercutRounding, i, &Face));
			UF_CALL(UF_MODL_ask_face_feats(Face, &ListTemp));
			UF_CALL(UF_MODL_ask_list_count(ListTemp, &iNumber2));
			for(int j=0;j<iNumber2;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(ListTemp, j, &Face));
				UF_CALL(UF_MODL_put_list_item(RoundingFeatures, Face));
			}
		}
		UF_CALL(UF_MODL_ask_list_count(FacesPunchRounding, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPunchRounding, i, &Face));
			UF_CALL(UF_MODL_ask_face_feats(Face, &ListTemp));
			UF_CALL(UF_MODL_ask_list_count(ListTemp, &iNumber2));
			for(int j=0;j<iNumber2;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(ListTemp, j, &Face));
				UF_CALL(UF_MODL_put_list_item(RoundingFeatures, Face));
			}
		}
		UF_CALL(UF_MODL_ask_list_count(FacesPressInRounding1, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPressInRounding1, i, &Face));
			UF_CALL(UF_MODL_ask_face_feats(Face, &ListTemp));
			UF_CALL(UF_MODL_ask_list_count(ListTemp, &iNumber2));
			for(int j=0;j<iNumber2;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(ListTemp, j, &Face));
				UF_CALL(UF_MODL_put_list_item(RoundingFeatures, Face));
			}
		}
		UF_CALL(UF_MODL_ask_list_count(FacesPressInRounding2, &iNumber1));
		for(int i=0;i<iNumber1;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPressInRounding2, i, &Face));
			UF_CALL(UF_MODL_ask_face_feats(Face, &ListTemp));
			UF_CALL(UF_MODL_ask_list_count(ListTemp, &iNumber2));
			for(int j=0;j<iNumber2;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(ListTemp, j, &Face));
				UF_CALL(UF_MODL_put_list_item(RoundingFeatures, Face));
			}
		}
		UF_CALL(UF_MODL_suppress_feature(RoundingFeatures));
	}

	CreateOuterPlanes(FaceBase, iIntervalsNum);
	CreateInnerPlanes(FaceBase, FacesPunchBevel, FacesPressInBevel, iIntervalsNum*dG_InnerCoef);

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].ScanRealPoints(FaceBase, FacesBevel, FacesUndercut);

	for(int i=0;i<iPunchesNum;i++)
	for(int j=0;j<Punches[i].iEdgesNum;j++)
		Punches[i].Edges[j].ScanRealPoints(FacesPunchBevel);

	for(int i=0;i<iPressInsNum;i++)
	for(int j=0;j<PressIns[i].iEdgesNum;j++)
		PressIns[i].Edges[j].ScanRealPoints(FacesPressInBevel);

	//      
	for(int i=0;i<iEdgesNum;i++)
		if(!Edges[i].bPS && !Edges[i].bNothing)
			Edges[i].PostCalcA(FaceBase);

	//     
	UF_CALL(UF_MODL_unsuppress_feature(RoundingFeatures));

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].ScanPoints(FacesBevel, FacesUndercut, FacesRounding, FacesUndercutRounding);

	for(int i=0;i<iPunchesNum;i++)
	for(int j=0;j<Punches[i].iEdgesNum;j++)
		Punches[i].Edges[j].ScanPoints(FacesPunchBevel, FacesPunchRounding);

	for(int i=0;i<iPressInsNum;i++)
	for(int j=0;j<PressIns[i].iEdgesNum;j++)
		PressIns[i].Edges[j].ScanPoints(FacesPressInBevel, FacesPressInRounding1, FacesPressInRounding2, FacesPressInBottom);

	//      
	for(int i=0;i<iEdgesNum;i++)
		if(!Edges[i].bPS && !Edges[i].bNothing)
			Edges[i].PostCalcB();

	DeleteOuterPlanes();
	DeleteInnerPlanes();

	//    
	LineUp();

	//  
	{
		double point[3], box[6], radius, rad_data;
		int type, norm_dir;
		UF_CALL(UF_MODL_ask_face_data(FaceBase, &type, point, Service.dBaseNormal, box, &radius, &rad_data, &norm_dir));
		// 
		Service.dBaseNormal[0] *= dOffset1<0 || dOffset2<0 ? -1 : 1;
		Service.dBaseNormal[1] *= dOffset1<0 || dOffset2<0 ? -1 : 1;
		Service.dBaseNormal[2] *= dOffset1<0 || dOffset2<0 ? -1 : 1;
//		dBaseNormal[0] *= fabs(dThickness)/dThickness;
//		dBaseNormal[1] *= fabs(dThickness)/dThickness;
//		dBaseNormal[2] *= fabs(dThickness)/dThickness;

		Normal = Service.dBaseNormal;
	}

	m_BaseFace = FaceBase;

	//    
	{
//		double dFrom[3], dTo[3];
//		double dSumAngle, dAngle;
//		for(int i=0;i<iEdgesNum;i++)
		{
//			UF_CALL(UF_CURVE_ask_curve_turn_angle(Edges[i].
		}
/*		Edges[iEdgesNum-1].Sections[Edges[iEdgesNum-1].iSectionsNum-1].PlaneSectionNormal.Fill(dFrom);
		Edges[0].Sections[0].PlaneSectionNormal.Fill(dTo);
		UF_VEC3_angle_between(dFrom, dTo, dBaseNormal, &dSumAngle);
		dSumAngle > PI/2 ? dSumAngle -= PI/2 : 0;
		for(int i=0;i<iEdgesNum;i++)
		for(int j=0;j<Edges[i].iSectionsNum-1;j++)
		{
			Edges[i].Sections[j].PlaneSectionNormal.Fill(dFrom);
			Edges[i].Sections[j+1].PlaneSectionNormal.Fill(dTo);
			UF_VEC3_angle_between(dFrom, dTo, dBaseNormal, &dAngle);
			dAngle > PI/2 ? dAngle -= PI/2 : 0;
			dSumAngle += dAngle;
		}
		for(int i=0;i<iEdgesNum-1;i++)
		{
			Edges[i].Sections[Edges[i].iSectionsNum-1].PlaneSectionNormal.Fill(dFrom);
			Edges[i+1].Sections[0].PlaneSectionNormal.Fill(dTo);
			UF_VEC3_angle_between(dFrom, dTo, dBaseNormal, &dAngle);
			dAngle > PI/2 ? dAngle -= PI/2 : 0;
			dSumAngle += dAngle;
		}*/
	}

	//       0XY
	RotateAndMoveToZero(Normal);

	//    
	UF_CALL(UF_MODL_unsuppress_feature(PartFeature));

	UF_CALL(UF_MODL_delete_list(&FacesRounding));
	UF_CALL(UF_MODL_delete_list(&FacesBevel));
	UF_CALL(UF_MODL_delete_list(&FacesUndercutRounding));
	UF_CALL(UF_MODL_delete_list(&FacesUndercut));
	UF_CALL(UF_MODL_delete_list(&FacesPunchRounding));
	UF_CALL(UF_MODL_delete_list(&FacesPunchBevel));
	UF_CALL(UF_MODL_delete_list(&FacesPressInRounding1));
	UF_CALL(UF_MODL_delete_list(&FacesPressInBevel));
	UF_CALL(UF_MODL_delete_list(&FacesPressInRounding2));
	UF_CALL(UF_MODL_delete_list(&FacesPressInBottom));

	UF_CALL(UF_MODL_delete_list(&PartFeature));
	UF_CALL(UF_MODL_delete_list(&RoundingFeatures));

	//   
	//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	UF_CALL(UF_UNDO_undo_to_mark(Undo, NULL));
	//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	return 0;
}

//        
int CPart::CreateOuterPlanes(tag_t Base, int iIntervalsNum)
{
uf_list_p_t OuterEdges;
tag_t Curve;
double dLength, dEdgesLength=0, dSegmentSize;
int iNumOfPoints;

	UF_CALL(UF_MODL_create_list(&OuterEdges));
	GetOuterEdges(Base, &OuterEdges);
	UF_CALL(UF_MODL_ask_list_count(OuterEdges, &iEdgesNum));

	//  
	Edges = new CEdge[iEdgesNum];

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(OuterEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));
		dEdgesLength += dLength;
	}

	//       
	dSegmentSize = dEdgesLength/double(iIntervalsNum);
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(OuterEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));

		iNumOfPoints = dLength/dSegmentSize + 1.5;
		if(iNumOfPoints<4)
			iNumOfPoints = 4;

		Edges[i].CreateOuterPlanes(Base, Curve, iNumOfPoints);
	}

	return 0;
}

//    CreateOuterPlanes  
void CPart::DeleteOuterPlanes(void)
{
int iNumber1, iNumber2;

	for(int i=0;i<iEdgesNum;i++)
	for(int j=0;j<Edges[i].iSectionsNum;j++)
		if(Edges[i].Sections[j].PlaneSection!=NULL_TAG)
			UF_CALL(UF_ASSEM_remove_instance(Edges[i].Sections[j].PlaneSection));
}

//        
int CPart::CreateInnerPlanes(tag_t Base, uf_list_p_t FacesPunchBevel, uf_list_p_t FacesPressInBevel, int iIntervalsNum)
{
uf_loop_p_t Loops;
uf_list_p_t Edges, Faces;
tag_t Curve, Face;
int iLoops, iType, iFaces, iNumber1, iNumber2;
int iHolesCursor=0, iPunchesCursor=0, iPressInsCursor=0;

	UF_CALL(UF_MODL_ask_face_loops(Base, &Loops));
	UF_CALL(UF_MODL_ask_loop_list_count(Loops, &iLoops));

	//  ,   
	for(int i=0;i<iLoops;i++)
	{
		UF_CALL(UF_MODL_create_list(&Edges));

		UF_CALL(UF_MODL_ask_loop_list_item(Loops, i, &iType, &Edges));
		if(iType==2)
		{
			//   .        (),    ( ) 
			UF_CALL(UF_MODL_ask_list_item(Edges, 0, &Curve));
			UF_CALL(UF_MODL_ask_edge_faces(Curve, &Faces));
			UF_CALL(UF_MODL_ask_list_count(Faces, &iFaces));
			if(iFaces==1)
				iHolesNum++;
			else
			{
				UF_CALL(UF_MODL_delete_list_item(&Faces, Base));
				//,       Face
				UF_CALL(UF_MODL_ask_list_item(Faces, 0, &Face));

				UF_CALL(UF_MODL_ask_list_count(FacesPunchBevel, &iNumber1));
				UF_CALL(UF_MODL_put_list_item(FacesPunchBevel, Face));
				UF_CALL(UF_MODL_ask_list_count(FacesPunchBevel, &iNumber2));
				if(iNumber1==iNumber2)
					iPunchesNum++;
				else
					UF_CALL(UF_MODL_delete_list_item(&FacesPunchBevel, Face));

				UF_CALL(UF_MODL_ask_list_count(FacesPressInBevel, &iNumber1));
				if(iNumber1)
				{
					UF_CALL(UF_MODL_put_list_item(FacesPressInBevel, Face));
					UF_CALL(UF_MODL_ask_list_count(FacesPressInBevel, &iNumber2));
					if(iNumber1==iNumber2)
						iPressInsNum++;
					else
						UF_CALL(UF_MODL_delete_list_item(&FacesPressInBevel, Face));
				}
			}
			UF_CALL(UF_MODL_delete_list(&Faces));
		}

		//UF_CALL(UF_MODL_delete_list(&Edges));
	}

	// ,   
	Holes = new CHole[iHolesNum];
	Punches = new CPunch[iPunchesNum];
	PressIns = new CPressIn[iPressInsNum];

	// ,   
	for(int i=0;i<iLoops;i++)
	{
		UF_CALL(UF_MODL_create_list(&Edges));

		UF_CALL(UF_MODL_ask_loop_list_item(Loops, i, &iType, &Edges));
		if(iType==2)
		{
			//   .        (),    ( ) 
			UF_CALL(UF_MODL_ask_list_item(Edges, 0, &Curve));
			UF_CALL(UF_MODL_ask_edge_faces(Curve, &Faces));
			UF_CALL(UF_MODL_ask_list_count(Faces, &iFaces));
			if(iFaces==1)
				Holes[iHolesCursor++].ScanHole(Edges, iIntervalsNum);
			else
			{
				UF_CALL(UF_MODL_delete_list_item(&Faces, Base));
				//,       Face
				UF_CALL(UF_MODL_ask_list_item(Faces, 0, &Face));

				UF_CALL(UF_MODL_ask_list_count(FacesPunchBevel, &iNumber1));
				UF_CALL(UF_MODL_put_list_item(FacesPunchBevel, Face));
				UF_CALL(UF_MODL_ask_list_count(FacesPunchBevel, &iNumber2));
				if(iNumber1==iNumber2)
					Punches[iPunchesCursor++].CreateInnerPlanes(Edges, iIntervalsNum);
				else
					UF_CALL(UF_MODL_delete_list_item(&FacesPunchBevel, Face));

				UF_CALL(UF_MODL_ask_list_count(FacesPressInBevel, &iNumber1));
				if(iNumber1)
				{
					UF_CALL(UF_MODL_put_list_item(FacesPressInBevel, Face));
					UF_CALL(UF_MODL_ask_list_count(FacesPressInBevel, &iNumber2));
					if(iNumber1==iNumber2)
						PressIns[iPressInsCursor++].CreateInnerPlanes(Edges, iIntervalsNum);
					else
						UF_CALL(UF_MODL_delete_list_item(&FacesPressInBevel, Face));
				}
			}
			UF_CALL(UF_MODL_delete_list(&Faces));
		}
	}

	UF_CALL(UF_MODL_delete_loop_list(&Loops));

	return 0;
}

//    CreateInnerPlanes  
void CPart::DeleteInnerPlanes(void)
{
int iNumber1, iNumber2, iNumber3;

	for(int i=0;i<iPunchesNum;i++)
	for(int j=0;j<Punches[i].iEdgesNum;j++)
	for(int k=0;k<Punches[i].Edges[j].iSectionsNum;k++)
		if(Punches[i].Edges[j].Sections[k].PlaneSection!=NULL_TAG)
			UF_CALL(UF_ASSEM_remove_instance(Punches[i].Edges[j].Sections[k].PlaneSection));

	for(int i=0;i<iPressInsNum;i++)
	for(int j=0;j<PressIns[i].iEdgesNum;j++)
	for(int k=0;k<PressIns[i].Edges[j].iSectionsNum;k++)
		if(PressIns[i].Edges[j].Sections[k].PlaneSection!=NULL_TAG)
			UF_CALL(UF_ASSEM_remove_instance(PressIns[i].Edges[j].Sections[k].PlaneSection));
}

//      
/*void CPart::LineUp(void)
{
	for(int i=1;i<iEdgesNum;i++)
		if(Edges[i].Sections[0].PointRealBase != Edges[i-1].Sections[Edges[i-1].iSectionsNum-1].PointRealBase)
			Edges[i].Reverse();

	for(int i=0;i<iHolesNum;i++)
	for(int j=1;j<Holes[i].iEdgesNum;j++)
		if(Holes[i].Edges[j].Sections[0].PointRealBase != Holes[i].Edges[j-1].Sections[Holes[i].Edges[j-1].iSectionsNum-1].PointRealBase)
			Holes[i].Edges[j].Reverse();

	for(int i=0;i<iPunchesNum;i++)
	for(int j=1;j<Punches[i].iEdgesNum;j++)
		if(Punches[i].Edges[j].Sections[0].PointRealBase != Punches[i].Edges[j-1].Sections[Punches[i].Edges[j-1].iSectionsNum-1].PointRealBase)
			Punches[i].Edges[j].Reverse();

	for(int i=0;i<iPressInsNum;i++)
	for(int j=1;j<PressIns[i].iEdgesNum;j++)
		if(PressIns[i].Edges[j].Sections[0].PointRealBase != PressIns[i].Edges[j-1].Sections[PressIns[i].Edges[j-1].iSectionsNum-1].PointRealBase)
			PressIns[i].Edges[j].Reverse();

	if(Edges[1].Sections[0].PointRealBase != Edges[0].Sections[Edges[0].iSectionsNum-1].PointRealBase)
	{
		Edges[0].Reverse();
		LineUp();
	}

	for(int i=0;i<iHolesNum;i++)
	for(int j=1;j<Holes[i].iEdgesNum;j++)
		if(Holes[i].Edges[1].Sections[0].PointRealBase != Holes[i].Edges[0].Sections[Holes[i].Edges[j-1].iSectionsNum-1].PointRealBase)
		{
			Holes[i].Edges[0].Reverse();
			LineUp();
		}

	for(int i=0;i<iPunchesNum;i++)
	for(int j=1;j<Punches[i].iEdgesNum;j++)
		if(Punches[i].Edges[1].Sections[0].PointRealBase != Punches[i].Edges[0].Sections[Punches[i].Edges[j-1].iSectionsNum-1].PointRealBase)
		{
			Punches[i].Edges[0].Reverse();
			LineUp();
		}

	for(int i=0;i<iPressInsNum;i++)
	for(int j=1;j<PressIns[i].iEdgesNum;j++)
		if(PressIns[i].Edges[1].Sections[0].PointRealBase != PressIns[i].Edges[0].Sections[PressIns[i].Edges[j-1].iSectionsNum-1].PointRealBase)
		{
			PressIns[i].Edges[0].Reverse();
			LineUp();
		}
}*/


void CPart::LineUp(void)
{
	for(int i=1;i<iEdgesNum;i++)
		if(!Edges[i].Sections[0].PointRealBase.Equal(&Edges[i-1].Sections[Edges[i-1].iSectionsNum-1].PointRealBase))
			Edges[i].Reverse();

	for(int i=0;i<iHolesNum;i++)
	for(int j=1;j<Holes[i].iEdgesNum;j++)
		if(!Holes[i].Edges[j].Sections[0].PointRealBase.Equal(&Holes[i].Edges[j-1].Sections[Holes[i].Edges[j-1].iSectionsNum-1].PointRealBase))
			Holes[i].Edges[j].Reverse();

	for(int i=0;i<iPunchesNum;i++)
	for(int j=1;j<Punches[i].iEdgesNum;j++)
		if(!Punches[i].Edges[j].Sections[0].PointRealBase.Equal(&Punches[i].Edges[j-1].Sections[Punches[i].Edges[j-1].iSectionsNum-1].PointRealBase))
			Punches[i].Edges[j].Reverse();

	for(int i=0;i<iPressInsNum;i++)
	for(int j=1;j<PressIns[i].iEdgesNum;j++)
		if(!PressIns[i].Edges[j].Sections[0].PointRealBase.Equal(&PressIns[i].Edges[j-1].Sections[PressIns[i].Edges[j-1].iSectionsNum-1].PointRealBase))
			PressIns[i].Edges[j].Reverse();

	if(!Edges[1].Sections[0].PointRealBase.Equal(&Edges[0].Sections[Edges[0].iSectionsNum-1].PointRealBase))
	{
		Edges[0].Reverse();
		LineUp();
	}

	for(int i=0;i<iHolesNum;i++)
	for(int j=1;j<Holes[i].iEdgesNum;j++)
		if(!Holes[i].Edges[1].Sections[0].PointRealBase.Equal(&Holes[i].Edges[0].Sections[Holes[i].Edges[j-1].iSectionsNum-1].PointRealBase))
		{
			Holes[i].Edges[0].Reverse();
			LineUp();
		}

	for(int i=0;i<iPunchesNum;i++)
	for(int j=1;j<Punches[i].iEdgesNum;j++)
		if(!Punches[i].Edges[1].Sections[0].PointRealBase.Equal(&Punches[i].Edges[0].Sections[Punches[i].Edges[j-1].iSectionsNum-1].PointRealBase))
		{
			Punches[i].Edges[0].Reverse();
			LineUp();
		}

	for(int i=0;i<iPressInsNum;i++)
	for(int j=1;j<PressIns[i].iEdgesNum;j++)
		if(!PressIns[i].Edges[1].Sections[0].PointRealBase.Equal(&PressIns[i].Edges[0].Sections[PressIns[i].Edges[j-1].iSectionsNum-1].PointRealBase))
		{
			PressIns[i].Edges[0].Reverse();
			LineUp();
		}
}


// 
void CPart::Correct(double E, double K, double n)
{
	//     
	for(int i=0;i<iEdgesNum;i++)
		if(!Edges[i].bPS && !Edges[i].bNothing)
			for(int j=0;j<Edges[i].iSectionsNum;j++)
				Edges[i].Sections[j].PostCalc(t, E, K, n, Service.dBaseNormal, m_BaseFace);
}

//        
//   , . .    
//    
void CPart::RotateAndMoveToZero(CPoint3D Normal)
{
	double DZ;
	CPoint3D Temp;

	//   --      OZ  
	CPoint3D Axis;
	Axis.X = Normal.Y;
	Axis.Y = -Normal.X;
	Axis.Z = 0;
	if(Axis.X==0 && Axis.Y==0 && Axis.Z==0)
		Axis.X = 1;

	//  
	double dAngle = acos(Normal.Z/sqrt(Normal.X*Normal.X + Normal.Y*Normal.Y + Normal.Z*Normal.Z));

	//   
	Temp = Service.dBaseNormal;
	::Rotate(&Temp, Axis, dAngle);
	Temp.Fill(Service.dBaseNormal);
//	dBaseNormal[0] *= -1.0;//!!!
//	dBaseNormal[1] *= -1.0;
//	dBaseNormal[2] *= -1.0;

	//  
	if(dAngle!=0)
	{
		for(int i=0;i<iPressEdgesNum;i++)
		for(int j=0;j<iPressPointsNum[i];j++)
			::Rotate(&PointsPress[i][j], Axis, dAngle);

		for(int i=0;i<iEdgesNum;i++)
		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			::Rotate(&Edges[i].Sections[j].PointBase, Axis, dAngle);
			::Rotate(&Edges[i].Sections[j].PointBevel, Axis, dAngle);
			::Rotate(&Edges[i].Sections[j].PointRealBase, Axis, dAngle);
			::Rotate(&Edges[i].Sections[j].PointRealBevel, Axis, dAngle);
			::Rotate(&Edges[i].Sections[j].PointRounding, Axis, dAngle);
			::Rotate(&Edges[i].Sections[j].PointUndercut, Axis, dAngle);
			::Rotate(&Edges[i].Sections[j].PointUndercutRounding, Axis, dAngle);
			::Rotate(&Edges[i].Sections[j].PlaneSectionNormal, Axis, dAngle);	//!!!
		}

		for(int i=0;i<iHolesNum;i++)
		for(int j=0;j<Holes[i].iEdgesNum;j++)
		for(int k=0;k<Holes[i].Edges[j].iSectionsNum;k++)
			::Rotate(&Holes[i].Edges[j].Sections[k].PointRealBase, Axis, dAngle);

		for(int i=0;i<iPunchesNum;i++)
		for(int j=0;j<Punches[i].iEdgesNum;j++)
		for(int k=0;k<Punches[i].Edges[j].iSectionsNum;k++)
		{
			::Rotate(&Punches[i].Edges[j].Sections[k].PointBase, Axis, dAngle);
			::Rotate(&Punches[i].Edges[j].Sections[k].PointBevel, Axis, dAngle);
			::Rotate(&Punches[i].Edges[j].Sections[k].PointRealBase, Axis, dAngle);
			::Rotate(&Punches[i].Edges[j].Sections[k].PointRounding, Axis, dAngle);
		}

		for(int i=0;i<iPressInsNum;i++)
		for(int j=0;j<PressIns[i].iEdgesNum;j++)
		for(int k=0;k<PressIns[i].Edges[j].iSectionsNum;k++)
		{
			::Rotate(&PressIns[i].Edges[j].Sections[k].PointBase, Axis, dAngle);
			::Rotate(&PressIns[i].Edges[j].Sections[k].PointBevel, Axis, dAngle);
			::Rotate(&PressIns[i].Edges[j].Sections[k].PointBottom, Axis, dAngle);
			::Rotate(&PressIns[i].Edges[j].Sections[k].PointRealBase, Axis, dAngle);
			::Rotate(&PressIns[i].Edges[j].Sections[k].PointRealBottom, Axis, dAngle);
			::Rotate(&PressIns[i].Edges[j].Sections[k].PointRounding1, Axis, dAngle);
		}
	}

	//  
	DZ = -Edges[0].Sections[0].PointRealBase.Z;

	//  
	for(int i=0;i<iPressEdgesNum;i++)
	for(int j=0;j<iPressPointsNum[i];j++)
		PointsPress[i][j].Z += DZ;

	for(int i=0;i<iEdgesNum;i++)
	for(int j=0;j<Edges[i].iSectionsNum;j++)
	{
		Edges[i].Sections[j].PointBase.Z += DZ;
		Edges[i].Sections[j].PointBevel.Z += DZ;
		Edges[i].Sections[j].PointRealBase.Z += DZ;
		Edges[i].Sections[j].PointRealBevel.Z += DZ;
		Edges[i].Sections[j].PointRounding.Z += DZ;
		Edges[i].Sections[j].PointUndercut.Z += DZ;
		Edges[i].Sections[j].PointUndercutRounding.Z += DZ;
	}

	for(int i=0;i<iHolesNum;i++)
	for(int j=0;j<Holes[i].iEdgesNum;j++)
	for(int k=0;k<Holes[i].Edges[j].iSectionsNum;k++)
		Holes[i].Edges[j].Sections[k].PointRealBase.Z += DZ;

	for(int i=0;i<iPunchesNum;i++)
	for(int j=0;j<Punches[i].iEdgesNum;j++)
	for(int k=0;k<Punches[i].Edges[j].iSectionsNum;k++)
	{
		Punches[i].Edges[j].Sections[k].PointBase.Z += DZ;
		Punches[i].Edges[j].Sections[k].PointBevel.Z += DZ;
		Punches[i].Edges[j].Sections[k].PointRealBase.Z += DZ;
		Punches[i].Edges[j].Sections[k].PointRounding.Z += DZ;
	}

	for(int i=0;i<iPressInsNum;i++)
	for(int j=0;j<PressIns[i].iEdgesNum;j++)
	for(int k=0;k<PressIns[i].Edges[j].iSectionsNum;k++)
	{
		PressIns[i].Edges[j].Sections[k].PointBase.Z += DZ;
		PressIns[i].Edges[j].Sections[k].PointBevel.Z += DZ;
		PressIns[i].Edges[j].Sections[k].PointBottom.Z += DZ;
		PressIns[i].Edges[j].Sections[k].PointRealBase.Z += DZ;
		PressIns[i].Edges[j].Sections[k].PointRealBottom.Z += DZ;
		PressIns[i].Edges[j].Sections[k].PointRounding1.Z += DZ;
	}
}

//      
//   false   
bool CPart::AskParametersInteractive(void)
{
//	if(!AskCoordsInteractive())
//		return false;

	if(!AskDistancesInteractive())
		return false;

//	if(!AskBridgesInteractive())
//		return false;

	return true;

//7777777777777777777777

tag_t *Splines = new tag_t[iEdgesNum], UGPart;
int iDegree, iNumber = 1, iPeriodicity;
CString string;
UF_CURVE_pt_slope_crvatr_s *Points;
CDlgOffsets DlgOffsets;
CDlgCoords DlgCoords(CWnd::FromHandle((HWND)UF_UI_get_default_parent()));
CDlgCoords *DlgCoords2 = new CDlgCoords();

	//  
	string.Format("Matrix %d", iNumber);
	while(UF_PART_new(string.GetBuffer(), 1, &UGPart))
		string.Format("Matrix %d", ++iNumber);

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Splines[i]));

		// 
		if(Edges[i].bNothing)
			UF_CALL(UF_OBJ_set_color(Splines[i], UF_OBJ_YELLOW));
		else
			UF_CALL(UF_OBJ_set_color(Splines[i], UF_OBJ_GREEN));

		delete [] Points;
	}

	//       
	UF_CALL(UF_VIEW_fit_view(NULL_TAG, .9));

	// 
	{
		double dPoint1[3]={0,0,0}, dPoint2[3]={1000,0,0}, dX[3], dY[3], dZ[3]={0,0,100}, dOffset[3];
		static char cMenu[3][37+1] = {
						"  ",
						"    0X",
						""};
		int iChoice, iResponse, iInput[2]={0,0};
		UF_OBJ_disp_props_t Attribs;
	
        Attribs.color = UF_OBJ_RED;
		Attribs.font = UF_OBJ_FONT_SOLID;
		Attribs.line_width = UF_OBJ_WIDTH_THICK;
		UF_CALL(UF_DISP_display_temporary_line(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, dPoint2, &Attribs));
		UF_CALL(UF_DISP_display_temporary_point(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, &Attribs, UF_DISP_DIAMOND));

		while(1)
		{
			iChoice = uc1603(" ", 3, cMenu, 3);

			if (iChoice == 1 || iChoice == 2) 
			{
				break;
			}

			if(iChoice==5)			//"  "
			{
				UF_CALL(uc1616("  ", iInput, 1, dPoint1));
				UF_DISP_refresh();
				UF_CALL(UF_DISP_display_temporary_line(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, dPoint2, &Attribs));
				UF_CALL(UF_DISP_display_temporary_point(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, &Attribs, UF_DISP_DIAMOND));
			}
			else if(iChoice==6)		//"    0X"
			{
				UF_CALL(uc1616("  ", iInput, 1, dPoint2));
				UF_DISP_refresh();
				UF_CALL(UF_DISP_display_temporary_line(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, dPoint2, &Attribs));
				UF_CALL(UF_DISP_display_temporary_point(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, &Attribs, UF_DISP_DIAMOND));
			}
			else if(iChoice==7)		//""
			{
				if(dPoint1[0]==dPoint2[0] && dPoint1[1]==dPoint2[1])
					uc1601("    ", 1);
				else
					break;
			}
		}

		if(iChoice==1 || iChoice==2)
			return false;

		//   
		dX[0] = dPoint2[0] - dPoint1[0];
		dX[1] = dPoint2[1] - dPoint1[1];
		dX[2] = 0;
		dY[0] = -dX[1];
		dY[1] = dX[0];
		dY[2] = 0;

		// 
		for(int i=0;i<iEdgesNum;i++)
		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			TransformCoordinateSystem(&Edges[i].Sections[j].PointBase, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].PointRealBase, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].PointRounding, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].PointBevel, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].PointRealBevel, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].PointUndercutRounding, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].PointUndercut, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].CalcPoints.PointRounding, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].CalcPoints.PointBevel, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Edges[i].Sections[j].PlaneSectionNormal, dX, dY, dZ, dPoint1);
		}

		for(int i=0;i<iHolesNum;i++)
		for(int j=0;j<Holes[i].iEdgesNum;j++)
		for(int k=0;k<Holes[i].Edges[j].iSectionsNum;k++)
			TransformCoordinateSystem(&Holes[i].Edges[j].Sections[k].PointRealBase, dX, dY, dZ, dPoint1);

		for(int i=0;i<iPunchesNum;i++)
		for(int j=0;j<Punches[i].iEdgesNum;j++)
		for(int k=0;k<Punches[i].Edges[j].iSectionsNum;k++)
		{
			TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointBase, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointBevel, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointRealBase, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointRounding, dX, dY, dZ, dPoint1);
		}

		for(int i=0;i<iPressInsNum;i++)
		for(int j=0;j<PressIns[i].iEdgesNum;j++)
		for(int k=0;k<PressIns[i].Edges[j].iSectionsNum;k++)
		{
			TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointBase, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointBevel, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointBottom, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointRealBase, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointRealBottom, dX, dY, dZ, dPoint1);
			TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointRounding1, dX, dY, dZ, dPoint1);
		}
	}

/*	SetParent(DlgCoords.m_hWnd, (HWND)UF_UI_get_default_parent());

	if(DlgCoords.DoModal()==IDOK)
	{
	}*/

	DlgOffsets.DoModal(Splines, Edges, iEdgesNum);

	delete [] Splines;
}

//    
//   false   
bool CPart::AskDistancesInteractive(void)
{
tag_t *Splines = new tag_t[iEdgesNum], UGPart;
int iDegree, iNumber = 1, iPeriodicity, iRes;
CString string;
UF_CURVE_pt_slope_crvatr_s *Points;
CDlgOffsets DlgOffsets;

	//  
	string.Format("Matrix %d", iNumber);
	while(UF_PART_new(string.GetBuffer(), 1, &UGPart))
		string.Format("Matrix %d", ++iNumber);

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Splines[i]));

		// 
		if(Edges[i].bNothing)
			UF_CALL(UF_OBJ_set_color(Splines[i], UF_OBJ_YELLOW));
		else
			UF_CALL(UF_OBJ_set_color(Splines[i], UF_OBJ_GREEN));

		delete [] Points;
	}

	//       
	UF_CALL(UF_VIEW_fit_view(NULL_TAG, .9));

	iRes = DlgOffsets.DoModal(Splines, Edges, iEdgesNum);

	delete [] Splines;

	UF_CALL(UF_PART_close(UGPart, 0, 1));

	return iRes==IDOK;
}

bool CPart::AskBridgesInteractive(void)
{
tag_t *Splines = new tag_t[iEdgesNum], UGPart, *OffCurve;
int iDegree, iNumber=1, iPeriodicity, iRes;
CString string;
UF_CURVE_pt_slope_crvatr_s *Points;
UF_CURVE_offset_data_t OffData;
UF_CURVE_offset_distance_data_t OffDist;
UF_STRING_t Curves;
CDlgBridges DlgBridges;

	//  
	string.Format("Matrix %d", iNumber);
	while(UF_PART_new(string.GetBuffer(), 1, &UGPart))
		string.Format("Matrix %d", ++iNumber);

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Splines[i]));

		// ,  
		if(Edges[i].bNothing && Edges[i].dOffset!=0)
		{
			int t1=1, t2=1;

			Curves.num = 1;
			Curves.string = &t1;
			Curves.dir = &t2;
			Curves.id = &Splines[i];

			//OffCurve = &Splines[i];

			string.Format("%f", Edges[i].dOffset);
			OffDist.distance = string.GetBuffer();

			OffData.approximation_tolerance = .1;
			OffData.string_tolerance = .1;
			OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
			OffData.offset_def.distance_type1 = &OffDist;
			OffData.input_curves = &Curves;

			UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

			UF_CALL(UF_ASSEM_remove_instance(Splines[i]));
			Splines[i] = *OffCurve;

			UF_free(OffCurve);
		}

		// 
		if(Edges[i].bNothing)
			UF_CALL(UF_OBJ_set_color(Splines[i], UF_OBJ_YELLOW));
		else
			UF_CALL(UF_OBJ_set_color(Splines[i], UF_OBJ_GREEN));

		delete [] Points;
	}

	//       
	UF_CALL(UF_VIEW_fit_view(NULL_TAG, .9));

	iRes = DlgBridges.DoModal(Splines, Edges, iEdgesNum);

	delete [] Splines;

	//UF_CALL(UF_PART_close(UGPart, 0, 1));

	return iRes==IDOK;
}

//     
//   false   
bool CPart::AskCoordsInteractive(void)
{
tag_t *Splines = new tag_t[iEdgesNum], UGPart;
int iDegree, iNumber=1, iPeriodicity;
CString string;
UF_CURVE_pt_slope_crvatr_s *Points;
double dPoint1[3]={0,0,0}, dPoint2[3]={1000,0,0}, dX[3], dY[3], dZ[3]={0,0,100}, dOffset[3], dZeros[3]={0,0,0};
static char cMenu[3][37+1] = {
	"  ",
	"    0X",
	""};
int iChoice, iInput[2]={0,0};
UF_OBJ_disp_props_t Attribs;

	//  
	string.Format("Matrix %d", iNumber);
	while(UF_PART_new(string.GetBuffer(), 1, &UGPart))
		string.Format("Matrix %d", ++iNumber);

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Splines[i]));

		delete [] Points;
	}

	//       
	UF_CALL(UF_VIEW_fit_view(NULL_TAG, .9));

    Attribs.color = UF_OBJ_RED;
	Attribs.font = UF_OBJ_FONT_SOLID;
	Attribs.line_width = UF_OBJ_WIDTH_THICK;
	UF_CALL(UF_DISP_display_temporary_line(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, dPoint2, &Attribs));
	UF_CALL(UF_DISP_display_temporary_point(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, &Attribs, UF_DISP_DIAMOND));

	while(1)
	{
		iChoice = uc1603(" ", 3, cMenu, 3);

		if (iChoice==1 || iChoice==2) 
		{
			delete [] Splines;
			UF_CALL(UF_PART_close(UGPart, 0, 1));
			return false;
		}

		if(iChoice==5)			//"  "
		{
			UF_CALL(uc1616("  ", iInput, 1, dPoint1));
			UF_DISP_refresh();
			UF_CALL(UF_DISP_display_temporary_line(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, dPoint2, &Attribs));
			UF_CALL(UF_DISP_display_temporary_point(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, &Attribs, UF_DISP_DIAMOND));
		}
		else if(iChoice==6)		//"    0X"
		{
			UF_CALL(uc1616("  ", iInput, 1, dPoint2));
			UF_DISP_refresh();
			UF_CALL(UF_DISP_display_temporary_line(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, dPoint2, &Attribs));
			UF_CALL(UF_DISP_display_temporary_point(NULL_TAG, UF_DISP_USE_WORK_VIEW, dPoint1, &Attribs, UF_DISP_DIAMOND));
		}
		else if(iChoice==7)		//""
		{
			if(dPoint1[0]==dPoint2[0] && dPoint1[1]==dPoint2[1])
				uc1601("    ", 1);
			else
				break;
		}
	}

	//   
	dX[0] = dPoint2[0] - dPoint1[0];
	dX[1] = dPoint2[1] - dPoint1[1];
	dX[2] = 0;
	dY[0] = -dX[1];
	dY[1] = dX[0];
	dY[2] = 0;

	// 
	for(int i=0;i<iPressEdgesNum;i++)
	for(int j=0;j<iPressPointsNum[i];j++)
		TransformCoordinateSystem(&PointsPress[i][j], dX, dY, dZ, dPoint1);

	for(int i=0;i<iEdgesNum;i++)
	for(int j=0;j<Edges[i].iSectionsNum;j++)
	{
		TransformCoordinateSystem(&Edges[i].Sections[j].PointBase, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].PointRealBase, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].PointRounding, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].PointBevel, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].PointRealBevel, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].PointUndercutRounding, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].PointUndercut, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].CalcPoints.PointRounding, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].CalcPoints.PointBevel, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Edges[i].Sections[j].PlaneSectionNormal, dX, dY, dZ, dZeros);
	}

	for(int i=0;i<iHolesNum;i++)
	for(int j=0;j<Holes[i].iEdgesNum;j++)
	for(int k=0;k<Holes[i].Edges[j].iSectionsNum;k++)
		TransformCoordinateSystem(&Holes[i].Edges[j].Sections[k].PointRealBase, dX, dY, dZ, dPoint1);

	for(int i=0;i<iPunchesNum;i++)
	for(int j=0;j<Punches[i].iEdgesNum;j++)
	for(int k=0;k<Punches[i].Edges[j].iSectionsNum;k++)
	{
		TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointBase, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointBevel, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointRealBase, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&Punches[i].Edges[j].Sections[k].PointRounding, dX, dY, dZ, dPoint1);
	}

	for(int i=0;i<iPressInsNum;i++)
	for(int j=0;j<PressIns[i].iEdgesNum;j++)
	for(int k=0;k<PressIns[i].Edges[j].iSectionsNum;k++)
	{
		TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointBase, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointBevel, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointBottom, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointRealBase, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointRealBottom, dX, dY, dZ, dPoint1);
		TransformCoordinateSystem(&PressIns[i].Edges[j].Sections[k].PointRounding1, dX, dY, dZ, dPoint1);
	}

	delete [] Splines;

	UF_CALL(UF_PART_close(UGPart, 0, 1));

	return true;
}

void CPart::WriteParameters(CFile *File)
{
CString str;

	str.Format("#NumOfObjects = %d\r\n", iHolesNum+iPunchesNum+iPressInsNum);
	File->Write(str, str.GetLength());	

	for(int i=0;i<iHolesNum;i++)
		Holes[i].WriteParameters(File);

	for(int i=0;i<iPunchesNum;i++)
		Punches[i].WriteParameters(File);

	for(int i=0;i<iPressInsNum;i++)
		PressIns[i].WriteParameters(File);
}

void CPart::ProcessIgnoredEdges(CEdge *NewEdges)
{
tag_t *SplinesTop = new tag_t[iEdgesNum], Curves[2], Bridge;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity, iFirst, iLast, iMax, iReverse[2]={0,0}, iSize=0;
double dParms[2]={1,0}, dParLimits[2]={0,1}, dPoint[3];
UF_EVAL_p_t Eval;

	//    
	for(int i=0;i<iEdgesNum;i++)
	{
		memcpy(&NewEdges[i], &Edges[i], sizeof(CEdge));
		NewEdges[i].Sections = new CSection[Edges[i].iSectionsNum];
		for(int j=0;j<Edges[i].iSectionsNum;j++)
			memcpy(&NewEdges[i].Sections[j], &Edges[i].Sections[j], sizeof(CSection));
	}

	//   
	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bIgnore)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesTop[i]));
	}

	// 
	iMax = iEdgesNum;
	for(int i=0;i<iMax;i++)
	if(Edges[i].bIgnore)
	{
		iFirst = iLast = i;
		iSize = 1;
		//    
		if(i==0)
			for(int j=iEdgesNum-1;j>=0;j--)
				if(Edges[j].bIgnore)
				{
					iFirst = j;
					iSize++;
				}
				else
				{
					iMax = j+1;
					break;
				}
		for(int j=i+1;j<iMax;j++)
			if(Edges[j].bIgnore)
			{
				iSize++;
				iLast = j;
			}
			else
			{
				i = j;
				break;
			}

		//      
		iFirst--;
		iLast++;
		if(iFirst < 0)
			iFirst += iEdgesNum;
		if(iLast> iEdgesNum-1)
			iLast -= iEdgesNum;
		Curves[0] = SplinesTop[iFirst];
		Curves[1] = SplinesTop[iLast];
		UF_CALL(UF_CURVE_create_bridge_curve(UF_CURVE_TANGENT, Curves, dParms, iReverse, &Bridge));

		// 
		UF_CALL(UF_EVAL_initialize(Bridge, &Eval));
		UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

		for(int j=0;j<iSize;j++)
		{
			int iCur = iFirst + j + 1;
			if(iCur> iEdgesNum-1)
				iCur -= iEdgesNum;

			for(int k=0;k<Edges[iCur].iSectionsNum;k++)
			{
				tag_t T;
				double dPar = (dParLimits[1]-dParLimits[0])*(double)j/(double)iSize + (double)k*(dParLimits[1]-dParLimits[0])/(((double)Edges[iCur].iSectionsNum-1.0)*(double)iSize);
				UF_CALL(UF_EVAL_evaluate(Eval, 0, dPar, dPoint, NULL));
				NewEdges[iCur].Sections[k].PointRealBase = dPoint;
			}
		}

		UF_CALL(UF_EVAL_free(Eval));

		UF_CALL(UF_OBJ_delete_object(Bridge));
	}

	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bIgnore)
		UF_CALL(UF_OBJ_delete_object(SplinesTop[i]));

	delete [] SplinesTop;
}

//   
tag_t CPart::CreateBase(void)
{
tag_t *SplinesTop = new tag_t[iEdgesNum], *SplinesBottom = new tag_t[iEdgesNum], *OffCurve, Base;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity, iNumber;
double dBaseHeight = 0, dPointOrigin[3], dVector[3];
UF_CURVE_offset_data_t OffData;
UF_CURVE_offset_distance_data_t OffDist;
UF_STRING_t Curves;
CString string;

/*	for(int i=0;i<iEdgesNum;i++)
		if(Edges[i].bIgnore)
		{
			Edges[i].bPS = true;
			Edges[i].bNothing = false;
		}*/

	//   ,   ,  
	for(int i=0;i<iEdgesNum;i++)
		SplinesTop[i] = SplinesBottom[i] = NULL_TAG;

	//   
	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bIgnore && !Edges[i].bPS)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesTop[i]));

		// ,  
		if(Edges[i].bNothing && !Edges[i].bIgnore && Edges[i].dOffset!=0)
		{
			int t1=1, t2=1;

			Curves.num = 1;
			Curves.string = &t1;
			Curves.dir = &t2;
			Curves.id = &SplinesTop[i];

			string.Format("%f", Edges[i].dOffset);
			OffDist.distance = string.GetBuffer();

			OffData.approximation_tolerance = .1;
			OffData.string_tolerance = .1;
			OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
			OffData.offset_def.distance_type1 = &OffDist;
			OffData.input_curves = &Curves;

			UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

			UF_CALL(UF_ASSEM_remove_instance(SplinesTop[i]));
			SplinesTop[i] = *OffCurve;

			UF_free(OffCurve);
		}

		delete [] Points;
	}
	//  ,  
	for(int i=0;i<iEdgesNum;i++)
	{
		int iNext = i!=iEdgesNum-1?i+1:0;

		if(!Edges[i].bIgnore && !Edges[i].bPS && !Edges[iNext].bIgnore && !Edges[iNext].bPS)
			ConnectCurves(SplinesTop[i], SplinesTop[iNext]);
	}

	//       
	UF_CALL(UF_VIEW_fit_view(NULL_TAG, .9));

	//  
	dBaseHeight = CalcMinBaseHeight() + dG_BaseAdditionalThickness;

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		if(!Edges[i].bPS && !Edges[i].bNothing)
		{
			int iSectionsNum = Edges[i].iSectionsNum, iSectionFirst = 0, iSectionLast = Edges[i].iSectionsNum-1;

			// /     / 
			for(int j=0;j<Edges[i].iSectionsNum;j++)
			if(!Edges[i].Sections[j].bIsRoundingPresent || !Edges[i].Sections[j].bIsBevelPresent)
			{
				iSectionsNum--;
				iSectionFirst++;
			}
			else
				break;
			for(int j=Edges[i].iSectionsNum-1;j>=0;j--)
			if(!Edges[i].Sections[j].bIsRoundingPresent || !Edges[i].Sections[j].bIsBevelPresent)
			{
				iSectionsNum--;
				iSectionLast--;
			}
			else
				break;

			// /          
			for(int j=iSectionFirst;j<iSectionLast+1;j++)
			{
				double dPointA0[3], dPointA1[3], dPointA2[3], dPointB0[3], dPointB1[3], dPointB2[3];
				Edges[i].Sections[j].PointRealBase.Fill(dPointA0);
				Edges[i].Sections[j+1].PointRealBase.Fill(dPointA1);
				Edges[i].Sections[j+2].PointRealBase.Fill(dPointA2);

				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(dPointB0);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB0[0] - dPointOrigin[0];
				dVector[1] = dPointB0[1] - dPointOrigin[1];
				dVector[2] = dPointB0[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB0, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j+1].CalcPoints.PointBevel.Fill(dPointB1);
				Edges[i].Sections[j+1].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB1[0] - dPointOrigin[0];
				dVector[1] = dPointB1[1] - dPointOrigin[1];
				dVector[2] = dPointB1[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB1, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j+2].CalcPoints.PointBevel.Fill(dPointB2);
				Edges[i].Sections[j+2].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB2[0] - dPointOrigin[0];
				dVector[1] = dPointB2[1] - dPointOrigin[1];
				dVector[2] = dPointB2[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB2, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				if((CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))>2 ||
					(CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))<.5)
				{
					iSectionsNum--;
					iSectionFirst++;
				}
				else
					break;
			}
			for(int j=iSectionLast;j>=0;j--)
			{
				double dPointA0[3], dPointA1[3], dPointA2[3], dPointB0[3], dPointB1[3], dPointB2[3];
				Edges[i].Sections[j].PointRealBase.Fill(dPointA0);
				Edges[i].Sections[j-1].PointRealBase.Fill(dPointA1);
				Edges[i].Sections[j-2].PointRealBase.Fill(dPointA2);

				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(dPointB0);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB0[0] - dPointOrigin[0];
				dVector[1] = dPointB0[1] - dPointOrigin[1];
				dVector[2] = dPointB0[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB0, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j-1].CalcPoints.PointBevel.Fill(dPointB1);
				Edges[i].Sections[j-1].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB1[0] - dPointOrigin[0];
				dVector[1] = dPointB1[1] - dPointOrigin[1];
				dVector[2] = dPointB1[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB1, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j-2].CalcPoints.PointBevel.Fill(dPointB2);
				Edges[i].Sections[j-2].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB2[0] - dPointOrigin[0];
				dVector[1] = dPointB2[1] - dPointOrigin[1];
				dVector[2] = dPointB2[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB2, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				if((CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))>2 ||
					(CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))<.5)
				{
					iSectionsNum--;
					iSectionLast--;
				}
				else
					break;
			}

/*{
CString s;

s.Format("First:\t%d\r\nLast:\t%d\r\nNum:\t%d\r\nRealNum:\t%d", iSectionFirst, iSectionLast, iSectionsNum, Edges[i].iSectionsNum);
//s.Format("First:\t%f\r\nLast:\t%f", Edges[i].Sections[iSectionFirst].PointRealBase-Edges[i].Sections[iSectionFirst].PointRealBevel, Edges[i].Sections[iSectionLast].PointRealBase-Edges[i].Sections[iSectionLast].PointRealBevel);
MessageBox(NULL, s, "", MB_OK);
}*/
			iDegree = iSplinesDegree;

			//,        
			if(iSectionsNum<iDegree+1)
				iDegree = iSectionsNum-1;

			Points = new UF_CURVE_pt_slope_crvatr_s[iSectionsNum];

			for(int j=iSectionFirst;j<iSectionLast+1;j++)
			{
				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(Points[j-iSectionFirst].point);
//				Edges[i].Sections[j].PointRealBevel.Fill(Points[j-iSectionFirst].point);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = Points[j-iSectionFirst].point[0] - dPointOrigin[0];
				dVector[1] = Points[j-iSectionFirst].point[1] - dPointOrigin[1];
				dVector[2] = Points[j-iSectionFirst].point[2] - dPointOrigin[2];
//				OffsetPoint(dPointOrigin, Points[j-iSectionFirst].point, dVector, dBaseHeight / cos(Edges[i].Sections[j].CalcPoints.Alpha));
				OffsetPoint(dPointOrigin, Points[j-iSectionFirst].point, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));
				Points[j-iSectionFirst].slope_type = UF_CURVE_SLOPE_NONE;
				Points[j-iSectionFirst].crvatr_type = UF_CURVE_CRVATR_NONE;
/*				tag_t Temp;
				UF_CALL(UF_CURVE_create_point(Points[j-iSectionFirst].point, &Temp));
				if(j==iSectionFirst || j==iSectionLast)
					UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_GREEN));
				else
					UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_RED));
				Edges[i].Sections[j].PointBase.Fill(dVector);
				UF_CALL(UF_CURVE_create_point(dVector, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
				Edges[i].Sections[j].PointRealBase.Fill(dVector);
				UF_CALL(UF_CURVE_create_point(dVector, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
				Edges[i].Sections[j].PointRealBevel.Fill(dVector);
				UF_CALL(UF_CURVE_create_point(dVector, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));*/
			}
/*			tag_t Temp;
			Edges[i].Sections[0].PointBevel.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[0].PointRealBase.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[0].PointRounding.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[Edges[i].iSectionsNum-1].PointBevel.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[Edges[i].iSectionsNum-1].PointRounding.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));*/


			if(iEdgesNum==1)
				iPeriodicity = 1;
			else
				iPeriodicity = 0;

			//    
			UF_CALL(UF_CURVE_create_spline_thru_pts(
				iDegree,
				iPeriodicity,
				iSectionsNum,
				Points,
				NULL,
				0,
				&SplinesBottom[i]));

			delete [] Points;

			// ,       
			{
			}
		}
		else if(Edges[i].bNothing && !Edges[i].bIgnore)
		{
			iDegree = iSplinesDegree;

			//,        
			if(Edges[i].iSectionsNum<iDegree+1)
				iDegree = Edges[i].iSectionsNum-1;

			Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = -Service.dBaseNormal[0];
				dVector[1] = -Service.dBaseNormal[1];
				dVector[2] = -Service.dBaseNormal[2];
				OffsetPoint(dPointOrigin, Points[j].point, dVector, dBaseHeight);
				Points[j].slope_type = UF_CURVE_SLOPE_NONE;
				Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
			}

			if(iEdgesNum==1)
				iPeriodicity = 1;
			else
				iPeriodicity = 0;

			//    
			UF_CALL(UF_CURVE_create_spline_thru_pts(
				iDegree,
				iPeriodicity,
				Edges[i].iSectionsNum,
				Points,
				NULL,
				0,
				&SplinesBottom[i]));

			// ,  
			if(Edges[i].bNothing && Edges[i].dOffset!=0)
			{
				int t1=1, t2=1;

				Curves.num = 1;
				Curves.string = &t1;
				Curves.dir = &t2;
				Curves.id = &SplinesBottom[i];

				string.Format("%f", Edges[i].dOffset);
				OffDist.distance = string.GetBuffer();

				OffData.approximation_tolerance = .1;
				OffData.string_tolerance = .1;
				OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
				OffData.offset_def.distance_type1 = &OffDist;
				OffData.input_curves = &Curves;

				UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

				UF_CALL(UF_ASSEM_remove_instance(SplinesBottom[i]));
				SplinesBottom[i] = *OffCurve;

				UF_free(OffCurve);
			}

			delete [] Points;
		}
/*		else	// 
		{
			UF_CURVE_pt_slope_crvatr_s Points;
			tag_t Temp;

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(Points.point);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = Points.point[0] - dPointOrigin[0];
				dVector[1] = Points.point[1] - dPointOrigin[1];
				dVector[2] = Points.point[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, Points.point, dVector, dBaseHeight / cos(Edges[i].Sections[j].CalcPoints.Alpha));
				UF_CALL(UF_CURVE_create_point(Points.point, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_RED));
			}
		}*/
	}
	//  ,  
	for(int i=0;i<iEdgesNum;i++)
	{
		int iNext = i!=iEdgesNum-1?i+1:0;

		if(!Edges[i].bIgnore && !Edges[i].bPS && !Edges[iNext].bIgnore && !Edges[iNext].bPS)
			ConnectCurves(SplinesBottom[i], SplinesBottom[iNext]);
	}

	//  
	{
		tag_t Plane, *SewTools, Curves[2];
		UF_STRING_t Section, Spine;
		int iAlignment=1, iEndPoint=0, iBodyType=0, iNumOfRuleds=0, iNumOfPS, iReverse[2]={0,0};
		double dValue[6], dTol[3]={.01, .01, 1}, dParms[2]={1,0};
		uf_list_p_t ListDisjoint;

		//    
		for(int i=0;i<iEdgesNum-1;i++)
			if(!Edges[i].bPS || !Edges[i+1].bPS)
				iNumOfRuleds++;
		if(!Edges[iEdgesNum-1].bPS || !Edges[0].bPS)
			iNumOfRuleds++;

		//   
		SewTools = new tag_t[iNumOfRuleds];
		iNumOfRuleds = 0;

		UF_MODL_create_string_list(2, 2, &Section);
		Section.num = 2;
		Section.string[0] = 1;
		Section.dir[0] = UF_MODL_CURVE_START_FROM_BEGIN;
		Section.string[1] = 1;
		Section.dir[1] = UF_MODL_CURVE_START_FROM_BEGIN;
		Spine.num = 0;

		//   
		for(int i=0;i<iEdgesNum;i++)
			if(!Edges[i].bPS && !Edges[i].bIgnore)
			{
				Section.id[0] = SplinesTop[i];
				Section.id[1] = SplinesBottom[i];

				UF_CALL(UF_MODL_create_ruled(&Section, &Spine, &iAlignment, dValue, &iEndPoint, &iBodyType, UF_NULLSIGN, dTol, &SewTools[iNumOfRuleds++]));
			}

		UF_MODL_free_string_list(&Section);

		//   
		for(int i=0;i<iEdgesNum;i++)
			if(!Edges[i].bPS && !Edges[i].bIgnore)
			{
				for(int j=i;j<iEdgesNum;j++)
					if(Edges[j].bPS || Edges[j].bIgnore)
					{
						iNumOfPS = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(Edges[k].bPS || Edges[k].bIgnore)
								iNumOfPS++;
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(Edges[k].bPS || Edges[k].bIgnore)
									iNumOfPS++;

						UF_MODL_create_string_list(iNumOfPS+1, iNumOfPS+1, &Section);
//						UF_MODL_create_string_list(2, 2, &Section);
						Section.num = 2;
						Section.string[0] = 1;
						Section.dir[0] = UF_MODL_CURVE_START_FROM_BEGIN;
						Section.string[1] = 1;
						Section.dir[1] = UF_MODL_CURVE_START_FROM_BEGIN;

						iNumOfPS = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(Edges[k].bPS || Edges[k].bIgnore)
								Section.id[iNumOfPS++] = SplinesTop[k];
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(Edges[k].bPS || Edges[k].bIgnore)
									Section.id[iNumOfPS++] = SplinesTop[k];

						//    
						Curves[0] = SplinesTop[j-1];
						if(j+iNumOfPS > iEdgesNum-1)
							Curves[1] = SplinesTop[j+iNumOfPS-iEdgesNum];
						else
							Curves[1] = SplinesTop[j+iNumOfPS];
						UF_CALL(UF_CURVE_create_bridge_curve(UF_CURVE_TANGENT, Curves, dParms, iReverse, &Section.id[0]));
						SplinesTop[j] = Section.id[0];

						//    
						Curves[0] = SplinesBottom[j-1];
						if(j+iNumOfPS > iEdgesNum-1)
							Curves[1] = SplinesBottom[j+iNumOfPS-iEdgesNum];
						else
							Curves[1] = SplinesBottom[j+iNumOfPS];
						UF_CALL(UF_CURVE_create_bridge_curve(UF_CURVE_TANGENT, Curves, dParms, iReverse, &Section.id[1]));
						SplinesBottom[j] = Section.id[1];

						UF_CALL(UF_MODL_create_ruled(&Section, &Spine, &iAlignment, dValue, &iEndPoint, &iBodyType, UF_NULLSIGN, dTol, &SewTools[iNumOfRuleds++]));

						UF_MODL_free_string_list(&Section);

						j += iNumOfPS;
					}
				break;
			}

		//  
		UF_MODL_create_string_list(iNumOfRuleds, iNumOfRuleds, &Section);
		Section.num = 1;
		Section.string[0] = iNumOfRuleds;
		iNumOfRuleds=0;
		for(int i=0;i<iEdgesNum;i++)
			if(SplinesTop[i] != NULL_TAG)
			{
				Section.dir[iNumOfRuleds] = UF_MODL_CURVE_START_FROM_BEGIN;
				Section.id[iNumOfRuleds++] = SplinesTop[i];
			}

		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &SewTools[iNumOfRuleds]));
		UF_MODL_free_string_list(&Section);

		//  
		UF_MODL_create_string_list(iNumOfRuleds, iNumOfRuleds, &Section);
		Section.num = 1;
		Section.string[0] = iNumOfRuleds;
		iNumOfRuleds=0;
		for(int i=0;i<iEdgesNum;i++)
			if(SplinesTop[i] != NULL_TAG)
			{
				Section.dir[iNumOfRuleds] = UF_MODL_CURVE_START_FROM_BEGIN;
				Section.id[iNumOfRuleds++] = SplinesBottom[i];
			}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane));
		UF_MODL_free_string_list(&Section);

		UF_CALL(UF_MODL_create_sew(0, 1, &Plane, iNumOfRuleds+1, SewTools, .01, 1, &ListDisjoint, &Base));

		//delete [] SewTools;
	}

    //   
	{
		int iNumBlendBlocks=1;
		int *iBlendBlocks;	// : , ,  

		//   
		for(int i=0;i<iEdgesNum;i++)
			if(Edges[i].bNothing && !Edges[i].bIgnore && !Edges[i].bPS)
				iNumBlendBlocks = 0;
		if(iNumBlendBlocks!=0)
		{
			iBlendBlocks = new int[3];
			iBlendBlocks[2] = iEdgesNum;
			for(int i=0;i<iEdgesNum;i++)
				if(!Edges[i].bNothing && !Edges[i].bPS && !Edges[i].bIgnore)
				{
					iBlendBlocks[0] = i;
					iBlendBlocks[1] = i-1;
					if(iBlendBlocks[1]<0)
						iBlendBlocks[1] += iEdgesNum;
					break;
				}
		}
		if(iNumBlendBlocks==0)
		{
			int iCur=0, iCurBlock=0;
			while(1)
			{
				if(iCur < iEdgesNum-1)
				{
					while(iCur<iEdgesNum && (Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
						iCur++;
					iNumBlendBlocks++;
				}
				else
					break;
				if(iCur < iEdgesNum-1)
				{
					while(iCur<iEdgesNum && !(Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
						iCur++;
					iNumBlendBlocks++;
				}
				else
					break;
			}
			iNumBlendBlocks = (iNumBlendBlocks)/2;
			iBlendBlocks = new int[(iNumBlendBlocks+1)*3];
			//   
			iCur=0;
			while(1)
			{
				if(iCur < iEdgesNum-1)
				{
					bool bF = false;
					iBlendBlocks[iCurBlock*3] = iCur;
					iBlendBlocks[iCurBlock*3+1] = iCur-1;
					while(iCur<iEdgesNum && (Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
					{
						iCur++;
						iBlendBlocks[iCurBlock*3+1]++;
						bF = true;
					}
					if(bF == true)
						iCurBlock++;

				}
				else
					break;
				if(iCur < iEdgesNum-1)
				{
					while(iCur<iEdgesNum && !(Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
						iCur++;
				}
				else
					break;
			}
			if(iCurBlock == iNumBlendBlocks + 1)
				iBlendBlocks[0] = iBlendBlocks[(iCurBlock-1)*3];
		}
		for(int i=0;i<iNumBlendBlocks;i++)
		{
			iBlendBlocks[i*3+2] = 1;
			if((iBlendBlocks[i*3+1]-iBlendBlocks[i*3]>0?iBlendBlocks[i*3+1]-iBlendBlocks[i*3]:iBlendBlocks[i*3+1]-iBlendBlocks[i*3]+iEdgesNum) > 0)
			{
				int iCur = iBlendBlocks[i*3];
				while(iCur != iBlendBlocks[i*3+1])
				{
					int iPrev = iCur;
					iCur++;
					if(iCur >= iEdgesNum)
						iCur -= iEdgesNum;
//					if(!Edges[iPrev].bNothing || !Edges[iCur].bNothing)
					if(SplinesTop[iCur] != NULL_TAG)
						iBlendBlocks[i*3+2]++;
				}
			}
		}

		int iBaseEdges;
		uf_list_p_t BaseEdges;

		for(int i=0;i<iNumBlendBlocks;i++)
		{
			//  
			tag_t Blend;
			UF_MODL_edge_blend_data_s BlendData;
			BlendData.blend_type = 0;
			BlendData.blend_instanced = FALSE;
			BlendData.blend_setback = FALSE;
			BlendData.vrb_tolerance = .1;
			BlendData.blend_radius[0] = '5';
			BlendData.blend_radius[1] = '\0';
			BlendData.smooth_overflow = TRUE;
			BlendData.cliff_overflow = TRUE;
			BlendData.notch_overflow = TRUE;

			//  
			UF_CALL(UF_MODL_ask_feat_edges(Base, &BaseEdges));
			UF_CALL(UF_MODL_ask_list_count(BaseEdges, &iBaseEdges));

			BlendData.number_edges = iBlendBlocks[i*3+2];
			BlendData.edge_data = new UF_MODL_blend_edge_s[iBlendBlocks[i*3+2]];
			int iCur = iBlendBlocks[i*3];
			int iCur2 = 0;
			while(iCur2 < iBlendBlocks[i*3+2])
			{
//				int iPrev = iCur-1;
//				if(iPrev<0)	iPrev += iEdgesNum;
				if(SplinesTop[iCur] != NULL_TAG)
				{
					BlendData.edge_data[iCur2].edge = FindClosestEdge(SplinesTop[iCur], &BaseEdges);

					//   
					BlendData.edge_data[iCur2].number_points = 0;

					if(!Edges[iCur].bPS && !Edges[iCur].bNothing)
					{
						double dGuess[3], dMinDist, dParm, dParLimits[2];
						UF_EVAL_p_t Eval;

						UF_CALL(UF_EVAL_initialize(BlendData.edge_data[iCur2].edge, &Eval));
						UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

						//   
						for(int k=0;k<Edges[iCur].iSectionsNum;k++)
						{
							Edges[iCur].Sections[k].PointRealBase.Fill(dGuess);
							UF_CALL(UF_MODL_ask_minimum_dist(BlendData.edge_data[iCur2].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
							if(dMinDist < dG_DistTol && Edges[iCur].Sections[k].CalcPoints.r > .1 && Edges[iCur].Sections[k].CalcPoints.r < 100)
								BlendData.edge_data[iCur2].number_points++;
						}

						//  
						BlendData.edge_data[iCur2].point_data = new UF_MODL_blend_point_s[BlendData.edge_data[iCur2].number_points];
						BlendData.edge_data[iCur2].number_points = 0;

						//  
						for(int k=0;k<Edges[iCur].iSectionsNum;k++)
						{
							Edges[iCur].Sections[k].PointRealBase.Fill(dGuess);
							UF_CALL(UF_MODL_ask_minimum_dist(BlendData.edge_data[iCur2].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
							if(dMinDist < dG_DistTol && Edges[iCur].Sections[k].CalcPoints.r > .1 && Edges[iCur].Sections[k].CalcPoints.r < 100)
							{
								Edges[iCur].Sections[k].PointRealBase.Fill(dGuess);
								UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
								BlendData.edge_data[iCur2].point_data[BlendData.edge_data[iCur2].number_points].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
								string.Format("%f", Edges[iCur].Sections[k].CalcPoints.r);
//										string.Format("%f", 70.0);
								string.CopyChars(BlendData.edge_data[iCur2].point_data[BlendData.edge_data[iCur2].number_points].radius, string.GetBuffer(), string.GetLength());
								BlendData.edge_data[iCur2].point_data[BlendData.edge_data[iCur2].number_points].radius[string.GetLength()] = '\0';
								BlendData.edge_data[iCur2].number_points++;
							}
						}

						UF_CALL(UF_EVAL_free(Eval));
					}
					iCur2++;

				}

//				iCur2++;
				iCur++;
				if(iCur>=iEdgesNum)	iCur -= iEdgesNum;
			}

			UF_MODL_create_edge_blend(&BlendData, &Blend);

			for(int j=0;j<BlendData.number_edges;j++)
				if(BlendData.edge_data[j].number_points)
					delete [] BlendData.edge_data[j].point_data;
			delete [] BlendData.edge_data;
		}
/*	for(int i=0;i<iEdgesNum;i++)
		if(Edges[i].bIgnore && Edges[i].bPS)
		{
			Edges[i].bPS = false;
			Edges[i].bNothing = true;
		}*/

	//   
	{
		tag_t Temp, Edge, *ReservedEdges = new tag_t[iEdgesNum*2];
		uf_list_p_t BodyEdges;
		int iBodyEdges;

		UF_CALL(UF_MODL_ask_feat_body(Base, &Temp));
		UF_CALL(UF_MODL_ask_body_edges(Temp, &BodyEdges));
		UF_CALL(UF_MODL_ask_list_count(BodyEdges, &iBodyEdges));
		for(int i=0;i<iEdgesNum;i++)
		{
			if(!Edges[i].bNothing || (Edges[i].bNothing && Edges[i].bIgnore))
				ReservedEdges[i] = FindClosestEdge(SplinesTop[i], &BodyEdges);
			else
				ReservedEdges[i] = NULL_TAG;
		}
		for(int i=0;i<iEdgesNum;i++)
			ReservedEdges[i+iEdgesNum] = FindClosestEdge(SplinesBottom[i], &BodyEdges);

		Service.dBaseRoundingRadius = RoundAllEdges(Temp, ReservedEdges, iEdgesNum*2, dG_RoundingsRadius);
		delete [] ReservedEdges;
	}
	}

	//  
	for(int i=0;i<iEdgesNum;i++)
	{
		if(SplinesTop[i] != NULL_TAG)
			UF_CALL(UF_OBJ_set_blank_status(SplinesTop[i], UF_OBJ_BLANKED));
		if(SplinesBottom[i] != NULL_TAG)
			UF_CALL(UF_OBJ_set_blank_status(SplinesBottom[i], UF_OBJ_BLANKED));
	}

	//   - ""
	{
		tag_t Temp;
		UF_CALL(UF_MODL_ask_feat_body(Base, &Temp));
		Base = Temp;
	}

	return Base;
}
tag_t CPart::CreateBase1(void)
{
tag_t *SplinesTop = new tag_t[iEdgesNum], *SplinesBottom = new tag_t[iEdgesNum], *OffCurve, Base;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity, iNumber;
double dBaseHeight = 0, dPointOrigin[3], dVector[3];
UF_CURVE_offset_data_t OffData;
UF_CURVE_offset_distance_data_t OffDist;
UF_STRING_t Curves;
CString string;

CEdge *Edges = new CEdge[iEdgesNum];

	ProcessIgnoredEdges(Edges);

//	return NULL_TAG;

	for(int i=0;i<iEdgesNum;i++)
		if(Edges[i].bIgnore)
		{
			Edges[i].bPS = true;
			Edges[i].bNothing = false;
		}

	//   ,   ,  
	for(int i=0;i<iEdgesNum;i++)
		SplinesTop[i] = SplinesBottom[i] = NULL_TAG;

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesTop[i]));

		// ,  
		if(Edges[i].bNothing && Edges[i].dOffset!=0)
		{
			int t1=1, t2=1;

			Curves.num = 1;
			Curves.string = &t1;
			Curves.dir = &t2;
			Curves.id = &SplinesTop[i];

			string.Format("%f", Edges[i].dOffset);
			OffDist.distance = string.GetBuffer();

			OffData.approximation_tolerance = .1;
			OffData.string_tolerance = .1;
			OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
			OffData.offset_def.distance_type1 = &OffDist;
			OffData.input_curves = &Curves;

			UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

			UF_CALL(UF_ASSEM_remove_instance(SplinesTop[i]));
			SplinesTop[i] = *OffCurve;

			UF_free(OffCurve);
		}

		delete [] Points;
	}
	//  ,  
	for(int i=0;i<iEdgesNum-1;i++)
		if(Edges[i].bNothing || Edges[i+1].bNothing)
			ConnectCurves(SplinesTop[i], SplinesTop[i+1]);
	if(Edges[iEdgesNum-1].bNothing || Edges[0].bNothing)
		ConnectCurves(SplinesTop[iEdgesNum-1], SplinesTop[0]);

	for(int i=0;i<iEdgesNum-1;i++)
		if(!Edges[i].bNothing && !Edges[i+1].bNothing)
			ConnectCurves(SplinesTop[i], SplinesTop[i+1]);
	if(!Edges[iEdgesNum-1].bNothing && !Edges[0].bNothing)
		ConnectCurves(SplinesTop[iEdgesNum-1], SplinesTop[0]);

//	for(int i=0;i<iEdgesNum-1;i++)
//		if((!Edges[i].bNothing && Edges[i+1].bNothing && !Edges[i+1].bIgnore) || (!Edges[i+1].bNothing && Edges[i].bNothing && !Edges[i].bIgnore))
//			ConnectCurves(SplinesTop[i], SplinesTop[i+1]);
//	if((!Edges[iEdgesNum-1].bNothing && Edges[0].bNothing && !Edges[0].bIgnore) || (!Edges[0].bNothing && Edges[iEdgesNum-1].bNothing && !Edges[iEdgesNum-1].bIgnore))
//		ConnectCurves(SplinesTop[iEdgesNum-1], SplinesTop[0]);

	//       
	UF_CALL(UF_VIEW_fit_view(NULL_TAG, .9));

	//  
	dBaseHeight = CalcMinBaseHeight() + dG_BaseAdditionalThickness;

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		if(!Edges[i].bPS && !Edges[i].bNothing)
		{
			int iSectionsNum = Edges[i].iSectionsNum, iSectionFirst = 0, iSectionLast = Edges[i].iSectionsNum-1;

			// /     / 
			for(int j=0;j<Edges[i].iSectionsNum;j++)
			if(!Edges[i].Sections[j].bIsRoundingPresent || !Edges[i].Sections[j].bIsBevelPresent)
			{
				iSectionsNum--;
				iSectionFirst++;
			}
			else
				break;
			for(int j=Edges[i].iSectionsNum-1;j>=0;j--)
			if(!Edges[i].Sections[j].bIsRoundingPresent || !Edges[i].Sections[j].bIsBevelPresent)
			{
				iSectionsNum--;
				iSectionLast--;
			}
			else
				break;

			// /          
			for(int j=iSectionFirst;j<iSectionLast+1;j++)
			{
				double dPointA0[3], dPointA1[3], dPointA2[3], dPointB0[3], dPointB1[3], dPointB2[3];
				Edges[i].Sections[j].PointRealBase.Fill(dPointA0);
				Edges[i].Sections[j+1].PointRealBase.Fill(dPointA1);
				Edges[i].Sections[j+2].PointRealBase.Fill(dPointA2);

				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(dPointB0);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB0[0] - dPointOrigin[0];
				dVector[1] = dPointB0[1] - dPointOrigin[1];
				dVector[2] = dPointB0[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB0, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j+1].CalcPoints.PointBevel.Fill(dPointB1);
				Edges[i].Sections[j+1].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB1[0] - dPointOrigin[0];
				dVector[1] = dPointB1[1] - dPointOrigin[1];
				dVector[2] = dPointB1[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB1, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j+2].CalcPoints.PointBevel.Fill(dPointB2);
				Edges[i].Sections[j+2].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB2[0] - dPointOrigin[0];
				dVector[1] = dPointB2[1] - dPointOrigin[1];
				dVector[2] = dPointB2[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB2, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				if((CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))>2 ||
					(CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))<.5)
				{
					iSectionsNum--;
					iSectionFirst++;
				}
				else
					break;
			}
			for(int j=iSectionLast;j>=0;j--)
			{
				double dPointA0[3], dPointA1[3], dPointA2[3], dPointB0[3], dPointB1[3], dPointB2[3];
				Edges[i].Sections[j].PointRealBase.Fill(dPointA0);
				Edges[i].Sections[j-1].PointRealBase.Fill(dPointA1);
				Edges[i].Sections[j-2].PointRealBase.Fill(dPointA2);

				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(dPointB0);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB0[0] - dPointOrigin[0];
				dVector[1] = dPointB0[1] - dPointOrigin[1];
				dVector[2] = dPointB0[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB0, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j-1].CalcPoints.PointBevel.Fill(dPointB1);
				Edges[i].Sections[j-1].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB1[0] - dPointOrigin[0];
				dVector[1] = dPointB1[1] - dPointOrigin[1];
				dVector[2] = dPointB1[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB1, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				Edges[i].Sections[j-2].CalcPoints.PointBevel.Fill(dPointB2);
				Edges[i].Sections[j-2].PointRealBase.Fill(dPointOrigin);
				dVector[0] = dPointB2[0] - dPointOrigin[0];
				dVector[1] = dPointB2[1] - dPointOrigin[1];
				dVector[2] = dPointB2[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, dPointB2, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));

				if((CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))>2 ||
					(CalcDist(dPointA0,dPointA1)*CalcDist(dPointB1,dPointB2))/(CalcDist(dPointA1,dPointA2)*CalcDist(dPointB0,dPointB1))<.5)
				{
					iSectionsNum--;
					iSectionLast--;
				}
				else
					break;
			}

/*{
CString s;

s.Format("First:\t%d\r\nLast:\t%d\r\nNum:\t%d\r\nRealNum:\t%d", iSectionFirst, iSectionLast, iSectionsNum, Edges[i].iSectionsNum);
//s.Format("First:\t%f\r\nLast:\t%f", Edges[i].Sections[iSectionFirst].PointRealBase-Edges[i].Sections[iSectionFirst].PointRealBevel, Edges[i].Sections[iSectionLast].PointRealBase-Edges[i].Sections[iSectionLast].PointRealBevel);
MessageBox(NULL, s, "", MB_OK);
}*/
			iDegree = iSplinesDegree;

			//,        
			if(iSectionsNum<iDegree+1)
				iDegree = iSectionsNum-1;

			Points = new UF_CURVE_pt_slope_crvatr_s[iSectionsNum];

			for(int j=iSectionFirst;j<iSectionLast+1;j++)
			{
				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(Points[j-iSectionFirst].point);
//				Edges[i].Sections[j].PointRealBevel.Fill(Points[j-iSectionFirst].point);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = Points[j-iSectionFirst].point[0] - dPointOrigin[0];
				dVector[1] = Points[j-iSectionFirst].point[1] - dPointOrigin[1];
				dVector[2] = Points[j-iSectionFirst].point[2] - dPointOrigin[2];
//				OffsetPoint(dPointOrigin, Points[j-iSectionFirst].point, dVector, dBaseHeight / cos(Edges[i].Sections[j].CalcPoints.Alpha));
				OffsetPoint(dPointOrigin, Points[j-iSectionFirst].point, dVector, dBaseHeight / fabs(cos(CalcAngle(dVector, Service.dBaseNormal))));
				Points[j-iSectionFirst].slope_type = UF_CURVE_SLOPE_NONE;
				Points[j-iSectionFirst].crvatr_type = UF_CURVE_CRVATR_NONE;
/*				tag_t Temp;
				UF_CALL(UF_CURVE_create_point(Points[j-iSectionFirst].point, &Temp));
				if(j==iSectionFirst || j==iSectionLast)
					UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_GREEN));
				else
					UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_RED));
				Edges[i].Sections[j].PointBase.Fill(dVector);
				UF_CALL(UF_CURVE_create_point(dVector, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
				Edges[i].Sections[j].PointRealBase.Fill(dVector);
				UF_CALL(UF_CURVE_create_point(dVector, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
				Edges[i].Sections[j].PointRealBevel.Fill(dVector);
				UF_CALL(UF_CURVE_create_point(dVector, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));*/
			}
/*			tag_t Temp;
			Edges[i].Sections[0].PointBevel.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[0].PointRealBase.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[0].PointRounding.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[Edges[i].iSectionsNum-1].PointBevel.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));
			Edges[i].Sections[Edges[i].iSectionsNum-1].PointRounding.Fill(dVector);
			UF_CALL(UF_CURVE_create_point(dVector, &Temp));
			UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_BLUE));*/


			if(iEdgesNum==1)
				iPeriodicity = 1;
			else
				iPeriodicity = 0;

			//    
			UF_CALL(UF_CURVE_create_spline_thru_pts(
				iDegree,
				iPeriodicity,
				iSectionsNum,
				Points,
				NULL,
				0,
				&SplinesBottom[i]));

			delete [] Points;

			// ,       
			{
			}
		}
		else if(Edges[i].bNothing)
		{
			iDegree = iSplinesDegree;

			//,        
			if(Edges[i].iSectionsNum<iDegree+1)
				iDegree = Edges[i].iSectionsNum-1;

			Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = -Service.dBaseNormal[0];
				dVector[1] = -Service.dBaseNormal[1];
				dVector[2] = -Service.dBaseNormal[2];
				OffsetPoint(dPointOrigin, Points[j].point, dVector, dBaseHeight);
				Points[j].slope_type = UF_CURVE_SLOPE_NONE;
				Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
			}

			if(iEdgesNum==1)
				iPeriodicity = 1;
			else
				iPeriodicity = 0;

			//    
			UF_CALL(UF_CURVE_create_spline_thru_pts(
				iDegree,
				iPeriodicity,
				Edges[i].iSectionsNum,
				Points,
				NULL,
				0,
				&SplinesBottom[i]));

			// ,  
			if(Edges[i].bNothing && Edges[i].dOffset!=0)
			{
				int t1=1, t2=1;

				Curves.num = 1;
				Curves.string = &t1;
				Curves.dir = &t2;
				Curves.id = &SplinesBottom[i];

				string.Format("%f", Edges[i].dOffset);
				OffDist.distance = string.GetBuffer();

				OffData.approximation_tolerance = .1;
				OffData.string_tolerance = .1;
				OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
				OffData.offset_def.distance_type1 = &OffDist;
				OffData.input_curves = &Curves;

				UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

				UF_CALL(UF_ASSEM_remove_instance(SplinesBottom[i]));
				SplinesBottom[i] = *OffCurve;

				UF_free(OffCurve);
			}

			delete [] Points;
		}
/*		else	// 
		{
			UF_CURVE_pt_slope_crvatr_s Points;
			tag_t Temp;

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].CalcPoints.PointBevel.Fill(Points.point);
				Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
				dVector[0] = Points.point[0] - dPointOrigin[0];
				dVector[1] = Points.point[1] - dPointOrigin[1];
				dVector[2] = Points.point[2] - dPointOrigin[2];
				OffsetPoint(dPointOrigin, Points.point, dVector, dBaseHeight / cos(Edges[i].Sections[j].CalcPoints.Alpha));
				UF_CALL(UF_CURVE_create_point(Points.point, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_RED));
			}
		}*/
	}

	//  ,  
	for(int i=0;i<iEdgesNum;i++)
	{
		int iNext = i!=iEdgesNum-1?i+1:0;

		if(!Edges[i].bIgnore && !Edges[i].bPS && !Edges[iNext].bIgnore && !Edges[iNext].bPS)
			ConnectCurves(SplinesBottom[i], SplinesBottom[iNext]);
	}

	//  
	{
		tag_t Plane, *SewTools, Curves[2];
		UF_STRING_t Section, Spine;
		int iAlignment=1, iEndPoint=0, iBodyType=0, iNumOfRuleds=0, iNumOfPS, iReverse[2]={0,0};
		double dValue[6], dTol[3]={.01, .01, 1}, dParms[2]={1,0};
		uf_list_p_t ListDisjoint;

		//    
		for(int i=0;i<iEdgesNum-1;i++)
			if(!Edges[i].bPS || !Edges[i+1].bPS)
				iNumOfRuleds++;
		if(!Edges[iEdgesNum-1].bPS || !Edges[0].bPS)
			iNumOfRuleds++;

		//   
		SewTools = new tag_t[iNumOfRuleds];
		iNumOfRuleds = 0;

		UF_MODL_create_string_list(2, 2, &Section);
		Section.num = 2;
		Section.string[0] = 1;
		Section.dir[0] = UF_MODL_CURVE_START_FROM_BEGIN;
		Section.string[1] = 1;
		Section.dir[1] = UF_MODL_CURVE_START_FROM_BEGIN;
		Spine.num = 0;

		//   
		for(int i=0;i<iEdgesNum;i++)
			if(!Edges[i].bPS)
			{
				Section.id[0] = SplinesTop[i];
				Section.id[1] = SplinesBottom[i];

				UF_CALL(UF_MODL_create_ruled(&Section, &Spine, &iAlignment, dValue, &iEndPoint, &iBodyType, UF_NULLSIGN, dTol, &SewTools[iNumOfRuleds++]));
			}

		UF_MODL_free_string_list(&Section);

		//   
		for(int i=0;i<iEdgesNum;i++)
			if(!Edges[i].bPS)
			{
				for(int j=i;j<iEdgesNum;j++)
					if(Edges[j].bPS)
					{
						iNumOfPS = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(Edges[k].bPS)
								iNumOfPS++;
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(Edges[k].bPS)
									iNumOfPS++;

						UF_MODL_create_string_list(iNumOfPS+1, iNumOfPS+1, &Section);
						Section.num = 2;
						Section.string[0] = iNumOfPS;
						Section.dir[0] = UF_MODL_CURVE_START_FROM_BEGIN;
						Section.string[1] = 1;
						Section.dir[1] = UF_MODL_CURVE_START_FROM_BEGIN;

						iNumOfPS = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(Edges[k].bPS)
								Section.id[iNumOfPS++] = SplinesTop[k];
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(Edges[k].bPS)
									Section.id[iNumOfPS++] = SplinesTop[k];

						//    
						Curves[0] = SplinesBottom[j-1];
						if(j+iNumOfPS > iEdgesNum-1)
							Curves[1] = SplinesBottom[j+iNumOfPS-iEdgesNum];
						else
							Curves[1] = SplinesBottom[j+iNumOfPS];
						UF_CALL(UF_CURVE_create_bridge_curve(UF_CURVE_TANGENT, Curves, dParms, iReverse, &Section.id[iNumOfPS]));
						SplinesBottom[j] = Section.id[iNumOfPS];

						UF_CALL(UF_MODL_create_ruled(&Section, &Spine, &iAlignment, dValue, &iEndPoint, &iBodyType, UF_NULLSIGN, dTol, &SewTools[iNumOfRuleds++]));

						UF_MODL_free_string_list(&Section);

						j += iNumOfPS;
					}
				break;
			}
		//  
		UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
		Section.num = 1;
		Section.string[0] = iEdgesNum;
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
			Section.id[i] = SplinesTop[i];
		}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &SewTools[iNumOfRuleds]));
		UF_MODL_free_string_list(&Section);

		//  
		UF_MODL_create_string_list(iNumOfRuleds, iNumOfRuleds, &Section);
		Section.num = 1;
		Section.string[0] = iNumOfRuleds;
		iNumOfRuleds=0;
		for(int i=0;i<iEdgesNum;i++)
			if(!Edges[i].bPS)
			{
				Section.dir[iNumOfRuleds] = UF_MODL_CURVE_START_FROM_BEGIN;
				Section.id[iNumOfRuleds++] = SplinesBottom[i];
			}
			else if(i>0 && Edges[i].bPS && !Edges[i-1].bPS)
			{
				Section.dir[iNumOfRuleds] = UF_MODL_CURVE_START_FROM_BEGIN;
				Section.id[iNumOfRuleds++] = SplinesBottom[i];
			}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane));
		UF_MODL_free_string_list(&Section);

		UF_CALL(UF_MODL_create_sew(0, 1, &Plane, iNumOfRuleds+1, SewTools, .01, 1, &ListDisjoint, &Base));

		//delete [] SewTools;
	}

    //   
	{
		int iNumBlendBlocks=1;
		int *iBlendBlocks;	// : , ,  

		//   
		for(int i=0;i<iEdgesNum;i++)
			if(Edges[i].bNothing && !Edges[i].bIgnore)
				iNumBlendBlocks = 0;
		if(iNumBlendBlocks!=0)
		{
			iBlendBlocks = new int[3];
			iBlendBlocks[2] = iEdgesNum;
			for(int i=0;i<iEdgesNum;i++)
				if(!Edges[i].bNothing && !Edges[i].bPS)
				{
					iBlendBlocks[0] = i;
					iBlendBlocks[1] = i-1;
					if(iBlendBlocks[1]<0)
						iBlendBlocks[1] += iEdgesNum;
					break;
				}
		}
		if(iNumBlendBlocks==0)
		{
			int iCur=0, iCurBlock=0;
			while(1)
			{
				if(iCur < iEdgesNum-1)
				{
					while(iCur<iEdgesNum && (Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
						iCur++;
					iNumBlendBlocks++;
				}
				else
					break;
				if(iCur < iEdgesNum-1)
				{
					while(iCur<iEdgesNum && !(Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
						iCur++;
					iNumBlendBlocks++;
				}
				else
					break;
			}
			iNumBlendBlocks = (iNumBlendBlocks)/2;
			iBlendBlocks = new int[(iNumBlendBlocks+1)*3];
			//   
			iCur=0;
			while(1)
			{
				if(iCur < iEdgesNum-1)
				{
					bool bF = false;
					iBlendBlocks[iCurBlock*3] = iCur;
					iBlendBlocks[iCurBlock*3+1] = iCur-1;
					while(iCur<iEdgesNum && (Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
					{
						iCur++;
						iBlendBlocks[iCurBlock*3+1]++;
						bF = true;
					}
					if(bF == true)
						iCurBlock++;

				}
				else
					break;
				if(iCur < iEdgesNum-1)
				{
					while(iCur<iEdgesNum && !(Edges[iCur].bPS || Edges[iCur].bIgnore || !Edges[iCur].bNothing))
						iCur++;
				}
				else
					break;
			}
			if(iCurBlock == iNumBlendBlocks + 1)
				iBlendBlocks[0] = iBlendBlocks[(iCurBlock-1)*3];
		}
		for(int i=0;i<iNumBlendBlocks;i++)
		{
			iBlendBlocks[i*3+2] = 1;
			if((iBlendBlocks[i*3+1]-iBlendBlocks[i*3]>0?iBlendBlocks[i*3+1]-iBlendBlocks[i*3]:iBlendBlocks[i*3+1]-iBlendBlocks[i*3]+iEdgesNum) > 0)
			{
				int iCur = iBlendBlocks[i*3];
				while(iCur != iBlendBlocks[i*3+1])
				{
					int iPrev = iCur;
					iCur++;
					if(iCur >= iEdgesNum)
						iCur -= iEdgesNum;
					if(!Edges[iPrev].bNothing || !Edges[iCur].bNothing)
						iBlendBlocks[i*3+2]++;
				}
			}
		}

		int iBaseEdges;
		uf_list_p_t BaseEdges;

		for(int i=0;i<iNumBlendBlocks;i++)
		{
			//  
			tag_t Blend;
			UF_MODL_edge_blend_data_s BlendData;
			BlendData.blend_type = 0;
			BlendData.blend_instanced = FALSE;
			BlendData.blend_setback = FALSE;
			BlendData.vrb_tolerance = .1;
			BlendData.blend_radius[0] = '5';
			BlendData.blend_radius[1] = '\0';
			BlendData.smooth_overflow = TRUE;
			BlendData.cliff_overflow = TRUE;
			BlendData.notch_overflow = TRUE;

			//  
			UF_CALL(UF_MODL_ask_feat_edges(Base, &BaseEdges));
			UF_CALL(UF_MODL_ask_list_count(BaseEdges, &iBaseEdges));

			BlendData.number_edges = iBlendBlocks[i*3+2];
			BlendData.edge_data = new UF_MODL_blend_edge_s[iBlendBlocks[i*3+2]];
			int iCur = iBlendBlocks[i*3];
			int iCur2 = 0;
			while(iCur2 < iBlendBlocks[i*3+2])
			{
				int iPrev = iCur-1;
				if(iPrev<0)	iPrev += iEdgesNum;
				if(!Edges[iPrev].bNothing || !Edges[iCur].bNothing)
				{
					BlendData.edge_data[iCur2].edge = FindClosestEdge(SplinesTop[iCur], &BaseEdges);

					//   
					BlendData.edge_data[iCur2].number_points = 0;

					if(!Edges[iCur].bPS && !Edges[iCur].bNothing)
					{
						double dGuess[3], dMinDist, dParm, dParLimits[2];
						UF_EVAL_p_t Eval;

						UF_CALL(UF_EVAL_initialize(BlendData.edge_data[iCur2].edge, &Eval));
						UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

						//   
						for(int k=0;k<Edges[iCur].iSectionsNum;k++)
						{
							Edges[iCur].Sections[k].PointRealBase.Fill(dGuess);
							UF_CALL(UF_MODL_ask_minimum_dist(BlendData.edge_data[iCur2].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
							if(dMinDist < dG_DistTol && Edges[iCur].Sections[k].CalcPoints.r > .1 && Edges[iCur].Sections[k].CalcPoints.r < 100)
								BlendData.edge_data[iCur2].number_points++;
						}

						//  
						BlendData.edge_data[iCur2].point_data = new UF_MODL_blend_point_s[BlendData.edge_data[iCur2].number_points];
						BlendData.edge_data[iCur2].number_points = 0;

						//  
						for(int k=0;k<Edges[iCur].iSectionsNum;k++)
						{
							Edges[iCur].Sections[k].PointRealBase.Fill(dGuess);
							UF_CALL(UF_MODL_ask_minimum_dist(BlendData.edge_data[iCur2].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
							if(dMinDist < dG_DistTol && Edges[iCur].Sections[k].CalcPoints.r > .1 && Edges[iCur].Sections[k].CalcPoints.r < 100)
							{
								Edges[iCur].Sections[k].PointRealBase.Fill(dGuess);
								UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
								BlendData.edge_data[iCur2].point_data[BlendData.edge_data[iCur2].number_points].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
								string.Format("%f", Edges[iCur].Sections[k].CalcPoints.r);
//										string.Format("%f", 70.0);
								string.CopyChars(BlendData.edge_data[iCur2].point_data[BlendData.edge_data[iCur2].number_points].radius, string.GetBuffer(), string.GetLength());
								BlendData.edge_data[iCur2].point_data[BlendData.edge_data[iCur2].number_points].radius[string.GetLength()] = '\0';
								BlendData.edge_data[iCur2].number_points++;
							}
						}

						UF_CALL(UF_EVAL_free(Eval));
					}

				}

				iCur2++;
				iCur++;
				if(iCur>=iEdgesNum)	iCur -= iEdgesNum;
			}

			UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

			for(int j=0;j<BlendData.number_edges;j++)
				if(BlendData.edge_data[j].number_points)
					delete [] BlendData.edge_data[j].point_data;
			delete [] BlendData.edge_data;
		}
	for(int i=0;i<iEdgesNum;i++)
		if(Edges[i].bIgnore && Edges[i].bPS)
		{
			Edges[i].bPS = false;
			Edges[i].bNothing = true;
		}

	//   
	{
		tag_t Temp, Edge, *ReservedEdges = new tag_t[iEdgesNum*2];
		uf_list_p_t BodyEdges;
		int iBodyEdges;

		UF_CALL(UF_MODL_ask_feat_body(Base, &Temp));
		UF_CALL(UF_MODL_ask_body_edges(Temp, &BodyEdges));
		UF_CALL(UF_MODL_ask_list_count(BodyEdges, &iBodyEdges));
		for(int i=0;i<iEdgesNum;i++)
		{
			if(!Edges[i].bNothing || (Edges[i].bNothing && Edges[i].bIgnore))
				ReservedEdges[i] = FindClosestEdge(SplinesTop[i], &BodyEdges);
			else
				ReservedEdges[i] = NULL_TAG;
		}
		for(int i=0;i<iEdgesNum;i++)
			ReservedEdges[i+iEdgesNum] = FindClosestEdge(SplinesBottom[i], &BodyEdges);

		Service.dBaseRoundingRadius = RoundAllEdges(Temp, ReservedEdges, iEdgesNum*2, dG_RoundingsRadius);
		delete [] ReservedEdges;
	}
	}
/*	{
		bool bAllEdges=true;
		int iBaseEdges, iInts, iNumOfNotNothing;
		tag_t Blend, Edge;
		UF_MODL_edge_blend_data_s BlendData;
		UF_MODL_blend_edge_s *BlendEdge;
		UF_MODL_blend_point_s *BlendPoints;
		uf_list_p_t BaseEdges;
		UF_MODL_intersect_info_p_t *Ints;

		BlendData.blend_type = 0;
		BlendData.blend_instanced = FALSE;
		BlendData.blend_setback = FALSE;
		BlendData.vrb_tolerance = .1;
		BlendData.blend_radius[0] = '0';
		BlendData.blend_radius[1] = '\0';
		BlendData.smooth_overflow = TRUE;
		BlendData.cliff_overflow = TRUE;
		BlendData.notch_overflow = TRUE;
		BlendData.number_edges = 1;

		//   
		UF_CALL(UF_MODL_ask_feat_edges(Base, &BaseEdges));
		UF_CALL(UF_MODL_ask_list_count(BaseEdges, &iBaseEdges));

		for(int i=0;i<iEdgesNum;i++)
			if(Edges[i].bNothing)
			{
				bAllEdges = false;
				break;
			}

		if(bAllEdges)
		{
			BlendData.number_edges = iEdgesNum;
			BlendEdge = new UF_MODL_blend_edge_s[iEdgesNum];

			//   
			for(int i=0;i<iEdgesNum;i++)
				//  
				for(int j=0;j<iBaseEdges;j++)
				{
					UF_CALL(UF_MODL_ask_list_item(BaseEdges, j, &BlendEdge[i].edge));;
					if(CurvesCoincide(SplinesTop[i], BlendEdge[i].edge))
					{
						//   
						if(!Edges[i].bPS)
						{
							int iValidPoints=0;
							double dGuess[3], dMinDist, dParm, dParLimits[2];
							UF_EVAL_p_t Eval;

							UF_CALL(UF_EVAL_initialize(BlendEdge[i].edge, &Eval));
							UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

							//   
							for(int m=0;m<Edges[i].iSectionsNum;m++)
							{
								Edges[i].Sections[m].PointRealBase.Fill(dGuess);
								UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[i].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
								if(dMinDist < dG_DistTol && Edges[i].Sections[m].CalcPoints.r > 1e-2)
									iValidPoints++;
							}

							BlendEdge[i].number_points = iValidPoints;
							BlendEdge[i].point_data = new UF_MODL_blend_point_s[iValidPoints];

							iValidPoints = 0;
							for(int m=0;m<Edges[i].iSectionsNum;m++)
							{
								Edges[i].Sections[m].PointRealBase.Fill(dGuess);
								UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[i].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
								if(dMinDist < dG_DistTol && Edges[i].Sections[m].CalcPoints.r > 1e-2)
								{
									Edges[i].Sections[m].PointRealBase.Fill(dGuess);
									UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
									BlendEdge[i].point_data[iValidPoints].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
									string.Format("%f", Edges[i].Sections[m].CalcPoints.r);
									string.CopyChars(BlendEdge[i].point_data[iValidPoints].radius, string.GetBuffer(), string.GetLength());
									BlendEdge[i].point_data[iValidPoints].radius[string.GetLength()] = '\0';
									iValidPoints++;
								}
							}

							UF_CALL(UF_EVAL_free(Eval));
						}
						else
							BlendEdge[i].number_points = 0;

						break;
					}
				}
			BlendData.edge_data = BlendEdge;

			UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

			// 
			for(int i=0;i<iEdgesNum;i++)
				if(BlendEdge[i].number_points)
					delete [] BlendEdge[i].point_data;
			delete [] BlendEdge;
		}

		if(!bAllEdges)
		for(int i=0;i<iEdgesNum;i++)
			if(Edges[i].bNothing)
			{
				for(int j=i;j<iEdgesNum;j++)
					if(!Edges[j].bNothing)
					{
						iNumOfNotNothing = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(!Edges[k].bNothing)
								iNumOfNotNothing++;
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(!Edges[k].bNothing)
									iNumOfNotNothing++;

						BlendData.number_edges = iNumOfNotNothing;
						BlendEdge = new UF_MODL_blend_edge_s[iNumOfNotNothing];

						//   
						iNumOfNotNothing = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(!Edges[k].bNothing)
								//  
								for(int l=0;l<iBaseEdges;l++)
								{
									UF_CALL(UF_MODL_ask_list_item(BaseEdges, l, &BlendEdge[iNumOfNotNothing].edge));;
									if(CurvesCoincide(SplinesTop[k], BlendEdge[iNumOfNotNothing].edge))
									{
										//   
										if(!Edges[k].bPS)
										{
											int iValidPoints=0;
											double dGuess[3], dMinDist, dParm, dParLimits[2];
											UF_EVAL_p_t Eval;

											UF_CALL(UF_EVAL_initialize(BlendEdge[iNumOfNotNothing].edge, &Eval));
											UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

											//   
											for(int m=0;m<Edges[k].iSectionsNum;m++)
											{
												Edges[k].Sections[m].PointRealBase.Fill(dGuess);
												UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
												if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
													iValidPoints++;
											}

											BlendEdge[iNumOfNotNothing].number_points = iValidPoints;
											BlendEdge[iNumOfNotNothing].point_data = new UF_MODL_blend_point_s[iValidPoints];

											iValidPoints = 0;
											for(int m=0;m<Edges[k].iSectionsNum;m++)
											{
												Edges[k].Sections[m].PointRealBase.Fill(dGuess);
												UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
												if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
												{
													Edges[k].Sections[m].PointRealBase.Fill(dGuess);
													UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
													BlendEdge[iNumOfNotNothing].point_data[iValidPoints].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
													string.Format("%f", Edges[k].Sections[m].CalcPoints.r);
													string.CopyChars(BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius, string.GetBuffer(), string.GetLength());
													BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius[string.GetLength()] = '\0';
													iValidPoints++;
												}
											}

											UF_CALL(UF_EVAL_free(Eval));
										}
										else
											BlendEdge[iNumOfNotNothing].number_points = 0;
										iNumOfNotNothing++;
										break;
									}
								}
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(!Edges[k].bNothing)
									//  
									for(int l=0;l<iBaseEdges;l++)
									{
										UF_CALL(UF_MODL_ask_list_item(BaseEdges, l, &BlendEdge[iNumOfNotNothing].edge));;
										if(CurvesCoincide(SplinesTop[k], BlendEdge[iNumOfNotNothing].edge))
										{
											//   
											if(!Edges[k].bPS)
											{
												int iValidPoints=0;
												double dGuess[3], dMinDist, dParm, dParLimits[2];
												UF_EVAL_p_t Eval;

												UF_CALL(UF_EVAL_initialize(BlendEdge[iNumOfNotNothing].edge, &Eval));
												UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

												//   
												for(int m=0;m<Edges[k].iSectionsNum;m++)
												{
													Edges[k].Sections[m].PointRealBase.Fill(dGuess);
													UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
													if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
														iValidPoints++;
												}

												BlendEdge[iNumOfNotNothing].number_points = iValidPoints;
												BlendEdge[iNumOfNotNothing].point_data = new UF_MODL_blend_point_s[iValidPoints];

												iValidPoints = 0;
												for(int m=0;m<Edges[k].iSectionsNum;m++)
												{
													Edges[k].Sections[m].PointRealBase.Fill(dGuess);
													UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
													if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
													{
														Edges[k].Sections[m].PointRealBase.Fill(dGuess);
														UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
														BlendEdge[iNumOfNotNothing].point_data[iValidPoints].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
														string.Format("%f", Edges[k].Sections[m].CalcPoints.r);
														string.CopyChars(BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius, string.GetBuffer(), string.GetLength());
														BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius[string.GetLength()] = '\0';
														iValidPoints++;
													}
												}

												UF_CALL(UF_EVAL_free(Eval));
											}
											else
												BlendEdge[iNumOfNotNothing].number_points = 0;
											iNumOfNotNothing++;
											break;
										}
									}

						BlendData.edge_data = BlendEdge;

						UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

						// 
						for(k=0;k<iNumOfNotNothing;k++)
							if(BlendEdge[k].number_points)
								delete [] BlendEdge[k].point_data;
						delete [] BlendEdge;

						//    ,    ,
						//   
						UF_CALL(UF_MODL_delete_list(&BaseEdges));
						UF_CALL(UF_MODL_ask_feat_edges(Base, &BaseEdges));
						UF_CALL(UF_MODL_ask_list_count(BaseEdges, &iBaseEdges));

						j += iNumOfNotNothing;
					}
				if(!(!Edges[0].bNothing && Edges[iEdgesNum-1].bNothing))
					break;
				for(int j=0;j<iEdgesNum;j++)
					if(!Edges[j].bNothing)
					{
						iNumOfNotNothing = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(!Edges[k].bNothing)
								iNumOfNotNothing++;
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(!Edges[k].bNothing)
									iNumOfNotNothing++;

						BlendData.number_edges = iNumOfNotNothing;
						BlendEdge = new UF_MODL_blend_edge_s[iNumOfNotNothing];

						//   
						iNumOfNotNothing = 0;
						for(int k=j;k<iEdgesNum;k++)
							if(!Edges[k].bNothing)
								//  
								for(int l=0;l<iBaseEdges;l++)
								{
									UF_CALL(UF_MODL_ask_list_item(BaseEdges, l, &BlendEdge[iNumOfNotNothing].edge));;
									if(CurvesCoincide(SplinesTop[k], BlendEdge[iNumOfNotNothing].edge))
									{
										//   
										if(!Edges[k].bPS)
										{
											int iValidPoints=0;
											double dGuess[3], dMinDist, dParm, dParLimits[2];
											UF_EVAL_p_t Eval;

											UF_CALL(UF_EVAL_initialize(BlendEdge[iNumOfNotNothing].edge, &Eval));
											UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

											//   
											for(int m=0;m<Edges[k].iSectionsNum;m++)
											{
												Edges[k].Sections[m].PointRealBase.Fill(dGuess);
												UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
												if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
													iValidPoints++;
											}

											BlendEdge[iNumOfNotNothing].number_points = iValidPoints;
											BlendEdge[iNumOfNotNothing].point_data = new UF_MODL_blend_point_s[iValidPoints];

											iValidPoints = 0;
											for(int m=0;m<Edges[k].iSectionsNum;m++)
											{
												Edges[k].Sections[m].PointRealBase.Fill(dGuess);
												UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
												if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
												{
													Edges[k].Sections[m].PointRealBase.Fill(dGuess);
													UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
													BlendEdge[iNumOfNotNothing].point_data[iValidPoints].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
													string.Format("%f", Edges[k].Sections[m].CalcPoints.r);
													string.CopyChars(BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius, string.GetBuffer(), string.GetLength());
													BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius[string.GetLength()] = '\0';
													iValidPoints++;
												}
											}

											UF_CALL(UF_EVAL_free(Eval));
										}
										else
											BlendEdge[iNumOfNotNothing].number_points = 0;
										iNumOfNotNothing++;
										break;
									}
								}
							else
								break;
						if(k==iEdgesNum)
							for(int k=0;k<i;k++)
								if(!Edges[k].bNothing)
									//  
									for(int l=0;l<iBaseEdges;l++)
									{
										UF_CALL(UF_MODL_ask_list_item(BaseEdges, l, &BlendEdge[iNumOfNotNothing].edge));;
										if(CurvesCoincide(SplinesTop[k], BlendEdge[iNumOfNotNothing].edge))
										{
											//   
											if(!Edges[k].bPS)
											{
												int iValidPoints=0;
												double dGuess[3], dMinDist, dParm, dParLimits[2];
												UF_EVAL_p_t Eval;

												UF_CALL(UF_EVAL_initialize(BlendEdge[iNumOfNotNothing].edge, &Eval));
												UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

												//   
												for(int m=0;m<Edges[k].iSectionsNum;m++)
												{
													Edges[k].Sections[m].PointRealBase.Fill(dGuess);
													UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
													if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
														iValidPoints++;
												}

												BlendEdge[iNumOfNotNothing].number_points = iValidPoints;
												BlendEdge[iNumOfNotNothing].point_data = new UF_MODL_blend_point_s[iValidPoints];

												iValidPoints = 0;
												for(int m=0;m<Edges[k].iSectionsNum;m++)
												{
													Edges[k].Sections[m].PointRealBase.Fill(dGuess);
													UF_CALL(UF_MODL_ask_minimum_dist(BlendEdge[iNumOfNotNothing].edge, NULL_TAG, 0, dGuess, 1, dGuess, &dMinDist, dGuess, dGuess));
													if(dMinDist < dG_DistTol && Edges[k].Sections[m].CalcPoints.r > 1e-2)
													{
														Edges[k].Sections[m].PointRealBase.Fill(dGuess);
														UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
														BlendEdge[iNumOfNotNothing].point_data[iValidPoints].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
														string.Format("%f", Edges[k].Sections[m].CalcPoints.r);
														string.CopyChars(BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius, string.GetBuffer(), string.GetLength());
														BlendEdge[iNumOfNotNothing].point_data[iValidPoints].radius[string.GetLength()] = '\0';
														iValidPoints++;
													}
												}

												UF_CALL(UF_EVAL_free(Eval));
											}
											else
												BlendEdge[iNumOfNotNothing].number_points = 0;
											iNumOfNotNothing++;
											break;
										}
									}

						BlendData.edge_data = BlendEdge;

						UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

						// 
						for(k=0;k<iNumOfNotNothing;k++)
							if(BlendEdge[k].number_points)
								delete [] BlendEdge[k].point_data;
						delete [] BlendEdge;

						//    ,    ,
						//   
						UF_CALL(UF_MODL_delete_list(&BaseEdges));
						UF_CALL(UF_MODL_ask_feat_edges(Base, &BaseEdges));
						UF_CALL(UF_MODL_ask_list_count(BaseEdges, &iBaseEdges));

						break;
					}

				break;
			}

	for(int i=0;i<iEdgesNum;i++)
		if(Edges[i].bIgnore && Edges[i].bPS)
		{
			Edges[i].bPS = false;
			Edges[i].bNothing = true;
		}

	//   
	{
		tag_t Temp, Edge, *ReservedEdges = new tag_t[iEdgesNum*2];
		uf_list_p_t BodyEdges;
		int iBodyEdges;

		UF_CALL(UF_MODL_ask_feat_body(Base, &Temp));
		UF_CALL(UF_MODL_ask_body_edges(Temp, &BodyEdges));
		UF_CALL(UF_MODL_ask_list_count(BodyEdges, &iBodyEdges));
		for(int i=0;i<iEdgesNum;i++)
		{
			ReservedEdges[i] = NULL_TAG;
			if(!Edges[i].bNothing || (Edges[i].bNothing && Edges[i].bIgnore))
			for(int j=0;j<iBodyEdges;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(BodyEdges, j, &Edge));
				if(CurvesCoincide(SplinesTop[i], Edge))
				{
					ReservedEdges[i] = Edge;
					break;
				}
			}
		}
		for(int i=0;i<iEdgesNum;i++)
		{
			ReservedEdges[i+iEdgesNum] = NULL_TAG;
			for(int j=0;j<iBodyEdges;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(BodyEdges, j, &Edge));
				if(CurvesCoincide(SplinesBottom[i], Edge, dG_DistTol*10))
				{
					ReservedEdges[i+iEdgesNum] = Edge;
					break;
				}
			}
		}

		RoundAllEdges(Temp, ReservedEdges, iEdgesNum*2, dG_RoundingsRadius);
		delete [] ReservedEdges;
	}

	}*/

	//  
	for(int i=0;i<iEdgesNum;i++)
	{
		if(SplinesTop[i] != NULL_TAG)
			UF_CALL(UF_OBJ_set_blank_status(SplinesTop[i], UF_OBJ_BLANKED));
		if(SplinesBottom[i] != NULL_TAG)
			UF_CALL(UF_OBJ_set_blank_status(SplinesBottom[i], UF_OBJ_BLANKED));
	}

	//   - ""
	{
		tag_t Temp;
		UF_CALL(UF_MODL_ask_feat_body(Base, &Temp));
		Base = Temp;
	}

	delete [] Edges;

	return Base;
}

//     ( )
double CPart::CalcMinBaseHeight(void)
{
double dBaseHeight=0;

	for(int i=0;i<iEdgesNum;i++)
		if(!Edges[i].bPS && !Edges[i].bNothing)
			for(int j=0;j<Edges[i].iSectionsNum;j++)
				if(Edges[i].Sections[j].CalcPoints.H * cos(Edges[i].Sections[j].CalcPoints.Alpha) > dBaseHeight)
					dBaseHeight = Edges[i].Sections[j].CalcPoints.H * cos(Edges[i].Sections[j].CalcPoints.Alpha);

	return dBaseHeight;
}

//   
void CPart::ScanPress(tag_t Base, int iIntervalsNum)
{
uf_list_p_t OuterEdges;
tag_t Curve;
double dLength, dEdgesLength=0, dSegmentSize;
double dCurveLength, dEvalLimits[2];
double dPoint[3], dTangent[3], dNormal[3], dBiNormal[3];
int iNumOfPoints, iNumber, iNumOfPointsNew;
UF_EVAL_p_t Evaluator;

	UF_CALL(UF_MODL_create_list(&OuterEdges));
	GetOuterEdges(Base, &OuterEdges);
	UF_CALL(UF_MODL_ask_list_count(OuterEdges, &iPressEdgesNum));

	// 
	iPressPointsNum = new int[iPressEdgesNum];
	PointsPress = new CPoint3D*[iPressEdgesNum];

	//   
	for(int i=0;i<iPressEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(OuterEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));
		dEdgesLength += dLength;
	}

	//       
	dSegmentSize = dEdgesLength/double(iIntervalsNum);
	for(int i=0;i<iPressEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(OuterEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));

		iNumOfPoints = dLength/dSegmentSize + 1.5;
		if(iNumOfPoints<4)
			iNumOfPoints = 4;

		iNumOfPointsNew = iNumOfPoints;

		//,   
		UF_CALL(UF_CURVE_ask_parameterization(Curve, dEvalLimits, &iNumber));
		if(iNumber==UF_CURVE_CLOSED_PERIODIC_CURVE || iNumber==UF_CURVE_CLOSED_NON_PERIODIC_CURVE)
			iNumOfPointsNew--;

		//   
		iPressPointsNum[i] = iNumOfPointsNew;
		PointsPress[i] = new CPoint3D[iNumOfPointsNew];

		// 
		UF_CALL(UF_EVAL_initialize(Curve, &Evaluator));
		UF_CALL(UF_EVAL_ask_limits(Evaluator, dEvalLimits));

		// 
		for(int j=0;j<iNumOfPointsNew;j++)
		{
			//   
			UF_CALL(UF_EVAL_evaluate_unit_vectors(
				Evaluator,
				dEvalLimits[0] + (double(j)/double(iNumOfPoints-1))*(dEvalLimits[1]-dEvalLimits[0]),
				dPoint,
				dTangent,
				dNormal,
				dBiNormal));

			//    
			PointsPress[i][j] = dPoint;
		}
	}

	UF_CALL(UF_MODL_delete_list(&OuterEdges));
}

//  
tag_t CPart::CreatePress(void)
{
tag_t Press;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity;
double dPointOrigin[3], dTol[3]={.01, .01, 1};
UF_STRING_t Section;

	UF_MODL_create_string_list(iPressEdgesNum, iPressEdgesNum, &Section);
	Section.num = 1;
	Section.string[0] = iPressEdgesNum;

	//  
	for(int i=0;i<iPressEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(iPressPointsNum[i]<iDegree+1)
			iDegree = iPressPointsNum[i]-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[iPressPointsNum[i]];

		for(int j=0;j<iPressPointsNum[i];j++)
		{
			PointsPress[i][j].Fill(dPointOrigin);
			OffsetPoint(dPointOrigin, Points[j].point, Service.dBaseNormal, dG_BasePressDistance);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			iPressPointsNum[i],
			Points,
			NULL,
			0,
			&Section.id[i]));

		Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;

		delete [] Points;
	}

	//  
	UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Press));

	//  
	for(int i=0;i<iPressEdgesNum;i++)
		UF_CALL(UF_OBJ_set_blank_status(Section.id[i], UF_OBJ_BLANKED));
	UF_CALL(UF_OBJ_set_blank_status(Press, UF_OBJ_BLANKED));
	UF_MODL_free_string_list(&Section);

	//    Extrude
	{
		uf_list_p_t ExtrObjects, ExtrFeats;
		char *cTaper="0", *cLimit[2]={"0", "0"};
		CString string;
	
		UF_CALL(UF_MODL_create_list(&ExtrObjects));
		UF_CALL(UF_MODL_put_list_item(ExtrObjects, Press));

		string.Format("%f", dG_PressThickness);
		cLimit[1] = string.GetBuffer();

		UF_CALL(UF_MODL_create_extruded(ExtrObjects, cTaper, cLimit, dPointOrigin, Service.dBaseNormal, UF_NULLSIGN, &ExtrFeats));

		UF_CALL(UF_MODL_ask_list_item(ExtrFeats, 0, &Press));

		UF_CALL(UF_MODL_delete_list(&ExtrObjects));
		UF_CALL(UF_MODL_delete_list(&ExtrFeats));
	}

	{
		tag_t Temp;
		UF_CALL(UF_MODL_ask_feat_body(Press, &Temp));
		Press = Temp;
	}

	return Press;
}
tag_t CPart::CreatePress1(void)
{
tag_t Press=NULL_TAG, Press1, BridgedCurves[2], Plane, Plane1;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity, iCur, iNext, iReverse[2]={0,0};
double dPointOrigin[3], dTol[3]={.01, .01, 1};
UF_STRING_t Section;
double dSign=-1, dMeanR, dParms[2]={1,0};
UF_CURVE_offset_data_t OffData;
UF_CURVE_offset_distance_data_t OffDist;
UF_STRING_t Curves;
CString string;
tag_t *ReservedEdges,  *ReservedEdges1;

	UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
	Section.num = 1;
	Section.string[0] = iEdgesNum;

	//       
	for(int i=0;i<iEdgesNum;i++)
		Section.id[i] = NULL_TAG;

	//  
	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bIgnore && !Edges[i].bPS)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
			OffsetPoint(dPointOrigin, Points[j].point, Service.dBaseNormal, dG_BasePressDistance);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Section.id[i]));

		Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;

		delete [] Points;

		// ,  
		if(!Edges[i].bNothing)
		{
			dMeanR = Edges[i].CalcMeanR();
			if(dMeanR>3)
			{
				int t1=1, t2=1, iNumber=1;
				tag_t *OffCurve;

				Curves.num = 1;
				Curves.string = &t1;
				Curves.dir = &t2;
				Curves.id = &Section.id[i];

				string.Format("%f", dSign*dMeanR/2.0);
				OffDist.distance = string.GetBuffer();

				OffData.approximation_tolerance = .1;
				OffData.string_tolerance = .1;
				OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
				OffData.offset_def.distance_type1 = &OffDist;
				OffData.input_curves = &Curves;

				UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

				UF_CALL(UF_ASSEM_remove_instance(Section.id[i]));
				Section.id[i] = *OffCurve;

				UF_free(OffCurve);
			}
		}
		if(Edges[i].bNothing && !Edges[i].bIgnore && Edges[i].dOffset!=0)
		{
			int t1=1, t2=1, iNumber=1;
			tag_t *OffCurve;

			Curves.num = 1;
			Curves.string = &t1;
			Curves.dir = &t2;
			Curves.id = &Section.id[i];

			string.Format("%f", Edges[i].dOffset<0?Edges[i].dOffset+Service.dBaseRoundingRadius:Edges[i].dOffset-Service.dBaseRoundingRadius);
			OffDist.distance = string.GetBuffer();

			OffData.approximation_tolerance = .1;
			OffData.string_tolerance = .1;
			OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
			OffData.offset_def.distance_type1 = &OffDist;
			OffData.input_curves = &Curves;

			UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

			UF_CALL(UF_ASSEM_remove_instance(Section.id[i]));
			Section.id[i] = *OffCurve;

			UF_free(OffCurve);
		}
	}

	// ,   
	for(int i=0;i<iEdgesNum;i++)
	{
		iNext = i!=iEdgesNum-1?i+1:0;

		if(!Edges[i].bIgnore && !Edges[i].bPS && !Edges[iNext].bIgnore && !Edges[iNext].bPS)
			ConnectCurves(Section.id[i], Section.id[iNext]);
	}

	//  ,  
	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bIgnore && !Edges[i].bPS)
	{
		for(int j=i;j<i+iEdgesNum;j++)
		{
			iCur = j>=iEdgesNum?j-iEdgesNum:j;
			iNext = iCur;
			do
				iNext = iNext>=iEdgesNum-1?0:iNext+1;
			while(Edges[iNext].bIgnore || Edges[iNext].bPS);
			if(iNext != (iCur>=iEdgesNum-1?0:iCur+1))
			{
				BridgedCurves[0] = Section.id[iCur];
				BridgedCurves[1] = Section.id[iNext];
				UF_CALL(UF_CURVE_create_bridge_curve(UF_CURVE_TANGENT, BridgedCurves, dParms, iReverse, &Section.id[iCur>=iEdgesNum-1?0:iCur+1]));
				j = iNext>iCur?iNext-1:iNext-1+iEdgesNum;
			}
			break;
		}
	}
	//      ,  
	iCur = 0;
	for(int i=0;i<iEdgesNum;i++)
	{
		if(Section.id[i] != NULL_TAG && iCur != i)
		{
			Section.id[iCur] = Section.id[i];
			Section.id[i] = NULL_TAG;
			iCur++;
		}
		if(Section.id[i] != NULL_TAG)
			iCur++;
	}
	Section.string[0] = iCur;

	for(int i=0;i<Section.string[0];i++)
		Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;

	//  
	{
		int iL;
		UF_CALL(UF_LAYER_ask_work_layer(&iL));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane));
		UF_CALL(UF_LAYER_set_status(iL, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_INACTIVE_LAYER));
	}

	//    
	for(int i=0;i<Section.string[0];i++)
		UF_CALL(UF_OBJ_set_blank_status(Section.id[i], UF_OBJ_BLANKED));
//	UF_CALL(UF_OBJ_set_translucency(Plane, 100));
	UF_CALL(UF_OBJ_set_blank_status(Plane, UF_OBJ_BLANKED));

	//    Extrude
	{
		uf_list_p_t ExtrObjects, ExtrFeats;
		char *cTaper="0", *cLimit[2]={"0", "0"};
		CString string;
	
		UF_CALL(UF_MODL_create_list(&ExtrObjects));
		UF_CALL(UF_MODL_put_list_item(ExtrObjects, Plane));

		string.Format("%f", dG_PressThickness);
		cLimit[1] = string.GetBuffer();

		UF_CALL(UF_MODL_create_extruded(ExtrObjects, cTaper, cLimit, dPointOrigin, Service.dBaseNormal, UF_NULLSIGN, &ExtrFeats));

		UF_CALL(UF_MODL_ask_list_item(ExtrFeats, 0, &Press));

		UF_CALL(UF_MODL_delete_list(&ExtrObjects));
		UF_CALL(UF_MODL_delete_list(&ExtrFeats));
	}

	{
		tag_t Temp;
		UF_CALL(UF_MODL_ask_feat_body(Press, &Temp));
		Press = Temp;
	}

	//    ,  
	{
		ReservedEdges = new tag_t[Section.string[0]];
		uf_list_p_t BodyEdges;
		int iBodyEdges;

		UF_CALL(UF_MODL_ask_body_edges(Press, &BodyEdges));
		for(int i=0;i<Section.string[0];i++)
			ReservedEdges[i] = FindClosestEdge(Section.id[i], &BodyEdges);

		UF_MODL_delete_list(&BodyEdges);
	}

	//     ,    
	dSign = 1;
	//       
	for(int i=0;i<iEdgesNum;i++)
		Section.id[i] = NULL_TAG;

	//  
	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bIgnore && !Edges[i].bPS)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
			OffsetPoint(dPointOrigin, Points[j].point, Service.dBaseNormal, dG_BasePressDistance);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Section.id[i]));

		Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;

		delete [] Points;

		// ,  
		if(!Edges[i].bNothing)
		{
			dMeanR = Edges[i].CalcMeanR();
			if(dMeanR>3)
			{
				int t1=1, t2=1, iNumber=1;
				tag_t *OffCurve;

				Curves.num = 1;
				Curves.string = &t1;
				Curves.dir = &t2;
				Curves.id = &Section.id[i];

				string.Format("%f", dSign*dMeanR/2.0);
				OffDist.distance = string.GetBuffer();

				OffData.approximation_tolerance = .1;
				OffData.string_tolerance = .1;
				OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
				OffData.offset_def.distance_type1 = &OffDist;
				OffData.input_curves = &Curves;

				UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

				UF_CALL(UF_ASSEM_remove_instance(Section.id[i]));
				Section.id[i] = *OffCurve;

				UF_free(OffCurve);
			}
		}
		if(Edges[i].bNothing && !Edges[i].bIgnore && Edges[i].dOffset!=0)
		{
			int t1=1, t2=1, iNumber=1;
			tag_t *OffCurve;

			Curves.num = 1;
			Curves.string = &t1;
			Curves.dir = &t2;
			Curves.id = &Section.id[i];

			string.Format("%f", Edges[i].dOffset);
			OffDist.distance = string.GetBuffer();

			OffData.approximation_tolerance = .1;
			OffData.string_tolerance = .1;
			OffData.offset_type = UF_CURVE_OFFSET_DISTANCE_NO_TRIM;
			OffData.offset_def.distance_type1 = &OffDist;
			OffData.input_curves = &Curves;

			UF_CALL(UF_CURVE_create_offset_curve(&OffData, &iNumber, &OffCurve));

			UF_CALL(UF_ASSEM_remove_instance(Section.id[i]));
			Section.id[i] = *OffCurve;

			UF_free(OffCurve);
		}
	}

	// ,   
	for(int i=0;i<iEdgesNum;i++)
	{
		iNext = i!=iEdgesNum-1?i+1:0;

		if(!Edges[i].bIgnore && !Edges[i].bPS && !Edges[iNext].bIgnore && !Edges[iNext].bPS)
			ConnectCurves(Section.id[i], Section.id[iNext]);
	}

	//  ,  
	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bIgnore && !Edges[i].bPS)
	{
		for(int j=i;j<i+iEdgesNum;j++)
		{
			iCur = j>=iEdgesNum?j-iEdgesNum:j;
			iNext = iCur;
			do
				iNext = iNext>=iEdgesNum-1?0:iNext+1;
			while(Edges[iNext].bIgnore || Edges[iNext].bPS);
			if(iNext != (iCur>=iEdgesNum-1?0:iCur+1))
			{
				BridgedCurves[0] = Section.id[iCur];
				BridgedCurves[1] = Section.id[iNext];
				UF_CALL(UF_CURVE_create_bridge_curve(UF_CURVE_TANGENT, BridgedCurves, dParms, iReverse, &Section.id[iCur>=iEdgesNum-1?0:iCur+1]));
				j = iNext>iCur?iNext-1:iNext-1+iEdgesNum;
			}
		}
		break;
	}

	//      ,  
	iCur = 0;
	for(int i=0;i<iEdgesNum;i++)
	{
		if(Section.id[i] != NULL_TAG && iCur != i)
			Section.id[iCur] = Section.id[i];
		if(Section.id[i] != NULL_TAG)
			iCur++;
	}
	Section.string[0] = iCur;

	//  
	{
		int iL;
		UF_CALL(UF_LAYER_ask_work_layer(&iL));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane1));
		UF_CALL(UF_LAYER_set_status(iL, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_INACTIVE_LAYER));
	}

	//    
	for(int i=0;i<Section.string[0];i++)
		UF_CALL(UF_OBJ_set_blank_status(Section.id[i], UF_OBJ_BLANKED));
//	UF_CALL(UF_OBJ_set_translucency(Plane1, 100));
	UF_CALL(UF_OBJ_set_blank_status(Plane1, UF_OBJ_BLANKED));

	//    Extrude
	{
		uf_list_p_t ExtrObjects, ExtrFeats;
		char *cTaper="0", *cLimit[2]={"0", "0"};
		CString string;
	
		UF_CALL(UF_MODL_create_list(&ExtrObjects));
		UF_CALL(UF_MODL_put_list_item(ExtrObjects, Plane1));

		string.Format("%f", dG_PressThickness);
		cLimit[1] = string.GetBuffer();

		UF_CALL(UF_MODL_create_extruded(ExtrObjects, cTaper, cLimit, dPointOrigin, Service.dBaseNormal, UF_NULLSIGN, &ExtrFeats));

		UF_CALL(UF_MODL_ask_list_item(ExtrFeats, 0, &Press1));

		UF_CALL(UF_MODL_delete_list(&ExtrObjects));
		UF_CALL(UF_MODL_delete_list(&ExtrFeats));
	}

	{
		tag_t Temp;
		UF_CALL(UF_MODL_ask_feat_body(Press1, &Temp));
		Press1 = Temp;
	}

	//    ,  
	{
		ReservedEdges1 = new tag_t[Section.string[0]];
		uf_list_p_t BodyEdges;
		int iBodyEdges;

		UF_CALL(UF_MODL_ask_body_edges(Press, &BodyEdges));
		for(int i=0;i<Section.string[0];i++)
			ReservedEdges[i] = FindClosestEdge(Section.id[i], &BodyEdges);

		UF_MODL_delete_list(&BodyEdges);
	}

	//    
	{
		double acc_value[11], mass_props[47], statistics[13], dVol, dVol1;

		acc_value[0] = .99;

		UF_CALL(UF_MODL_ask_mass_props_3d(&Press, 1, 1, 4, 1, 1, acc_value, mass_props, statistics));
		dVol = mass_props[1];

		UF_CALL(UF_MODL_ask_mass_props_3d(&Press1, 1, 1, 4, 1, 1, acc_value, mass_props, statistics));
		dVol1 = mass_props[1];

		if(dVol<dVol1)
		{
			UF_CALL(UF_OBJ_delete_object(Press1));
			UF_CALL(UF_OBJ_delete_object(Plane1));
			//  ,    
			RoundAllEdges(Press, ReservedEdges, Section.string[0], dG_RoundingsRadius);
		}
		else
		{
			UF_CALL(UF_OBJ_delete_object(Press));
			UF_CALL(UF_OBJ_delete_object(Plane));
			Press = Press1;
			RoundAllEdges(Press, ReservedEdges1, Section.string[0], dG_RoundingsRadius);
		}
	}

	delete [] ReservedEdges;
	delete [] ReservedEdges1;
	UF_MODL_free_string_list(&Section);

	return Press;
}

//  
tag_t CPart::CreateInvolute(void)
{
tag_t *Splines = new tag_t[iEdgesNum], *Splines2 = new tag_t[iEdgesNum], *PunchSplines, Plane, Plane2;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity, iSplinesNum=0;
UF_STRING_t Section;
double dTol[3]={.01, .01, 1}, dL, dAngle;
double dPoint[3], dVector[3], dMinDist1, dMinDist2;
CPoint3D P1, P2;
bool bInner;
UF_CURVE_line_s LineCoords;

	//   ,   ,  
	for(int i=0;i<iEdgesNum;i++)
		Splines[i] = Splines2[i] = NULL_TAG;

	//  RealBase
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Splines[i]));

		delete [] Points;
	}

	int iCounter = 0;
	for(int i=0;i<iPunchesNum;i++)
		iCounter += Punches[i].iEdgesNum;
	PunchSplines = new tag_t[iCounter];
	UF_MODL_create_string_list(iEdgesNum+iCounter, iEdgesNum+iCounter, &Section);
	Section.num = 1 + iPunchesNum;
	Section.string[0] = iEdgesNum;
	for(int i=0;i<iEdgesNum;i++)
	{
		Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
		Section.id[i] = Splines[i];
	}
	iCounter = iEdgesNum;
	//    
	for(int h=0;h<iPunchesNum;h++)
	{
		Section.string[h+1] = Punches[h].iEdgesNum;
		for(int i=0;i<Punches[h].iEdgesNum;i++)
		{
			Section.dir[iCounter] = UF_MODL_CURVE_START_FROM_BEGIN;

			iDegree = iSplinesDegree;

			//,        
			if(Punches[h].Edges[i].iSectionsNum<iDegree+1)
				iDegree = Punches[h].Edges[i].iSectionsNum-1;

			Points = new UF_CURVE_pt_slope_crvatr_s[Punches[h].Edges[i].iSectionsNum];

			for(int j=0;j<Punches[h].Edges[i].iSectionsNum;j++)
			{
				Punches[h].Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
				Points[j].slope_type = UF_CURVE_SLOPE_NONE;
				Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
			}

			if(Punches[h].iEdgesNum==1)
				iPeriodicity = 1;
			else
				iPeriodicity = 0;

			//    
			UF_CALL(UF_CURVE_create_spline_thru_pts(
				iDegree,
				iPeriodicity,
				Punches[h].Edges[i].iSectionsNum,
				Points,
				NULL,
				0,
				&Section.id[iCounter]));

			UF_CALL(UF_OBJ_set_blank_status(Section.id[iCounter++], UF_OBJ_BLANKED));

			delete [] Points;
		}
	}
	UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane));

	//     
	iCounter = iEdgesNum;
	for(int h=0;h<iPunchesNum;h++)
	for(int i=0;i<Punches[h].iEdgesNum;i++)
	{
		double dGuess[3], dMinDist, dParm, dParLimits[2];
		double dPoint1[3], dTangent1[3], dNormal1[3], dBiNormal1[3];
		UF_EVAL_p_t Eval;

		iDegree = iSplinesDegree;

		//,        
		if(Punches[h].Edges[i].iSectionsNum<iDegree+1)
			iDegree = Punches[h].Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Punches[h].Edges[i].iSectionsNum];

		UF_CALL(UF_EVAL_initialize(Section.id[iCounter], &Eval));
		UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

		for(int j=0;j<Punches[h].Edges[i].iSectionsNum;j++)
		{
			//     
			dL = (Punches[h].Edges[i].Sections[j].PointRounding-Punches[h].Edges[i].Sections[j].PointBevel);

			P1.X = Punches[h].Edges[i].Sections[j].PointBevel.X - Punches[h].Edges[i].Sections[j].PointRealBase.X;
			P1.Y = Punches[h].Edges[i].Sections[j].PointBevel.Y - Punches[h].Edges[i].Sections[j].PointRealBase.Y;
			P1.Z = Punches[h].Edges[i].Sections[j].PointBevel.Z - Punches[h].Edges[i].Sections[j].PointRealBase.Z;
			P2.X = Punches[h].Edges[i].Sections[j].PointBase.X - Punches[h].Edges[i].Sections[j].PointRealBase.X;
			P2.Y = Punches[h].Edges[i].Sections[j].PointBase.Y - Punches[h].Edges[i].Sections[j].PointRealBase.Y;
			P2.Z = Punches[h].Edges[i].Sections[j].PointBase.Z - Punches[h].Edges[i].Sections[j].PointRealBase.Z;
			dAngle = CalcAngle(P1, P2);

			dL += sqrt(Punches[h].Edges[i].Sections[j].r*(Punches[h].Edges[i].Sections[j].r+t)) * dAngle;

			//   
			dL -= Punches[h].Edges[i].Sections[j].PointRealBase - Punches[h].Edges[i].Sections[j].PointRounding;

			Punches[h].Edges[i].Sections[j].PointRealBase.Fill(dGuess);
			UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
			UF_CALL(UF_EVAL_evaluate_unit_vectors(
				Eval,
				dParm,
				dPoint1,
				dTangent1,
				dNormal1,
				dBiNormal1));

			FindThirdVector(Service.dBaseNormal, dTangent1, dVector);
			Punches[h].Edges[i].Sections[j].PointRealBase.Fill(dPoint1);
			OffsetPoint(dPoint1, Points[j].point, dVector, 1);

			//,     
			UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, Points[j].point, 1, Points[j].point, &dMinDist1, dPoint1, dPoint1));
			Punches[h].Edges[i].Sections[j].PointRealBase.Fill(dPoint1);
			OffsetPoint(dPoint1, Points[j].point, dVector, -1);
			UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, Points[j].point, 1, Points[j].point, &dMinDist2, dPoint1, dPoint1));
			if(dMinDist1<dMinDist2)
			{
				dVector[0] *= -1.0;
				dVector[1] *= -1.0;
				dVector[2] *= -1.0;
			}

			Punches[h].Edges[i].Sections[j].PointRealBase.Fill(dPoint);
			OffsetPoint(dPoint,	Points[j].point, dVector, dL);

			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(Punches[h].iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Punches[h].Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&PunchSplines[iCounter-iEdgesNum]));

		UF_CALL(UF_OBJ_set_blank_status(PunchSplines[iCounter++ - iEdgesNum], UF_OBJ_BLANKED));

		delete [] Points;
	}
	UF_MODL_free_string_list(&Section);

	//   
	for(int i=0;i<iEdgesNum;i++)
	if(!Edges[i].bNothing && !Edges[i].bPS)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			//     
			dL = (Edges[i].Sections[j].PointRounding-Edges[i].Sections[j].PointBevel) + (Edges[i].Sections[j].PointUndercutRounding-Edges[i].Sections[j].PointUndercut);

			if(Edges[i].Sections[j].bIsUndercutRoundingPresent)
			{
				P1.X = Edges[i].Sections[j].PointRealBase.X - Edges[i].Sections[j].PointBase.X;
				P1.Y = Edges[i].Sections[j].PointRealBase.Y - Edges[i].Sections[j].PointBase.Y;
				P1.Z = Edges[i].Sections[j].PointRealBase.Z - Edges[i].Sections[j].PointBase.Z;
				P2.X = Edges[i].Sections[j].PointRealBevel.X - Edges[i].Sections[j].PointUndercutRounding.X;
				P2.Y = Edges[i].Sections[j].PointRealBevel.Y - Edges[i].Sections[j].PointUndercutRounding.Y;
				P2.Z = Edges[i].Sections[j].PointRealBevel.Z - Edges[i].Sections[j].PointUndercutRounding.Z;
				if(CalcAngle(P1, P2)<PI/2)
					bInner = true;
				else
					bInner = false;

				//     
				P1.X = Edges[i].Sections[j].PointBevel.X - Edges[i].Sections[j].PointRealBevel.X;
				P1.Y = Edges[i].Sections[j].PointBevel.Y - Edges[i].Sections[j].PointRealBevel.Y;
				P1.Z = Edges[i].Sections[j].PointBevel.Z - Edges[i].Sections[j].PointRealBevel.Z;
				P2.X = Edges[i].Sections[j].PointUndercutRounding.X - Edges[i].Sections[j].PointRealBevel.X;
				P2.Y = Edges[i].Sections[j].PointUndercutRounding.Y - Edges[i].Sections[j].PointRealBevel.Y;
				P2.Z = Edges[i].Sections[j].PointUndercutRounding.Z - Edges[i].Sections[j].PointRealBevel.Z;
				dAngle = PI - CalcAngle(P1, P2);

				bInner ? dL += sqrt(Edges[i].Sections[j].rUndercut*(Edges[i].Sections[j].rUndercut+t)) * dAngle : dL += sqrt(Edges[i].Sections[j].rUndercut*(Edges[i].Sections[j].rUndercut-t)) * dAngle;
			}

			if(Edges[i].Sections[j].bIsRoundingPresent)
			{
				if(Edges[i].Sections[j].Alpha < PI/2)
					dL += sqrt(Edges[i].Sections[j].r*(Edges[i].Sections[j].r+t)) * (PI/2+Edges[i].Sections[j].Alpha);
				else
					dL += sqrt(Edges[i].Sections[j].r*(Edges[i].Sections[j].r-t)) * (PI/2+Edges[i].Sections[j].Alpha);

				//   
				dL -= Edges[i].Sections[j].PointRealBase - Edges[i].Sections[j].PointRounding;
			}

			Edges[i].Sections[j].PlaneSectionNormal.Fill(dPoint);
			FindThirdVector(Service.dBaseNormal, dPoint, dVector);
			Edges[i].Sections[j].PointRealBase.Fill(dPoint);
			OffsetPoint(dPoint,	Points[j].point, dVector, 1);

			//,     
			UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, Points[j].point, 1, Points[j].point, &dMinDist1, dPoint, dPoint));
			Edges[i].Sections[j].PointRealBase.Fill(dPoint);
			OffsetPoint(dPoint,	Points[j].point, dVector, -1);
			UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, Points[j].point, 1, Points[j].point, &dMinDist2, dPoint, dPoint));
			if(dMinDist1<dMinDist2)
			{
				dVector[0] *= -1.0;
				dVector[1] *= -1.0;
				dVector[2] *= -1.0;
			}

			Edges[i].Sections[j].PointRealBase.Fill(dPoint);
			OffsetPoint(dPoint,	Points[j].point, dVector, dL);

			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			1,//iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Splines2[i]));

		delete [] Points;
	}

	// 
	for(int i=0;i<iEdgesNum;i++)
	{
		int j = i-1;
		j<0?j+=iEdgesNum:0;
		if(Edges[i].bPS && !Edges[j].bPS)
		{
			bool bFound = false;
			double dParms[2]={1,0};
			int iReverse[2]={0,0};
			tag_t Curves[2];

			//  ,    
			for(int k=i+1;k<iEdgesNum;k++)
				if(!Edges[k].bPS)
				{
					bFound = true;
					break;
				}
			if(!bFound)
				for(k=0;k<i;k++)
					if(!Edges[k].bPS)
					{
						bFound = true;
						break;
					}
			//   j  k
			Curves[0] = Splines2[j];
			Curves[1] = Splines2[k];
			UF_CALL(UF_CURVE_create_bridge_curve(UF_CURVE_TANGENT, Curves, dParms, iReverse, &Splines2[i]));
			i = (k<i?iEdgesNum:k);
		}
	}


	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		int j = i-1, k = i+1;
		j<0?j+=iEdgesNum:0;
		k>iEdgesNum-1?k-=iEdgesNum:0;
		if(!Edges[i].bNothing && Edges[j].bNothing)
			iSplinesNum++;
		if(!Edges[i].bNothing && Edges[k].bNothing)
			iSplinesNum++;
		if(Edges[i].bPS && !Edges[j].bPS)
			iSplinesNum++;
	}

	//  
	{
		int iCounter = 0;
		for(int i=0;i<iHolesNum;i++)
			iCounter += Holes[i].iEdgesNum;
		for(int i=0;i<iPunchesNum;i++)
			iCounter += Punches[i].iEdgesNum;
		UF_MODL_create_string_list(iEdgesNum+iSplinesNum+iCounter, iEdgesNum+iSplinesNum+iCounter, &Section);
		Section.num = 1 + iHolesNum + iPunchesNum;
		Section.string[0] = iEdgesNum+iSplinesNum;
		for(int i=0;i<iEdgesNum;i++)
			if(Edges[i].bPS)
				Section.string[0]--;
		iSplinesNum = 0;
		for(int i=0;i<iEdgesNum;i++)
		{
			int j = i-1, k = i+1;
			j<0?j+=iEdgesNum:0;
			k>iEdgesNum-1?k-=iEdgesNum:0;
			if(!Edges[i].bNothing && Edges[j].bNothing)
			{
				Edges[i].Sections[0].PointRealBase.Fill(LineCoords.start_point);
				//     
				dL = (Edges[i].Sections[0].PointRounding-Edges[i].Sections[0].PointBevel) + (Edges[i].Sections[0].PointUndercutRounding-Edges[i].Sections[0].PointUndercut);

				if(Edges[i].Sections[0].bIsUndercutRoundingPresent)
				{
					P1.X = Edges[i].Sections[0].PointRealBase.X - Edges[i].Sections[0].PointBase.X;
					P1.Y = Edges[i].Sections[0].PointRealBase.Y - Edges[i].Sections[0].PointBase.Y;
					P1.Z = Edges[i].Sections[0].PointRealBase.Z - Edges[i].Sections[0].PointBase.Z;
					P2.X = Edges[i].Sections[0].PointRealBevel.X - Edges[i].Sections[0].PointUndercutRounding.X;
					P2.Y = Edges[i].Sections[0].PointRealBevel.Y - Edges[i].Sections[0].PointUndercutRounding.Y;
					P2.Z = Edges[i].Sections[0].PointRealBevel.Z - Edges[i].Sections[0].PointUndercutRounding.Z;
					if(CalcAngle(P1, P2)<PI/2)
						bInner = true;
					else
						bInner = false;

					//     
					P1.X = Edges[i].Sections[0].PointBevel.X - Edges[i].Sections[0].PointRealBevel.X;
					P1.Y = Edges[i].Sections[0].PointBevel.Y - Edges[i].Sections[0].PointRealBevel.Y;
					P1.Z = Edges[i].Sections[0].PointBevel.Z - Edges[i].Sections[0].PointRealBevel.Z;
					P2.X = Edges[i].Sections[0].PointUndercutRounding.X - Edges[i].Sections[0].PointRealBevel.X;
					P2.Y = Edges[i].Sections[0].PointUndercutRounding.Y - Edges[i].Sections[0].PointRealBevel.Y;
					P2.Z = Edges[i].Sections[0].PointUndercutRounding.Z - Edges[i].Sections[0].PointRealBevel.Z;
					dAngle = PI - CalcAngle(P1, P2);

					bInner ? dL += sqrt(Edges[i].Sections[0].rUndercut*(Edges[i].Sections[0].rUndercut+t)) * dAngle : dL += sqrt(Edges[i].Sections[0].rUndercut*(Edges[i].Sections[0].rUndercut-t)) * dAngle;
				}

				if(Edges[i].Sections[0].bIsRoundingPresent)
				{
					if(Edges[i].Sections[0].Alpha < PI/2)
						dL += sqrt(Edges[i].Sections[0].r*(Edges[i].Sections[0].r+t)) * (PI/2+Edges[i].Sections[0].Alpha);
					else
						dL += sqrt(Edges[i].Sections[0].r*(Edges[i].Sections[0].r-t)) * (PI/2+Edges[i].Sections[0].Alpha);

					//   
					dL -= Edges[i].Sections[0].PointRealBase - Edges[i].Sections[0].PointRounding;
				}

				Edges[i].Sections[0].PlaneSectionNormal.Fill(dPoint);
				FindThirdVector(Service.dBaseNormal, dPoint, dVector);
				Edges[i].Sections[0].PointRealBase.Fill(dPoint);
				OffsetPoint(dPoint,	LineCoords.end_point, dVector, 1);
				//,     
				UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, LineCoords.end_point, 1, LineCoords.end_point, &dMinDist1, dPoint, dPoint));
				Edges[i].Sections[0].PointRealBase.Fill(dPoint);
				OffsetPoint(dPoint,	LineCoords.end_point, dVector, -1);
				UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, LineCoords.end_point, 1, LineCoords.end_point, &dMinDist2, dPoint, dPoint));
				if(dMinDist1<dMinDist2)
				{
					dVector[0] *= -1.0;
					dVector[1] *= -1.0;
					dVector[2] *= -1.0;
				}

				Edges[i].Sections[0].PointRealBase.Fill(dPoint);
				OffsetPoint(dPoint,	LineCoords.end_point, dVector, dL);
				UF_CALL(UF_CURVE_create_line(&LineCoords, &Section.id[iSplinesNum]));
				Section.dir[iSplinesNum] = UF_MODL_CURVE_START_FROM_BEGIN;
				iSplinesNum++;
			}
			if(!Edges[i].bNothing && !Edges[i].bPS)
			{
				Section.id[iSplinesNum] = Splines2[i];
				Section.dir[iSplinesNum] = UF_MODL_CURVE_START_FROM_BEGIN;
				iSplinesNum++;
			}
			if(Edges[i].bPS && !Edges[j].bPS)
			{
				Section.id[iSplinesNum] = Splines2[i];
				Section.dir[iSplinesNum] = UF_MODL_CURVE_START_FROM_BEGIN;
				iSplinesNum++;
			}
			if(Edges[i].bNothing)
			{
				Section.id[iSplinesNum] = Splines[i];
				Section.dir[iSplinesNum] = UF_MODL_CURVE_START_FROM_BEGIN;
				iSplinesNum++;
			}
			if(!Edges[i].bNothing && Edges[k].bNothing)
			{
				Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Fill(LineCoords.end_point);
				//     
				dL = (Edges[i].Sections[Edges[i].iSectionsNum-1].PointRounding-Edges[i].Sections[Edges[i].iSectionsNum-1].PointBevel) + (Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercutRounding-Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercut);

				if(Edges[i].Sections[Edges[i].iSectionsNum-1].bIsUndercutRoundingPresent)
				{
					P1.X = Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.X - Edges[i].Sections[Edges[i].iSectionsNum-1].PointBase.X;
					P1.Y = Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Y - Edges[i].Sections[Edges[i].iSectionsNum-1].PointBase.Y;
					P1.Z = Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Z - Edges[i].Sections[Edges[i].iSectionsNum-1].PointBase.Z;
					P2.X = Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.X - Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercutRounding.X;
					P2.Y = Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.Y - Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercutRounding.Y;
					P2.Z = Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.Z - Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercutRounding.Z;
					if(CalcAngle(P1, P2)<PI/2)
						bInner = true;
					else
						bInner = false;

					//     
					P1.X = Edges[i].Sections[Edges[i].iSectionsNum-1].PointBevel.X - Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.X;
					P1.Y = Edges[i].Sections[Edges[i].iSectionsNum-1].PointBevel.Y - Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.Y;
					P1.Z = Edges[i].Sections[Edges[i].iSectionsNum-1].PointBevel.Z - Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.Z;
					P2.X = Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercutRounding.X - Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.X;
					P2.Y = Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercutRounding.Y - Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.Y;
					P2.Z = Edges[i].Sections[Edges[i].iSectionsNum-1].PointUndercutRounding.Z - Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBevel.Z;
					dAngle = PI - CalcAngle(P1, P2);

					bInner ? dL += sqrt(Edges[i].Sections[Edges[i].iSectionsNum-1].rUndercut*(Edges[i].Sections[Edges[i].iSectionsNum-1].rUndercut+t)) * dAngle : dL += sqrt(Edges[i].Sections[Edges[i].iSectionsNum-1].rUndercut*(Edges[i].Sections[Edges[i].iSectionsNum-1].rUndercut-t)) * dAngle;
				}

				if(Edges[i].Sections[Edges[i].iSectionsNum-1].bIsRoundingPresent)
				{
					if(Edges[i].Sections[Edges[i].iSectionsNum-1].Alpha < PI/2)
						dL += sqrt(Edges[i].Sections[Edges[i].iSectionsNum-1].r*(Edges[i].Sections[Edges[i].iSectionsNum-1].r+t)) * (PI/2+Edges[i].Sections[Edges[i].iSectionsNum-1].Alpha);
					else
						dL += sqrt(Edges[i].Sections[Edges[i].iSectionsNum-1].r*(Edges[i].Sections[Edges[i].iSectionsNum-1].r-t)) * (PI/2+Edges[i].Sections[Edges[i].iSectionsNum-1].Alpha);

					//   
					dL -= Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase - Edges[i].Sections[Edges[i].iSectionsNum-1].PointRounding;
				}
				Edges[i].Sections[Edges[i].iSectionsNum-1].PlaneSectionNormal.Fill(dPoint);
				FindThirdVector(Service.dBaseNormal, dPoint, dVector);
				Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Fill(dPoint);
				OffsetPoint(dPoint,	LineCoords.start_point, dVector, 1);
				//,     
				UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, LineCoords.start_point, 1, LineCoords.start_point, &dMinDist1, dPoint, dPoint));
				Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Fill(dPoint);
				OffsetPoint(dPoint,	LineCoords.start_point, dVector, -1);
				UF_CALL(UF_MODL_ask_minimum_dist(Plane, NULL_TAG, 0, LineCoords.start_point, 1, LineCoords.start_point, &dMinDist2, dPoint, dPoint));
				if(dMinDist1<dMinDist2)
				{
					dVector[0] *= -1.0;
					dVector[1] *= -1.0;
					dVector[2] *= -1.0;
				}

				Edges[i].Sections[Edges[i].iSectionsNum-1].PointRealBase.Fill(dPoint);
				OffsetPoint(dPoint,	LineCoords.start_point, dVector, dL);
				UF_CALL(UF_CURVE_create_line(&LineCoords, &Section.id[iSplinesNum]));
				Section.dir[iSplinesNum] = UF_MODL_CURVE_START_FROM_BEGIN;
				iSplinesNum++;
			}
		}

		for(int h=0;h<iHolesNum;h++)
		{
			Section.string[h+1] = Holes[h].iEdgesNum;
			for(int i=0;i<Holes[h].iEdgesNum;i++)
			{
				Section.dir[iSplinesNum] = UF_MODL_CURVE_START_FROM_BEGIN;

				iDegree = iSplinesDegree;

				//,        
				if(Holes[h].Edges[i].iSectionsNum<iDegree+1)
					iDegree = Holes[h].Edges[i].iSectionsNum-1;

				Points = new UF_CURVE_pt_slope_crvatr_s[Holes[h].Edges[i].iSectionsNum];

				for(int j=0;j<Holes[h].Edges[i].iSectionsNum;j++)
				{
					Holes[h].Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
					Points[j].slope_type = UF_CURVE_SLOPE_NONE;
					Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
				}

				if(Holes[h].iEdgesNum==1)
					iPeriodicity = 1;
				else
					iPeriodicity = 0;

				//    
				UF_CALL(UF_CURVE_create_spline_thru_pts(
					iDegree,
					iPeriodicity,
					Holes[h].Edges[i].iSectionsNum,
					Points,
					NULL,
					0,
					&Section.id[iSplinesNum++]));

				delete [] Points;
			}
		}

		iCounter = 0;
		for(int h=0;h<iPunchesNum;h++)
		{
			Section.string[h+1+iHolesNum] = Punches[h].iEdgesNum;
			for(int i=0;i<Punches[h].iEdgesNum;i++)
			{
				Section.dir[iSplinesNum] = UF_MODL_CURVE_START_FROM_BEGIN;
				Section.id[iSplinesNum++] = PunchSplines[iCounter++];
			}
		}

		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane2));
		for(int i=0;i<iSplinesNum;i++)
			UF_CALL(UF_OBJ_set_blank_status(Section.id[i], UF_OBJ_BLANKED));
		UF_MODL_free_string_list(&Section);
	}

	// Ѩ!!!
	for(int i=0;i<iEdgesNum;i++)
	{
		if(Splines[i] != NULL_TAG)
			UF_CALL(UF_OBJ_set_blank_status(Splines[i], UF_OBJ_BLANKED));
		if(Splines2[i] != NULL_TAG)
			UF_CALL(UF_OBJ_set_blank_status(Splines2[i], UF_OBJ_BLANKED));
	}
	UF_CALL(UF_OBJ_delete_object(Plane));
	delete [] Splines;
	delete [] Splines2;
	delete [] PunchSplines;

	//    
	{
		tag_t Temp;
		CString string;
		char *cOffset1="0", *cOffset2;

		string.Format("%f", t);
		cOffset2 = string.GetBuffer();
		UF_CALL(UF_MODL_create_thicken_sheet(Plane2, cOffset1, cOffset2, UF_NULLSIGN, &Temp));
		UF_CALL(UF_OBJ_set_blank_status(Plane2, UF_OBJ_BLANKED));
		UF_CALL(UF_MODL_ask_feat_body(Temp, &Plane2));
	}

	return Plane2;
}

//####################################################################################################################################################
//####################################################################################################################################################
//####################################################################################################################################################

CEdge::CEdge(void)
{
	dOffset = 0;
	iBridgeType = 2;
	bNothing = false;
	bPS = false;
	bIgnore = false;
}

CEdge::~CEdge(void)
{
}

//      
int CEdge::CreateOuterPlanes(tag_t Base, tag_t Edge, int iIntervalsNum)
{
int iNumber, iNumber2, iIntervalsNumOld = iIntervalsNum;	//    
double dMaxAngle = 1e-5;	//     GetSmoothAdjFace
double dLength, dEvalLimits[2];
double dPoint[3], dPointRounding[3], dTangent[3], dNormal[3], dBiNormal[3], dDistance=10000000;
double dCurvaRange[4], dCurva[5];
int iCurvaStatus;
tag_t PlaneSection, FaceRounding, Face, SheetBody, Feature;
uf_list_p_t EdgeFaces, BodyFeats, FeatFaces;
UF_EVAL_p_t Evaluator;
UF_ATTR_value_t Attr;

	//     
	UF_CALL(UF_MODL_ask_face_body(Base, &SheetBody));
	UF_CALL(UF_MODL_ask_body_feats(SheetBody, &BodyFeats));

	//,   
	UF_CALL(UF_CURVE_ask_parameterization(Edge, dEvalLimits, &iNumber));
	if(iNumber==UF_CURVE_CLOSED_PERIODIC_CURVE || iNumber==UF_CURVE_CLOSED_NON_PERIODIC_CURVE)
		iIntervalsNum--;

	Sections = new CSection[iIntervalsNum];
	iSectionsNum = iIntervalsNum;

	// 
	UF_CALL(UF_CURVE_ask_arc_length(Edge, 0, 1, UF_MODL_MMETER, &dLength));
	UF_CALL(UF_EVAL_initialize(Edge, &Evaluator));
	UF_CALL(UF_EVAL_ask_limits(Evaluator, dEvalLimits));

	//,     
	UF_CALL(UF_MODL_create_list(&EdgeFaces));
	UF_CALL(UF_MODL_ask_edge_faces(Edge, &EdgeFaces));
	UF_CALL(UF_MODL_ask_list_count(EdgeFaces, &iNumber));
	bNothing = !(iNumber-1);

	//,       
	if(iNumber==2)
	{
		int iIndex;
		char *cTitle = "PS";
		tag_t *CF;

		UF_CALL(UF_MODL_delete_list_item(&EdgeFaces, Base));
		UF_CALL(UF_MODL_ask_list_item(EdgeFaces, 0, &Face));

		//,     Face
/*		UF_CALL(UF_MODL_ask_list_count(BodyFeats, &iNumber));
		for(int i=0;i<iNumber;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(BodyFeats, i, &Feature));
			UF_CALL(UF_MODL_ask_feat_faces(Feature, &FeatFaces));
			UF_CALL(UF_MODL_ask_list_count(FeatFaces, &iNumber2));
			if(iNumber2==1)
			{
				UF_CALL(UF_MODL_put_list_item(FeatFaces, Face));
				UF_CALL(UF_MODL_ask_list_count(FeatFaces, &iNumber2));
				if(iNumber2==1)
				{
					UF_CALL(UF_MODL_delete_list(&FeatFaces));
					break;
				}
			}
			UF_CALL(UF_MODL_delete_list(&FeatFaces));
		}*/

		UF_CALL(UF_ATTR_read_value(Face, "PS", UF_ATTR_any, &Attr));
		bPS = Attr.type;

/*		{
			iIndex=0;
			char cTitle[UF_ATTR_MAX_TITLE_LEN+1];
			UF_CALL(UF_ATTR_cycle(Feature, &iIndex, UF_ATTR_null, cTitle, &Attr));
		}*/
	}

	UF_CALL(UF_MODL_delete_list(&EdgeFaces));

	//      
	for(int i=0;i<iIntervalsNum;i++)
	{
		//   
		UF_CALL(UF_EVAL_evaluate_unit_vectors(
			Evaluator,
			dEvalLimits[0] + (double(i)/double(iIntervalsNumOld-1))*(dEvalLimits[1]-dEvalLimits[0]),
			dPoint,
			dTangent,
			dNormal,
			dBiNormal));

		//    
		Sections[i].PointRealBase = dPoint;

		if(bNothing)
		{
			Sections[i].bIsBevelPresent = Sections[i].bIsRoundingPresent = Sections[i].bIsUndercutPresent = Sections[i].bIsUndercutRoundingPresent = false;
			Sections[i].PlaneSection = NULL_TAG;
			continue;
		}
		if(bPS)
		{
			Sections[i].PlaneSection = NULL_TAG;
			continue;
		}

		//       
		UF_CALL(UF_MODL_create_plane(
			dPoint,
			dTangent,
			&Sections[i].PlaneSection));

		//   
		Sections[i].PlaneSectionNormal = dTangent;
	}

	UF_CALL(UF_MODL_delete_list(&BodyFeats));

	return 0;
}

//     :
//PointRealBase
//PointRealBevel
//PointUndercut
int CEdge::ScanRealPoints(tag_t FaceBase, uf_list_p_t FacesBevel, uf_list_p_t FacesUndercut)
{
int iNumber;
double dPoint1[3], dPoint2[3];
uf_list_p_t Face;
CPoint3D P1, P2;
	
	if(bPS || bNothing)
	{
		for(int i=0;i<iSectionsNum;i++)
		{
			Sections[i].PointRealBevel = Sections[i].PointUndercut = Sections[i].PointBase = Sections[i].PointRounding = Sections[i].PointBevel = Sections[i].PointUndercutRounding = Sections[i].PointRealBase;
			Sections[i].bIsRoundingPresent = Sections[i].bIsBevelPresent = Sections[i].bIsUndercutRoundingPresent = Sections[i].bIsUndercutPresent = false;
		}
		return 0;
	}
	else
	{
		UF_CALL(UF_MODL_create_list(&Face));
		UF_CALL(UF_MODL_put_list_item(Face, FaceBase));

		for(int i=0;i<iSectionsNum;i++)
		{
			Sections[i].PointRealBase.Fill(dPoint1);
			if(ScanSectionParms(Sections[i].PlaneSection, FacesBevel, dPoint1, dPoint2, NULL, &Sections[i].H)==1)
				;//MessageBox(NULL, "        ", " ", MB_OK);
			else
				Sections[i].PointRealBevel = dPoint2;
		}
		for(int i=0;i<iSectionsNum;i++)
		{
			Sections[i].PointRealBevel.Fill(dPoint1);
			if(ScanSectionParms(Sections[i].PlaneSection, FacesUndercut, dPoint1, dPoint2)==1)
			{
				Sections[i].PointUndercut = Sections[i].PointUndercutRounding = Sections[i].PointBevel = Sections[i].PointRealBevel;
				Sections[i].bIsUndercutRoundingPresent = Sections[i].bIsUndercutPresent = false;
			}
			else
				Sections[i].PointUndercut = dPoint2;
		}
		//     Base    Alpha
		for(int i=0;i<iSectionsNum;i++)
		{
			double dTemp, dDist;
			Sections[i].PointRealBase.Fill(dPoint1);
			if(ScanSectionParms(Sections[i].PlaneSection, Face, dPoint1, dPoint2, &dTemp, &dDist)==1)
				Sections[i].Alpha = 500;
			else if(dDist > dG_DistTol)
			{
				Sections[i].PointRealBevel.Fill(dPoint1);
				P1.X = dPoint1[0] - Sections[i].PointRealBase.X;
				P1.Y = dPoint1[1] - Sections[i].PointRealBase.Y;
				P1.Z = dPoint1[2] - Sections[i].PointRealBase.Z;
				P2.X = dPoint2[0] - Sections[i].PointRealBase.X;
				P2.Y = dPoint2[1] - Sections[i].PointRealBase.Y;
				P2.Z = dPoint2[2] - Sections[i].PointRealBase.Z;
				Sections[i].Alpha = CalcAngle(P1, P2) - PI/2;
			}
			else
				Sections[i].Alpha = 500;
		}

		UF_CALL(UF_MODL_create_list(&Face));
	}

	return 0;
}

//     :
//PointBase
//PointRounding
//PointBevel
//PointUndercutRounding
//    ScanRealPoints
int CEdge::ScanPoints(uf_list_p_t FacesBevel, uf_list_p_t FacesUndercut, uf_list_p_t FacesRounding, uf_list_p_t FacesUndercutRounding)
{
int iNumber;
double dPoint1[3], dPoint2[3];

	if(bPS || bNothing)
		return 0;
	else
	{
		for(int i=0;i<iSectionsNum;i++)
		if(Sections[i].bIsUndercutPresent)
		{
			Sections[i].PointUndercut.Fill(dPoint1);
			if(ScanSectionParms(Sections[i].PlaneSection, FacesUndercut, dPoint1, dPoint2)==1)
			{
				Sections[i].PointUndercutRounding = Sections[i].PointUndercut;
				Sections[i].bIsUndercutPresent = false;
			}
			else
				Sections[i].PointUndercutRounding = dPoint2;

			Sections[i].PointUndercutRounding.Fill(dPoint1);
			if(ScanSectionParms(Sections[i].PlaneSection, FacesUndercutRounding, dPoint1, dPoint2, &Sections[i].rUndercut)==1)
				;//MessageBox(NULL, "         ", " ", MB_OK);
			else
				Sections[i].PointBevel = dPoint2;
		}

		for(int i=0;i<iSectionsNum;i++)
		if(Sections[i].bIsBevelPresent)
		{
			Sections[i].PointBevel.Fill(dPoint1);
			if(ScanSectionParms(Sections[i].PlaneSection, FacesBevel, dPoint1, dPoint2)==1)
			{
				Sections[i].PointRounding = Sections[i].PointBevel;
				Sections[i].bIsBevelPresent = false;
			}
			else
				Sections[i].PointRounding = dPoint2;

			Sections[i].PointRounding.Fill(dPoint1);
			if(ScanSectionParms(Sections[i].PlaneSection, FacesRounding, dPoint1, dPoint2, &Sections[i].r)==1)
				;//MessageBox(NULL, "        ", " ", MB_OK);
			else
				Sections[i].PointBase = dPoint2;
		}
	}

	return 0;
}

//     
//   :
//{0} -   iSectionsNum   
//{1} -  bPS ,          
//{2} -  bNothing ,       
void CEdge::WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections, int iIndex)
{
CString str;

	str.Format("%d", iSectionsNum);
	sFormatEdge.Replace("{0}", str);

	str.Format("%d", bPS==true);
	sFormatEdge.Replace("{1}", str);

	str.Format("%d", bNothing==true);
	sFormatEdge.Replace("{2}", str);

	File->Write(sFormatEdge, sFormatEdge.GetLength());

	for(int i=0;i<iSectionsNum;i++)
		Sections[i].WriteSection(File, sFormatSections, i + iIndex);
}

//       
//    ,  
void CEdge::PostCalcA(tag_t FaceBase)
{
double dPoint[3], dDist, daTemp[3];
bool bNC1, bNC2;

	// 
	for(int i=1;i<iSectionsNum-1;i++)
		Sections[i].R = CalcR(Sections[i-1].PointRealBase, Sections[i].PointRealBase, Sections[i+1].PointRealBase, &Sections[i].bNullCurva);
	//  
	for(int i=1;i<iSectionsNum-1;i++)
	{
		dPoint[0] = Sections[i-1].PointRealBase.X + (Sections[i+1].PointRealBase.X - Sections[i-1].PointRealBase.X)/2;
		dPoint[1] = Sections[i-1].PointRealBase.Y + (Sections[i+1].PointRealBase.Y - Sections[i-1].PointRealBase.Y)/2;
		dPoint[2] = Sections[i-1].PointRealBase.Z + (Sections[i+1].PointRealBase.Z - Sections[i-1].PointRealBase.Z)/2;
		UF_CALL(UF_MODL_ask_minimum_dist(
			FaceBase,
			NULL_TAG,
			0,
			daTemp,
			1,
			dPoint,
			&dDist,
			daTemp,
			daTemp));
		if(dDist>0)
			Sections[i].R *= -1.0;
	}

	//  
	if(Sections[1].bNullCurva || Sections[2].bNullCurva)
	{
		Sections[0].R = 0;
		Sections[0].bNullCurva = true;
	}
	else
		Sections[0].R = Sections[1].R*Sections[2].R/(2*Sections[2].R-Sections[1].R);
	if(Sections[iSectionsNum-2].bNullCurva || Sections[iSectionsNum-3].bNullCurva)
	{
		Sections[iSectionsNum-1].R = 0;
		Sections[iSectionsNum-1].bNullCurva = true;
	}
	else
		Sections[iSectionsNum-1].R = Sections[iSectionsNum-2].R*Sections[iSectionsNum-3].R/(2*Sections[iSectionsNum-3].R-Sections[iSectionsNum-2].R);

	if(bG_SingleRadius)
	{
		for(int i=0;i<iSectionsNum;i++)
		{
			Sections[i].R = (CalcR(Sections[0].PointRealBase, Sections[iSectionsNum/2].PointRealBase, Sections[iSectionsNum-1].PointRealBase, &bNC1)+CalcR(Sections[0].PointRealBase, Sections[(iSectionsNum-1)/2].PointRealBase, Sections[iSectionsNum-1].PointRealBase, &bNC2))/2;
			Sections[i].bNullCurva = bNC1 || bNC2;
		}
		//  
		dPoint[0] = Sections[0].PointRealBase.X + (Sections[iSectionsNum-1].PointRealBase.X - Sections[0].PointRealBase.X)/2;
		dPoint[1] = Sections[0].PointRealBase.Y + (Sections[iSectionsNum-1].PointRealBase.Y - Sections[0].PointRealBase.Y)/2;
		dPoint[2] = Sections[0].PointRealBase.Z + (Sections[iSectionsNum-1].PointRealBase.Z - Sections[0].PointRealBase.Z)/2;
		UF_CALL(UF_MODL_ask_minimum_dist(
			FaceBase,
			NULL_TAG,
			0,
			daTemp,
			1,
			dPoint,
			&dDist,
			daTemp,
			daTemp));
		if(dDist>0)
			for(int i=0;i<iSectionsNum;i++)
				Sections[i].R *= -1.0;
	}

	if(bG_SingleRadius)
	{
		double dH=0;
		int iH=0;
		for(int i=0;i<iSectionsNum;i++)
			if(Sections[i].H > .1)
			{
				dH += Sections[i].H;
				iH++;
			}
		if(iH!=0)
			for(int i=0;i<iSectionsNum;i++)
				Sections[i].H = dH/double(iH);
	}
}

//       
//    ,  
void CEdge::PostCalcB(void)
{
	//   
	if(Sections[0].r==0)
		Sections[0].r = 2*Sections[1].r - Sections[2].r;
	if(Sections[iSectionsNum-1].r==0)
		Sections[iSectionsNum-1].r = 2*Sections[iSectionsNum-2].r - Sections[iSectionsNum-3].r;

	//   
	if(Sections[0].Alpha==500)
		Sections[0].Alpha = 2*Sections[1].Alpha - Sections[2].Alpha;
	if(Sections[iSectionsNum-1].Alpha==500)
		Sections[iSectionsNum-1].Alpha = 2*Sections[iSectionsNum-2].Alpha - Sections[iSectionsNum-3].Alpha;
}

//        
void CEdge::Reverse(void)
{
char cMemBlock[sizeof(CSection)];

	for(int i=0;i<iSectionsNum/2;i++)
	{
		memcpy(cMemBlock, &Sections[iSectionsNum-i-1], sizeof(CSection));
		memcpy(&Sections[iSectionsNum-i-1], &Sections[i], sizeof(CSection));
		memcpy(&Sections[i], cMemBlock, sizeof(CSection));
	}
}

//      
double CEdge::CalcMeanR(void)
{
double dR=0;

	for(int i=0;i<iSectionsNum;i++)
		dR += Sections[i].r;

	return dR / (double)iSectionsNum;
}

//####################################################################################################################################################
//####################################################################################################################################################
//####################################################################################################################################################

CSection::CSection(void)
{
double dPoint[3]={0,0,0};

	bIsRoundingPresent = bIsBevelPresent = bIsUndercutRoundingPresent = bIsUndercutPresent = true;
	
	PointBase = dPoint;
	PointRealBase = dPoint;
	PointRounding = dPoint;
	PointBevel = dPoint;
	PointRealBevel = dPoint;
	PointUndercutRounding = dPoint;
	PointUndercut = dPoint;

	r = rUndercut = R = H = Alpha = 0;

	bNullCurva = false;
}

//      
//   :
//{0x},{0y},{0z} -   PointRealBase
//{1x},{1y},{1z} -   PointRealBevel
//{2x},{2y},{2z} -   PointUndecut
//{3x},{3y},{3z} -   PointBase
//{4x},{4y},{4z} -   PointRounding
//{c4x},{c4y},{c4z} -   PointRounding  
//{5x},{5y},{5z} -   PointBevel
//{5x},{5y},{5z} -   PointBevel  
//{6x},{6y},{6z} -   PointUndercutRounding
//{7} -   r   
//{7} -   r     
//{8} -   R   
//{9} -   H   
//{9} -   H      
//{10g},{10r}  -  Alpha  , 
//{10g},{10r}  -  Alpha  ,   
//{11} -   rUndercut 
//{12} - 
//{13} -    bIsRoundingPresent
//{14} -    bIsBevelPresent
//{15} -     bIsUndercutRoundingPresent
//{16} -    bIsUndercutPresent
//{i} -   (  )  
void CSection::WriteSection(CFile *File, CString sFormat, int iIndex)
{
CString str;

	str.Format("%f", PointRealBase.X);
	sFormat.Replace("{0x}", str);
	str.Format("%f", PointRealBase.Y);
	sFormat.Replace("{0y}", str);
	str.Format("%f", PointRealBase.Z);
	sFormat.Replace("{0z}", str);

	str.Format("%f", PointRealBevel.X);
	sFormat.Replace("{1x}", str);
	str.Format("%f", PointRealBevel.Y);
	sFormat.Replace("{1y}", str);
	str.Format("%f", PointRealBevel.Z);
	sFormat.Replace("{1z}", str);

	str.Format("%f", PointUndercut.X);
	sFormat.Replace("{2x}", str);
	str.Format("%f", PointUndercut.Y);
	sFormat.Replace("{2y}", str);
	str.Format("%f", PointUndercut.Z);
	sFormat.Replace("{2z}", str);

	str.Format("%f", PointBase.X);
	sFormat.Replace("{3x}", str);
	str.Format("%f", PointBase.Y);
	sFormat.Replace("{3y}", str);
	str.Format("%f", PointBase.Z);
	sFormat.Replace("{3z}", str);

	str.Format("%f", PointRounding.X);
	sFormat.Replace("{4x}", str);
	str.Format("%f", PointRounding.Y);
	sFormat.Replace("{4y}", str);
	str.Format("%f", PointRounding.Z);
	sFormat.Replace("{4z}", str);

	str.Format("%f", CalcPoints.PointRounding.X);
	sFormat.Replace("{c4x}", str);
	str.Format("%f", CalcPoints.PointRounding.Y);
	sFormat.Replace("{c4y}", str);
	str.Format("%f", CalcPoints.PointRounding.Z);
	sFormat.Replace("{c4z}", str);

	str.Format("%f", PointBevel.X);
	sFormat.Replace("{5x}", str);
	str.Format("%f", PointBevel.Y);
	sFormat.Replace("{5y}", str);
	str.Format("%f", PointBevel.Z);
	sFormat.Replace("{5z}", str);

	str.Format("%f", CalcPoints.PointBevel.X);
	sFormat.Replace("{c5x}", str);
	str.Format("%f", CalcPoints.PointBevel.Y);
	sFormat.Replace("{c5y}", str);
	str.Format("%f", CalcPoints.PointBevel.Z);
	sFormat.Replace("{c5z}", str);

	str.Format("%f", PointUndercutRounding.X);
	sFormat.Replace("{6x}", str);
	str.Format("%f", PointUndercutRounding.Y);
	sFormat.Replace("{6y}", str);
	str.Format("%f", PointUndercutRounding.Z);
	sFormat.Replace("{6z}", str);

	str.Format("%f", r);
	sFormat.Replace("{7}", str);

	str.Format("%f", CalcPoints.r);
	sFormat.Replace("{c7}", str);

	if(!bNullCurva)
	{
		str.Format("%f", R);
		sFormat.Replace("{8}", str);
	}
	else
		sFormat.Replace("{8}", "#Inf");

	str.Format("%f", H);
	sFormat.Replace("{9}", str);

	str.Format("%f", CalcPoints.H);
	sFormat.Replace("{c9}", str);

	str.Format("%f", Alpha);
	sFormat.Replace("{10r}", str);
	str.Format("%f", Alpha*180/PI);
	sFormat.Replace("{10g}", str);

	str.Format("%f", CalcPoints.Alpha);
	sFormat.Replace("{c10r}", str);
	str.Format("%f", CalcPoints.Alpha*180/PI);
	sFormat.Replace("{c10g}", str);

	str.Format("%f", rUndercut);
	sFormat.Replace("{11}", str);

	str.Format("%d", bIsRoundingPresent==true);
	sFormat.Replace("{13}", str);

	str.Format("%d", bIsBevelPresent==true);
	sFormat.Replace("{14}", str);

	str.Format("%d", bIsUndercutRoundingPresent==true);
	sFormat.Replace("{15}", str);

	str.Format("%d", bIsUndercutPresent==true);
	sFormat.Replace("{16}", str);

	str.Format("%d", iIndex);
	sFormat.Replace("{i}", str);

	File->Write(sFormat, sFormat.GetLength());
}

//       
void CSection::PostCalc(double t, double E, double K, double n, double *dNormal, tag_t BaseFace)
{
CPoint3D P1, P2, P3, N, V1, V2;
double dAngle, dL;
bool bInner;

	//  
	//     
	if(bIsUndercutRoundingPresent)
	{
		//,   ,   
		P1.X = PointRealBase.X - PointBase.X;
		P1.Y = PointRealBase.Y - PointBase.Y;
		P1.Z = PointRealBase.Z - PointBase.Z;
		P2.X = PointRealBevel.X - PointUndercutRounding.X;
		P2.Y = PointRealBevel.Y - PointUndercutRounding.Y;
		P2.Z = PointRealBevel.Z - PointUndercutRounding.Z;
		if(CalcAngle(P1, P2)<PI/2)
			bInner = true;
		else
			bInner = false;

		//     
		P1.X = PointBevel.X - PointRealBevel.X;
		P1.Y = PointBevel.Y - PointRealBevel.Y;
		P1.Z = PointBevel.Z - PointRealBevel.Z;
		P2.X = PointUndercutRounding.X - PointRealBevel.X;
		P2.Y = PointUndercutRounding.Y - PointRealBevel.Y;
		P2.Z = PointUndercutRounding.Z - PointRealBevel.Z;
		dAngle = PI - CalcAngle(P1, P2);

		//      
		if(bInner)
			dL = sqrt(rUndercut*(rUndercut+t)) * dAngle;
		else
			dL = sqrt(rUndercut*(rUndercut-t)) * dAngle;
		
		//   
		CalcPoints.H = (PointRealBase-PointBevel) + (PointUndercutRounding-PointUndercut) + dL;
	}
	else
		CalcPoints.H = H;

	//  
	CalcPoints.Alpha = Alpha - Gamma(K, n, t, CalcPoints.H, R, r, Alpha, E);

	//      
	//P1.X = PointBase.X - PointRealBase.X;
	//P1.Y = PointBase.Y - PointRealBase.Y;
	//P1.Z = PointBase.Z - PointRealBase.Z;
	P1.X = dNormal[0];
	P1.Y = dNormal[1];
	P1.Z = dNormal[2];
	P2.X = PointRounding.X - PointRealBase.X;
	P2.Y = PointRounding.Y - PointRealBase.Y;
	P2.Z = PointRounding.Z - PointRealBase.Z;
	//N.X = P2.Z*P1.Y - P2.Y*P1.Z;
	//N.Y = P1.Z*P2.X - P2.Z*P1.X;
	//N.Z = P1.X*P2.Y - P2.X*P1.Y;

	N = PlaneSectionNormal;
	
	//   
/*	{
		double dFrom[3], dTo[3];

		PointRealBase.Fill(dFrom);
		dTo[0] = dFrom[0] + PlaneSectionNormal.X*10 - dBasePoint[0];
		dTo[1] = dFrom[1] + PlaneSectionNormal.Y*10 - dBasePoint[1];
		dTo[2] = dFrom[2] + PlaneSectionNormal.Z*10 - dBasePoint[2];
		dFrom[0] -= dBasePoint[0];
		dFrom[1] -= dBasePoint[1];
		dFrom[2] -= dBasePoint[2];

		UF_VEC3_angle_between(dFrom, dTo, dNormal, &dAngle);

		if(dAngle<PI/2)
			Rotate(&P2, N,  Alpha - CalcPoints.Alpha);
		else
			Rotate(&P2, N,  Alpha - CalcPoints.Alpha);
	}*/

	V1 = V2 = P2;
	Rotate(&V1, N,  Alpha - CalcPoints.Alpha);
	Rotate(&V2, N,  CalcPoints.Alpha - Alpha);
	if(Alpha < -.01)
		P2 = CalcAngle(V1, P1) < CalcAngle(V2, P1) ? V1 : V2;
	else if(Alpha > .01)
		P2 = CalcAngle(V1, P1) < CalcAngle(V2, P1) ? V2 : V1;
	else
	{
		double dPPP[3], dDist1, dDist2;
		CPoint3D Po1, Po2;
		V1 = V2 = P2;
		Rotate(&V1, N,  Alpha - CalcPoints.Alpha);
		Rotate(&V2, N,  CalcPoints.Alpha - Alpha);
		Po1 = Po2 = PointRealBase;
		dL = sqrt(V1.X*V1.X + V1.Y*V1.Y + V1.Z*V1.Z);
		Po1.X += V1.X*100.0/dL;
		Po1.Y += V1.Y*100.0/dL;
		Po1.Z += V1.Z*100.0/dL;
		dL = sqrt(V2.X*V2.X + V2.Y*V2.Y + V2.Z*V2.Z);
		Po2.X += V2.X*100.0/dL;
		Po2.Y += V2.Y*100.0/dL;
		Po2.Z += V2.Z*100.0/dL;
		
		Po1.Fill(dPPP);
		UF_CALL(UF_MODL_ask_minimum_dist(BaseFace, NULL_TAG, 0, dPPP, 1, dPPP, &dDist1, dPPP, dPPP));
		Po2.Fill(dPPP);
		UF_CALL(UF_MODL_ask_minimum_dist(BaseFace, NULL_TAG, 0, dPPP, 1, dPPP, &dDist2, dPPP, dPPP));
		if(dDist1>dDist2)
			P2 = V2;
		else
			P2 = V1;
	}

	//     
//	Rotate(&P2, N,  Alpha - CalcPoints.Alpha);
//	Rotate(&P2, N,  CalcPoints.Alpha - Alpha);
	//
	dL = sqrt(P2.X*P2.X + P2.Y*P2.Y + P2.Z*P2.Z);
	P2.X /= dL;
	P2.Y /= dL;
	P2.Z /= dL;

	//      
	CalcPoints.PointBevel.X = PointRealBase.X + P2.X*CalcPoints.H;
	CalcPoints.PointBevel.Y = PointRealBase.Y + P2.Y*CalcPoints.H;
	CalcPoints.PointBevel.Z = PointRealBase.Z + P2.Z*CalcPoints.H;
	CalcPoints.PointRounding.X = PointRealBase.X + P2.X*(PointRounding-PointRealBase);
	CalcPoints.PointRounding.Y = PointRealBase.Y + P2.Y*(PointRounding-PointRealBase);
	CalcPoints.PointRounding.Z = PointRealBase.Z + P2.Z*(PointRounding-PointRealBase);

	//   
	CalcPoints.r = (PointRounding-PointRealBase)*tan((Alpha+PI/2)/2.0);
	CalcPoints.r = (PointRounding-PointRealBase)*tan((CalcPoints.Alpha+PI/2)/2.0);

//	if(bIsRoundingPresent
}


//####################################################################################################################################################
//####################################################################################################################################################
//####################################################################################################################################################

int CHole::ScanHole(uf_list_p_t InnerEdges, int iIntervalsNum)
{
tag_t Curve;
double dLength, dEdgesLength=0, dSegmentSize;
int iNumOfPoints;

	UF_CALL(UF_MODL_ask_list_count(InnerEdges, &iEdgesNum));

	//  
	Edges = new CHoleEdge[iEdgesNum];

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(InnerEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));
		dEdgesLength += dLength;
	}

	//       
	dSegmentSize = dEdgesLength/double(iIntervalsNum);
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(InnerEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));

		iNumOfPoints = dLength/dSegmentSize + 1.5;
		if(iNumOfPoints<4)
			iNumOfPoints=4;

		Edges[i].ScanEdge(Curve, iNumOfPoints);
	}

	return 0;
}

//     
//   :
//{0} -   iEdgesNum   
void CHole::WriteHole(CFile *File, CString sFormat, CString sFormatEdge, CString sFormatSections)
{
CString str;

	str.Format("%d", iEdgesNum);
	sFormat.Replace("{0}", str);

	File->Write(sFormat, sFormat.GetLength());

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].WriteEdge(File, sFormatEdge, sFormatSections);
}

//     
void CHole::WriteParameters(CFile *File)
{
CString str;

	str.Format("#TypeOfObject = Hole\r\n#NumOfEdges = %d\r\n", iEdgesNum);
	File->Write(str, str.GetLength());

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].Parameters.WriteParameters(File);
}

int CHoleEdge::ScanEdge(tag_t Edge, int iIntervalsNum)
{
int iNumber, iIntervalsNumOld = iIntervalsNum;	//    
double dLength, dEvalLimits[2];
double dPoint[3], dTangent[3], dNormal[3], dBiNormal[3];
UF_EVAL_p_t Evaluator;

	//,   
	UF_CALL(UF_CURVE_ask_parameterization(Edge, dEvalLimits, &iNumber));
	if(iNumber==UF_CURVE_CLOSED_PERIODIC_CURVE || iNumber==UF_CURVE_CLOSED_NON_PERIODIC_CURVE)
		iIntervalsNum--;

	Sections = new CHoleSection[iIntervalsNum];
	iSectionsNum = iIntervalsNum;

	// 
	UF_CALL(UF_CURVE_ask_arc_length(Edge, 0, 1, UF_MODL_MMETER, &dLength));
	UF_CALL(UF_EVAL_initialize(Edge, &Evaluator));
	UF_CALL(UF_EVAL_ask_limits(Evaluator, dEvalLimits));


	//      
	for(int i=0;i<iIntervalsNum;i++)
	{
		//   
		UF_CALL(UF_EVAL_evaluate_unit_vectors(
			Evaluator,
			dEvalLimits[0] + (double(i)/double(iIntervalsNumOld-1))*(dEvalLimits[1]-dEvalLimits[0]),
			dPoint,
			dTangent,
			dNormal,
			dBiNormal));

		//    
		Sections[i].PointRealBase = dPoint;
	}

	//  
	Parameters.ScanEdge(Edge);

	return 0;
}

//      
//   :
//{0} -   iSectionsNum   
void CHoleEdge::WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections)
{
CString str;

	str.Format("%d", iSectionsNum);
	sFormatEdge.Replace("{0}", str);

	File->Write(sFormatEdge, sFormatEdge.GetLength());

	for(int i=0;i<iSectionsNum;i++)
		Sections[i].WriteSection(File, sFormatSections);
}


//        
void CHoleEdge::Reverse(void)
{
char cMemBlock[sizeof(CHoleSection)];

	for(int i=0;i<iSectionsNum/2;i++)
	{
		memcpy(cMemBlock, &Sections[iSectionsNum-i-1], sizeof(CHoleSection));
		memcpy(&Sections[iSectionsNum-i-1], &Sections[i], sizeof(CHoleSection));
		memcpy(&Sections[i], cMemBlock, sizeof(CHoleSection));
	}
}

//      
//   :
//{0x},{0y},{0z} -   PointRealBase
void CHoleSection::WriteSection(CFile *File, CString sFormatSections)
{
CString str;

	str.Format("%f", PointRealBase.X);
	sFormatSections.Replace("{0x}", str);
	str.Format("%f", PointRealBase.Y);
	sFormatSections.Replace("{0y}", str);
	str.Format("%f", PointRealBase.Z);
	sFormatSections.Replace("{0z}", str);

	File->Write(sFormatSections, sFormatSections.GetLength());
}

//####################################################################################################################################################
//####################################################################################################################################################
//####################################################################################################################################################

int CPunch::CreateInnerPlanes(uf_list_p_t InnerEdges, int iIntervalsNum)
{
tag_t Curve;
double dLength, dEdgesLength=0, dSegmentSize;
int iNumOfPoints;

	UF_CALL(UF_MODL_ask_list_count(InnerEdges, &iEdgesNum));

	//  
	Edges = new CPunchEdge[iEdgesNum];

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(InnerEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));
		dEdgesLength += dLength;
	}

	//       
	dSegmentSize = dEdgesLength/double(iIntervalsNum);
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(InnerEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));

		iNumOfPoints = dLength/dSegmentSize + 1.5;
		if(iNumOfPoints<4)
			iNumOfPoints=4;

		Edges[i].CreateInnerPlanes(Curve, iNumOfPoints);
	}

	return 0;
}

//     
//   :
//{0} -   iEdgesNum   
void CPunch::WritePunch(CFile *File, CString sFormat, CString sFormatEdge, CString sFormatSections)
{
CString str;

	str.Format("%d", iEdgesNum);
	sFormat.Replace("{0}", str);

	File->Write(sFormat, sFormat.GetLength());

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].WriteEdge(File, sFormatEdge, sFormatSections);
}

//     
void CPunch::WriteParameters(CFile *File)
{
CString str;

	str.Format("#TypeOfObject = Punch\r\n#NumOfEdges = %d\r\n", iEdgesNum);
	File->Write(str, str.GetLength());

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].Parameters.WriteParameters(File);
}

//  ""     
void CPunch::CreatePunchBase(tag_t *Base, double dBaseHeight, double dBaseNormal[3])
{
tag_t *SplinesTop = new tag_t[iEdgesNum], *SplinesBottom = new tag_t[iEdgesNum], Cylinder;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity;
double dPointOrigin[3], dVector[3], dAlpha;
CString string;

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesTop[i]));

		delete [] Points;
	}

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointBevel.Fill(Points[j].point);
			Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
			dVector[0] = Points[j].point[0] - dPointOrigin[0];
			dVector[1] = Points[j].point[1] - dPointOrigin[1];
			dVector[2] = Points[j].point[2] - dPointOrigin[2];
			dAlpha = CalcAngle(dVector, dBaseNormal);
			OffsetPoint(dPointOrigin, Points[j].point, dVector, -(dBaseHeight+dG_BaseAdditionalThickness) / cos(dAlpha));
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesBottom[i]));

		delete [] Points;
	}
	//  ,  
	if(iEdgesNum>1)
	{
		for(int i=0;i<iEdgesNum-1;i++)
			ConnectCurves(SplinesBottom[i], SplinesBottom[i+1]);
		ConnectCurves(SplinesBottom[iEdgesNum-1], SplinesBottom[0]);
	}

	//  
	{
		tag_t Plane, *SewTools;//, Curves[2];
		UF_STRING_t Section, Spine;
		int iAlignment=1, iEndPoint=0, iBodyType=0;//, iNumOfRuleds=0, iNumOfPS, iReverse[2]={0,0};
		double dValue[6], dTol[3]={.01, .01, 1};//, dParms[2]={1,0};
		uf_list_p_t ListDisjoint;

		//   
		SewTools = new tag_t[iEdgesNum+1];

		UF_MODL_create_string_list(2, 2, &Section);
		Section.num = 2;
		Section.string[0] = 1;
		Section.dir[0] = UF_MODL_CURVE_START_FROM_BEGIN;
		Section.string[1] = 1;
		Section.dir[1] = UF_MODL_CURVE_START_FROM_BEGIN;
		Spine.num = 0;

		//   
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.id[0] = SplinesTop[i];
			Section.id[1] = SplinesBottom[i];

			UF_CALL(UF_MODL_create_ruled(&Section, &Spine, &iAlignment, dValue, &iEndPoint, &iBodyType, UF_NULLSIGN, dTol, &SewTools[i]));
		}

		UF_MODL_free_string_list(&Section);

		//  
		UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
		Section.num = 1;
		Section.string[0] = iEdgesNum;
		// 
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
			Section.id[i] = SplinesTop[i];
		}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &SewTools[iEdgesNum]));
		// 
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
			Section.id[i] = SplinesBottom[i];
		}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane));
		UF_MODL_free_string_list(&Section);

		//   
		UF_CALL(UF_MODL_create_sew(0, 1, &Plane, iEdgesNum+1, SewTools, .1, 1, &ListDisjoint, &Cylinder));
		{
			tag_t Temp;
			UF_CALL(UF_MODL_ask_feat_body(Cylinder, &Temp));
			Cylinder = Temp;
		}

		//    ""
		UF_CALL(UF_MODL_operations(*Base, Cylinder, UF_NEGATIVE));
	}

	//  (       )
	{
		int iBaseEdges;
		tag_t Blend, Edge;
		UF_MODL_edge_blend_data_s BlendData;
		UF_MODL_blend_edge_s *BlendEdge;
		UF_MODL_blend_point_s *BlendPoints;
		uf_list_p_t BaseEdges;

		BlendData.blend_type = 0;
		BlendData.blend_instanced = FALSE;
		BlendData.blend_setback = FALSE;
		BlendData.vrb_tolerance = .01;
		BlendData.blend_radius[0] = '0';
		BlendData.blend_radius[1] = '\0';
		BlendData.smooth_overflow = TRUE;
		BlendData.cliff_overflow = TRUE;
		BlendData.notch_overflow = TRUE;
		BlendData.number_edges = iEdgesNum;

		BlendEdge = new UF_MODL_blend_edge_s[iEdgesNum];

		//   
		UF_CALL(UF_MODL_ask_body_edges(*Base, &BaseEdges));
		UF_CALL(UF_MODL_ask_list_count(BaseEdges, &iBaseEdges));

		for(int i=0;i<iEdgesNum;i++)
		{
			UF_EVAL_p_t Eval;
			double dGuess[3], dMinDist, dParm, dParLimits[2];

			//  
			for(int j=0;j<iBaseEdges;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(BaseEdges, j, &BlendEdge[i].edge));;
				if(CurvesCoincide(SplinesTop[i], BlendEdge[i].edge))
					break;
			}

			BlendEdge[i].number_points = Edges[i].iSectionsNum;
			//BlendEdge[i].number_points = 0;
			BlendEdge[i].point_data = new UF_MODL_blend_point_s[Edges[i].iSectionsNum];

			UF_CALL(UF_EVAL_initialize(BlendEdge[i].edge, &Eval));
			UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].PointRealBase.Fill(dGuess);
				UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
				BlendEdge[i].point_data[j].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
				//    
				if(iEdgesNum==1 && j==0 && BlendEdge[i].point_data[j].parameter==dParLimits[1])
					BlendEdge[i].point_data[j].parameter = dParLimits[0];
				if(iEdgesNum==1 && j==Edges[i].iSectionsNum-1 && BlendEdge[i].point_data[j].parameter==dParLimits[0])
					BlendEdge[i].point_data[j].parameter = dParLimits[1];
				string.Format("%f", Edges[i].Sections[j].r);
				string.CopyChars(BlendEdge[i].point_data[j].radius, string.GetBuffer(), string.GetLength());
				BlendEdge[i].point_data[j].radius[string.GetLength()] = '\0';
			}

			UF_CALL(UF_EVAL_free(Eval));
		}

		BlendData.edge_data = BlendEdge;
		UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

		// 
		for(int i=0;i<iEdgesNum;i++)
			delete [] BlendEdge[i].point_data;
		delete [] BlendEdge;
	}

	//  
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_OBJ_set_blank_status(SplinesTop[i], UF_OBJ_BLANKED));
		UF_CALL(UF_OBJ_set_blank_status(SplinesBottom[i], UF_OBJ_BLANKED));
	}

	delete [] SplinesTop;
	delete [] SplinesBottom;
}

//  ""     
void CPunch::CreatePunchPress(tag_t *Press, double dBaseNormal[3])
{
tag_t Cylinder;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity;
double dPointOrigin[3], dTol[3]={.01, .01, 1};
UF_STRING_t Section;

	UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
	Section.num = 1;
	Section.string[0] = iEdgesNum;

	//  
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointBase.Fill(dPointOrigin);
			OffsetPoint(dPointOrigin, Points[j].point, dBaseNormal, dG_BasePressDistance + dG_PressThickness);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Section.id[i]));

		Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;

		delete [] Points;
	}

	//  
	{
		int iL;
		UF_CALL(UF_LAYER_ask_work_layer(&iL));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Cylinder));
		UF_CALL(UF_LAYER_set_status(iL, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_INACTIVE_LAYER));
	}
//	UF_CALL(UF_OBJ_set_translucency(Cylinder, 100));
	UF_CALL(UF_OBJ_set_blank_status(Cylinder, UF_OBJ_BLANKED));

	//    Extrude
	{
		uf_list_p_t ExtrObjects, ExtrFeats;
		char *cTaper="0", *cLimit[2]={"0", "0"};
		CString string;
		double dCylAxis[3]={-dBaseNormal[0], -dBaseNormal[1], -dBaseNormal[2]};
	
		UF_CALL(UF_MODL_create_list(&ExtrObjects));
		UF_CALL(UF_MODL_put_list_item(ExtrObjects, Cylinder));

		string.Format("%f", dG_PressThickness);
		cLimit[1] = string.GetBuffer();

		UF_CALL(UF_MODL_create_extruded(ExtrObjects, cTaper, cLimit, dPointOrigin, dCylAxis, UF_NULLSIGN, &ExtrFeats));

		UF_CALL(UF_MODL_ask_list_item(ExtrFeats, 0, &Cylinder));

		UF_CALL(UF_MODL_delete_list(&ExtrObjects));
		UF_CALL(UF_MODL_delete_list(&ExtrFeats));

		//    ""
		{
			tag_t Temp;
			UF_CALL(UF_MODL_ask_feat_body(Cylinder, &Temp));
			Cylinder = Temp;
		}
		UF_CALL(UF_MODL_operations(*Press, Cylinder, UF_NEGATIVE));
	}

	//   5
	{
		int iPressEdges;
		tag_t Blend, Edge;
		UF_MODL_edge_blend_data_s BlendData;
		UF_MODL_blend_edge_s *BlendEdge;
		UF_MODL_blend_point_s *BlendPoints;
		uf_list_p_t PressEdges;

		BlendData.blend_type = 0;
		BlendData.blend_instanced = FALSE;
		BlendData.blend_setback = FALSE;
		BlendData.vrb_tolerance = .01;
		BlendData.blend_radius[0] = '5';
		BlendData.blend_radius[1] = '\0';
		BlendData.smooth_overflow = TRUE;
		BlendData.cliff_overflow = TRUE;
		BlendData.notch_overflow = TRUE;
		BlendData.number_edges = iEdgesNum;

		BlendEdge = new UF_MODL_blend_edge_s[iEdgesNum];

		//   
		UF_CALL(UF_MODL_ask_body_edges(*Press, &PressEdges));
		UF_CALL(UF_MODL_ask_list_count(PressEdges, &iPressEdges));

		for(int i=0;i<iEdgesNum;i++)
		{
			//  
			for(int j=0;j<iPressEdges;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(PressEdges, j, &BlendEdge[i].edge));;
				if(CurvesCoincide(Section.id[i], BlendEdge[i].edge, .1))
					break;
			}

			BlendEdge[i].number_points = 0;
		}

		BlendData.edge_data = BlendEdge;
		UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

		// 
		delete [] BlendEdge;
	}

	//  
	for(int i=0;i<iEdgesNum;i++)
		UF_CALL(UF_OBJ_set_blank_status(Section.id[i], UF_OBJ_BLANKED));
	UF_MODL_free_string_list(&Section);
}

int CPunchEdge::CreateInnerPlanes(tag_t Edge, int iIntervalsNum)
{
int iNumber, iIntervalsNumOld = iIntervalsNum;	//    
double dLength, dEvalLimits[2];
double dPoint[3], dTangent[3], dNormal[3], dBiNormal[3];
UF_EVAL_p_t Evaluator;

	//,   
	UF_CALL(UF_CURVE_ask_parameterization(Edge, dEvalLimits, &iNumber));
	if(iNumber==UF_CURVE_CLOSED_PERIODIC_CURVE || iNumber==UF_CURVE_CLOSED_NON_PERIODIC_CURVE)
		iIntervalsNum--;

	Sections = new CPunchSection[iIntervalsNum];
	iSectionsNum = iIntervalsNum;

	// 
	UF_CALL(UF_CURVE_ask_arc_length(Edge, 0, 1, UF_MODL_MMETER, &dLength));
	UF_CALL(UF_EVAL_initialize(Edge, &Evaluator));
	UF_CALL(UF_EVAL_ask_limits(Evaluator, dEvalLimits));


	//      
	for(int i=0;i<iIntervalsNum;i++)
	{
		//   
		UF_CALL(UF_EVAL_evaluate_unit_vectors(
			Evaluator,
			dEvalLimits[0] + (double(i)/double(iIntervalsNumOld-1))*(dEvalLimits[1]-dEvalLimits[0]),
			dPoint,
			dTangent,
			dNormal,
			dBiNormal));

		//    
		Sections[i].PointRealBase = dPoint;

		//       
		UF_CALL(UF_MODL_create_plane(
			dPoint,
			dTangent,
			&Sections[i].PlaneSection));
	}

	//  
	Parameters.ScanEdge(Edge);

	return 0;
}

//    :
//PointBevel
int CPunchEdge::ScanRealPoints(uf_list_p_t FacesPunchBevel)
{
int iNumber;
double dPoint1[3], dPoint2[3];

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointRealBase.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPunchBevel, dPoint1, dPoint2)==1)
			MessageBox(NULL, "         ", " ", MB_OK);
		else
			Sections[i].PointBevel = dPoint2;
	}

	return 0;
}

//    :
//PointRounding
//PointBase
int CPunchEdge::ScanPoints(uf_list_p_t FacesPunchBevel, uf_list_p_t FacesPunchRounding)
{
int iNumber;
double dPoint1[3], dPoint2[3];

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointBevel.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPunchBevel, dPoint1, dPoint2)==1)
			MessageBox(NULL, "         ", " ", MB_OK);
		else
			Sections[i].PointRounding = dPoint2;
	}

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointRounding.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPunchRounding, dPoint1, dPoint2, &Sections[i].r)==1)
			MessageBox(NULL, "         ", " ", MB_OK);
		else
			Sections[i].PointBase = dPoint2;
	}

	return 0;
}

//      
//   :
//{0} -   iSectionsNum   
void CPunchEdge::WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections)
{
CString str;

	str.Format("%d", iSectionsNum);
	sFormatEdge.Replace("{0}", str);

	File->Write(sFormatEdge, sFormatEdge.GetLength());

	for(int i=0;i<iSectionsNum;i++)
		Sections[i].WriteSection(File, sFormatSections);
}

//        
void CPunchEdge::Reverse(void)
{
char cMemBlock[sizeof(CPunchSection)];

	for(int i=0;i<iSectionsNum/2;i++)
	{
		memcpy(cMemBlock, &Sections[iSectionsNum-i-1], sizeof(CPunchSection));
		memcpy(&Sections[iSectionsNum-i-1], &Sections[i], sizeof(CPunchSection));
		memcpy(&Sections[i], cMemBlock, sizeof(CPunchSection));
	}
}

//      
//   :
//{0x},{0y},{0z} -   PointRealBase
//{1x},{1y},{1z} -   PointBevel
//{2x},{2y},{2z} -   PointBase
//{3x},{3y},{3z} -   PointRounding
void CPunchSection::WriteSection(CFile *File, CString sFormatSections)
{
CString str;

	str.Format("%f", PointRealBase.X);
	sFormatSections.Replace("{0x}", str);
	str.Format("%f", PointRealBase.Y);
	sFormatSections.Replace("{0y}", str);
	str.Format("%f", PointRealBase.Z);
	sFormatSections.Replace("{0z}", str);

	str.Format("%f", PointBevel.X);
	sFormatSections.Replace("{1x}", str);
	str.Format("%f", PointBevel.Y);
	sFormatSections.Replace("{1y}", str);
	str.Format("%f", PointBevel.Z);
	sFormatSections.Replace("{1z}", str);

	str.Format("%f", PointBevel.X);
	sFormatSections.Replace("{2x}", str);
	str.Format("%f", PointBevel.Y);
	sFormatSections.Replace("{2y}", str);
	str.Format("%f", PointBevel.Z);
	sFormatSections.Replace("{2z}", str);

	str.Format("%f", PointBevel.X);
	sFormatSections.Replace("{3x}", str);
	str.Format("%f", PointBevel.Y);
	sFormatSections.Replace("{3y}", str);
	str.Format("%f", PointBevel.Z);
	sFormatSections.Replace("{3z}", str);

	File->Write(sFormatSections, sFormatSections.GetLength());
}

//####################################################################################################################################################
//####################################################################################################################################################
//####################################################################################################################################################

int CPressIn::CreateInnerPlanes(uf_list_p_t InnerEdges, int iIntervalsNum)
{
tag_t Curve;
double dLength, dEdgesLength=0, dSegmentSize;
int iNumOfPoints;

	UF_CALL(UF_MODL_ask_list_count(InnerEdges, &iEdgesNum));

	//  
	Edges = new CPressInEdge[iEdgesNum];

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(InnerEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));
		dEdgesLength += dLength;
	}

	//       
	dSegmentSize = dEdgesLength/double(iIntervalsNum);
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_MODL_ask_list_item(InnerEdges, i, &Curve));
		UF_CALL(UF_CURVE_ask_arc_length(Curve, 0, 1, UF_MODL_MMETER, &dLength));

		iNumOfPoints = dLength/dSegmentSize + 1.5;
		if(iNumOfPoints==1)
			iNumOfPoints++;

		Edges[i].CreateInnerPlanes(Curve, iNumOfPoints);
	}

	return 0;
}

//     
//   :
//{0} -   iEdgesNum   
void CPressIn::WritePressIn(CFile *File, CString sFormat, CString sFormatEdge, CString sFormatSections)
{
CString str;

	str.Format("%d", iEdgesNum);
	sFormat.Replace("{0}", str);

	File->Write(sFormat, sFormat.GetLength());

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].WriteEdge(File, sFormatEdge, sFormatSections);
}

//     
void CPressIn::WriteParameters(CFile *File)
{
CString str;

	str.Format("#TypeOfObject = PressIn\r\n#NumOfEdges = %d\r\n", iEdgesNum);
	File->Write(str, str.GetLength());

	for(int i=0;i<iEdgesNum;i++)
		Edges[i].Parameters.WriteParameters(File);
}

//  ""     
void CPressIn::CreatePressInBase(tag_t *Base, double dBaseNormal[3])
{
tag_t *SplinesTop = new tag_t[iEdgesNum], *SplinesBottom = new tag_t[iEdgesNum], Cylinder;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity;
double dPointOrigin[3], dVector[3], dAlpha;
CString string;

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesTop[i]));

		delete [] Points;
	}

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBottom.Fill(Points[j].point);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesBottom[i]));

		delete [] Points;
	}

	//  
	{
		tag_t Plane, *SewTools;//, Curves[2];
		UF_STRING_t Section, Spine;
		int iAlignment=1, iEndPoint=0, iBodyType=0;//, iNumOfRuleds=0, iNumOfPS, iReverse[2]={0,0};
		double dValue[6], dTol[3]={.01, .01, 1};//, dParms[2]={1,0};
		uf_list_p_t ListDisjoint;

		//   
		SewTools = new tag_t[iEdgesNum+1];

		UF_MODL_create_string_list(2, 2, &Section);
		Section.num = 2;
		Section.string[0] = 1;
		Section.dir[0] = UF_MODL_CURVE_START_FROM_BEGIN;
		Section.string[1] = 1;
		Section.dir[1] = UF_MODL_CURVE_START_FROM_BEGIN;
		Spine.num = 0;

		//   
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.id[0] = SplinesTop[i];
			Section.id[1] = SplinesBottom[i];

			UF_CALL(UF_MODL_create_ruled(&Section, &Spine, &iAlignment, dValue, &iEndPoint, &iBodyType, UF_NULLSIGN, dTol, &SewTools[i]));
		}

		UF_MODL_free_string_list(&Section);

		//  
		UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
		Section.num = 1;
		Section.string[0] = iEdgesNum;
		// 
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
			Section.id[i] = SplinesTop[i];
		}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &SewTools[iEdgesNum]));
		// 
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
			Section.id[i] = SplinesBottom[i];
		}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane));
		UF_MODL_free_string_list(&Section);

		//   
		UF_CALL(UF_MODL_create_sew(0, 1, &Plane, iEdgesNum+1, SewTools, .1, 1, &ListDisjoint, &Cylinder));
		{
			tag_t Temp;
			UF_CALL(UF_MODL_ask_feat_body(Cylinder, &Temp));
			Cylinder = Temp;
		}

		//    ""
		UF_CALL(UF_MODL_operations(*Base, Cylinder, UF_NEGATIVE));
	}

	//  
	{
		int iBaseEdges;
		tag_t Blend, Edge;
		UF_MODL_edge_blend_data_s BlendData;
		UF_MODL_blend_edge_s *BlendEdge;
		UF_MODL_blend_point_s *BlendPoints;
		uf_list_p_t BaseEdges;

		BlendData.blend_type = 0;
		BlendData.blend_instanced = FALSE;
		BlendData.blend_setback = FALSE;
		BlendData.vrb_tolerance = .01;
		BlendData.blend_radius[0] = '0';
		BlendData.blend_radius[1] = '\0';
		BlendData.smooth_overflow = TRUE;
		BlendData.cliff_overflow = TRUE;
		BlendData.notch_overflow = TRUE;
		BlendData.number_edges = iEdgesNum;

		BlendEdge = new UF_MODL_blend_edge_s[iEdgesNum];

		//   
		UF_CALL(UF_MODL_ask_body_edges(*Base, &BaseEdges));
		UF_CALL(UF_MODL_ask_list_count(BaseEdges, &iBaseEdges));

		//   (       )
		for(int i=0;i<iEdgesNum;i++)
		{
			UF_EVAL_p_t Eval;
			double dGuess[3], dMinDist, dParm, dParLimits[2];

			//  
			for(int j=0;j<iBaseEdges;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(BaseEdges, j, &BlendEdge[i].edge));;
				if(CurvesCoincide(SplinesTop[i], BlendEdge[i].edge))
					break;
			}

			BlendEdge[i].number_points = Edges[i].iSectionsNum;
			BlendEdge[i].point_data = new UF_MODL_blend_point_s[Edges[i].iSectionsNum];

			UF_CALL(UF_EVAL_initialize(BlendEdge[i].edge, &Eval));
			UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].PointRealBase.Fill(dGuess);
				UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
				BlendEdge[i].point_data[j].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
				//    
				if(iEdgesNum==1 && j==0 && BlendEdge[i].point_data[j].parameter==dParLimits[1])
					BlendEdge[i].point_data[j].parameter = dParLimits[0];
				if(iEdgesNum==1 && j==Edges[i].iSectionsNum-1 && BlendEdge[i].point_data[j].parameter==dParLimits[0])
					BlendEdge[i].point_data[j].parameter = dParLimits[1];
				string.Format("%f", Edges[i].Sections[j].r1);
				string.CopyChars(BlendEdge[i].point_data[j].radius, string.GetBuffer(), string.GetLength());
				BlendEdge[i].point_data[j].radius[string.GetLength()] = '\0';
			}

			UF_CALL(UF_EVAL_free(Eval));
		}

		BlendData.edge_data = BlendEdge;
		UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

		//   (       )
		for(int i=0;i<iEdgesNum;i++)
		{
			UF_EVAL_p_t Eval;
			double dGuess[3], dMinDist, dParm, dParLimits[2];

			//  
			for(int j=0;j<iBaseEdges;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(BaseEdges, j, &BlendEdge[i].edge));;
				if(CurvesCoincide(SplinesBottom[i], BlendEdge[i].edge))
					break;
			}

			UF_CALL(UF_EVAL_initialize(BlendEdge[i].edge, &Eval));
			UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].PointRealBottom.Fill(dGuess);
				UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
				BlendEdge[i].point_data[j].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
				//    
				if(iEdgesNum==1 && j==0 && BlendEdge[i].point_data[j].parameter==dParLimits[1])
					BlendEdge[i].point_data[j].parameter = dParLimits[0];
				if(iEdgesNum==1 && j==Edges[i].iSectionsNum-1 && BlendEdge[i].point_data[j].parameter==dParLimits[0])
					BlendEdge[i].point_data[j].parameter = dParLimits[1];
				string.Format("%f", Edges[i].Sections[j].r2);
				string.CopyChars(BlendEdge[i].point_data[j].radius, string.GetBuffer(), string.GetLength());
				BlendEdge[i].point_data[j].radius[string.GetLength()] = '\0';
			}

			UF_CALL(UF_EVAL_free(Eval));
		}

		BlendData.edge_data = BlendEdge;
		UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

		// 
		for(int i=0;i<iEdgesNum;i++)
			delete [] BlendEdge[i].point_data;
		delete [] BlendEdge;
	}

	//  
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_OBJ_set_blank_status(SplinesTop[i], UF_OBJ_BLANKED));
		UF_CALL(UF_OBJ_set_blank_status(SplinesBottom[i], UF_OBJ_BLANKED));
	}

	delete [] SplinesTop;
	delete [] SplinesBottom;
}

//       
void CPressIn::CreatePressInPress(tag_t *Press, double dBaseNormal[3], double dPartThickness)
{
tag_t Cylinder;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity;
double dPointOrigin[3], dTol[3]={.01, .01, 1};
UF_STRING_t Section;

	UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
	Section.num = 1;
	Section.string[0] = iEdgesNum;

	//  
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
			OffsetPoint(dPointOrigin, Points[j].point, dBaseNormal, dG_BasePressDistance - dPartThickness);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&Section.id[i]));

		Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;

		delete [] Points;
	}

	//  
	{
		int iL;
		UF_CALL(UF_LAYER_ask_work_layer(&iL));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Cylinder));
		UF_CALL(UF_LAYER_set_status(iL, UF_LAYER_WORK_LAYER));
		UF_CALL(UF_LAYER_set_status(111, UF_LAYER_INACTIVE_LAYER));
	}
//	UF_CALL(UF_OBJ_set_translucency(Cylinder, 100));
	UF_CALL(UF_OBJ_set_blank_status(Cylinder, UF_OBJ_BLANKED));

	//    Extrude
	{
		uf_list_p_t ExtrObjects, ExtrFeats;
		char *cTaper="0", *cLimit[2]={"0", "0"};
		CString string;
	
		UF_CALL(UF_MODL_create_list(&ExtrObjects));
		UF_CALL(UF_MODL_put_list_item(ExtrObjects, Cylinder));

		string.Format("%f", dG_BushInPressThickness + 2*dPartThickness);
		cLimit[1] = string.GetBuffer();

		UF_CALL(UF_MODL_create_extruded(ExtrObjects, cTaper, cLimit, dPointOrigin, dBaseNormal, UF_NULLSIGN, &ExtrFeats));

		UF_CALL(UF_MODL_ask_list_item(ExtrFeats, 0, &Cylinder));

		UF_CALL(UF_MODL_delete_list(&ExtrObjects));
		UF_CALL(UF_MODL_delete_list(&ExtrFeats));

		//         ""
		{
			tag_t Temp;
			UF_CALL(UF_MODL_ask_feat_body(Cylinder, &Temp));
			Cylinder = Temp;
		}
		PeelBody(Cylinder, dPartThickness, true);
/*		{
			uf_list_p_t BodyFaces;
			tag_t Temp;
			int iFaces;

			UF_CALL(UF_MODL_ask_body_faces(Cylinder, &BodyFaces));
			UF_CALL(UF_MODL_ask_list_count(BodyFaces, &iFaces));
			for(int i=0;i<iFaces;i++)
			{
				UF_CALL(UF_MODL_ask_list_item(BodyFaces, i, &Temp));
				UF_CALL(UF_OBJ_set_color(Temp, UF_OBJ_YELLOW));
			}
		}*/
		UF_CALL(UF_MODL_operations(*Press, Cylinder, UF_NEGATIVE));
	}

	//  
	for(int i=0;i<iEdgesNum;i++)
		UF_CALL(UF_OBJ_set_blank_status(Section.id[i], UF_OBJ_BLANKED));
	UF_MODL_free_string_list(&Section);
}

//     
tag_t CPressIn::CreatePressInBush(double dBaseNormal[3], double dPartThickness)
{
tag_t *SplinesTop = new tag_t[iEdgesNum], *SplinesBottom = new tag_t[iEdgesNum], Cylinder;
UF_CURVE_pt_slope_crvatr_s *Points;
int iDegree, iPeriodicity;
double dPointOrigin[3];
CString string;

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBase.Fill(dPointOrigin);
			OffsetPoint(dPointOrigin, Points[j].point, dBaseNormal, dG_BasePressDistance/2);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesTop[i]));

		delete [] Points;
	}

	//   
	for(int i=0;i<iEdgesNum;i++)
	{
		iDegree = iSplinesDegree;

		//,        
		if(Edges[i].iSectionsNum<iDegree+1)
			iDegree = Edges[i].iSectionsNum-1;

		Points = new UF_CURVE_pt_slope_crvatr_s[Edges[i].iSectionsNum];

		for(int j=0;j<Edges[i].iSectionsNum;j++)
		{
			Edges[i].Sections[j].PointRealBottom.Fill(dPointOrigin);
			OffsetPoint(dPointOrigin, Points[j].point, dBaseNormal, dG_BasePressDistance/2);
			Points[j].slope_type = UF_CURVE_SLOPE_NONE;
			Points[j].crvatr_type = UF_CURVE_CRVATR_NONE;
		}

		if(iEdgesNum==1)
			iPeriodicity = 1;
		else
			iPeriodicity = 0;

		//    
		UF_CALL(UF_CURVE_create_spline_thru_pts(
			iDegree,
			iPeriodicity,
			Edges[i].iSectionsNum,
			Points,
			NULL,
			0,
			&SplinesBottom[i]));

		delete [] Points;
	}

	//  
	{
		tag_t Plane, *SewTools;//, Curves[2];
		UF_STRING_t Section, Spine;
		int iAlignment=1, iEndPoint=0, iBodyType=0;//, iNumOfRuleds=0, iNumOfPS, iReverse[2]={0,0};
		double dValue[6], dTol[3]={.01, .01, 1};//, dParms[2]={1,0};
		uf_list_p_t ListDisjoint;

		//   
		SewTools = new tag_t[iEdgesNum+1];

		UF_MODL_create_string_list(2, 2, &Section);
		Section.num = 2;
		Section.string[0] = 1;
		Section.dir[0] = UF_MODL_CURVE_START_FROM_BEGIN;
		Section.string[1] = 1;
		Section.dir[1] = UF_MODL_CURVE_START_FROM_BEGIN;
		Spine.num = 0;

		//   
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.id[0] = SplinesTop[i];
			Section.id[1] = SplinesBottom[i];

			UF_CALL(UF_MODL_create_ruled(&Section, &Spine, &iAlignment, dValue, &iEndPoint, &iBodyType, UF_NULLSIGN, dTol, &SewTools[i]));
		}

		UF_MODL_free_string_list(&Section);

		//  
		UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
		Section.num = 1;
		Section.string[0] = iEdgesNum;
		// 
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
			Section.id[i] = SplinesTop[i];
		}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &SewTools[iEdgesNum]));
		// 
		for(int i=0;i<iEdgesNum;i++)
		{
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
			Section.id[i] = SplinesBottom[i];
		}
		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Plane));
		UF_MODL_free_string_list(&Section);

		//   
		UF_CALL(UF_MODL_create_sew(0, 1, &Plane, iEdgesNum+1, SewTools, .1, 1, &ListDisjoint, &Cylinder));
		{
			tag_t Temp;
			UF_CALL(UF_MODL_ask_feat_body(Cylinder, &Temp));
			Cylinder = Temp;
		}
	}

	//   ,   
	{
		tag_t Cyl2;
		UF_STRING_t Section;
		double dTol[3]={.01, .01, 1};
		uf_list_p_t ExtrObjects, ExtrFeats;
		char *cTaper="0", *cLimit[2]={"0", "0"};
		CString string;

		UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
		Section.num = 1;
		Section.string[0] = iEdgesNum;

		for(int i=0;i<iEdgesNum;i++)
		{
			Section.id[i] = SplinesTop[i];
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
		}

		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Cyl2));
		UF_CALL(UF_OBJ_set_blank_status(Cyl2, UF_OBJ_BLANKED));

		UF_CALL(UF_MODL_create_list(&ExtrObjects));
		UF_CALL(UF_MODL_put_list_item(ExtrObjects, Cyl2));

		string.Format("%f", dG_BushInPressThickness + 2*dPartThickness);
		cLimit[1] = string.GetBuffer();

		UF_CALL(UF_MODL_create_extruded(ExtrObjects, cTaper, cLimit, dPointOrigin, dBaseNormal, UF_NULLSIGN, &ExtrFeats));

		UF_CALL(UF_MODL_ask_list_item(ExtrFeats, 0, &Cyl2));

		UF_CALL(UF_MODL_delete_list(&ExtrObjects));
		UF_CALL(UF_MODL_delete_list(&ExtrFeats));

		{
			tag_t Temp;
			UF_CALL(UF_MODL_ask_feat_body(Cyl2, &Temp));
			Cyl2 = Temp;
		}
		UF_CALL(UF_MODL_operations(Cylinder, Cyl2, UF_POSITIVE));
	}

	// 
	{
		int iCylinderEdges;
		tag_t Blend, Edge;
		UF_MODL_edge_blend_data_s BlendData;
		UF_MODL_blend_edge_s *BlendEdge;
		UF_MODL_blend_point_s *BlendPoints;
		uf_list_p_t CylinderEdges;

		BlendData.blend_type = 0;
		BlendData.blend_instanced = FALSE;
		BlendData.blend_setback = FALSE;
		BlendData.vrb_tolerance = .01;
		BlendData.blend_radius[0] = '0';
		BlendData.blend_radius[1] = '\0';
		BlendData.smooth_overflow = TRUE;
		BlendData.cliff_overflow = TRUE;
		BlendData.notch_overflow = TRUE;
		BlendData.number_edges = iEdgesNum;

		BlendEdge = new UF_MODL_blend_edge_s[iEdgesNum];

		//   
		UF_CALL(UF_MODL_ask_body_edges(Cylinder, &CylinderEdges));
		UF_CALL(UF_MODL_ask_list_count(CylinderEdges, &iCylinderEdges));

		//   (       )
		for(int i=0;i<iEdgesNum;i++)
		{
			UF_EVAL_p_t Eval;
			double dGuess[3], dMinDist, dParm, dParLimits[2];

			//  
			for(int j=0;j<iCylinderEdges;j++)
			{
				UF_CALL(UF_MODL_ask_list_item(CylinderEdges, j, &BlendEdge[i].edge));;
				if(CurvesCoincide(SplinesBottom[i], BlendEdge[i].edge, .1))
					break;
			}

			BlendEdge[i].number_points = Edges[i].iSectionsNum;
			BlendEdge[i].point_data = new UF_MODL_blend_point_s[Edges[i].iSectionsNum];

			UF_CALL(UF_EVAL_initialize(BlendEdge[i].edge, &Eval));
			UF_CALL(UF_EVAL_ask_limits(Eval, dParLimits));

			for(int j=0;j<Edges[i].iSectionsNum;j++)
			{
				Edges[i].Sections[j].PointRealBottom.Fill(dPointOrigin);
				OffsetPoint(dPointOrigin, dGuess, dBaseNormal, dG_BasePressDistance/2);
				UF_CALL(UF_EVAL_evaluate_closest_point(Eval, dGuess, &dParm, dGuess));
				BlendEdge[i].point_data[j].parameter = (dParm-dParLimits[0])/(dParLimits[1]-dParLimits[0]);
				//    
				if(iEdgesNum==1 && j==0 && BlendEdge[i].point_data[j].parameter==dParLimits[1])
					BlendEdge[i].point_data[j].parameter = dParLimits[0];
				if(iEdgesNum==1 && j==Edges[i].iSectionsNum-1 && BlendEdge[i].point_data[j].parameter==dParLimits[0])
					BlendEdge[i].point_data[j].parameter = dParLimits[1];
				string.Format("%f", Edges[i].Sections[j].r2);
				string.CopyChars(BlendEdge[i].point_data[j].radius, string.GetBuffer(), string.GetLength());
				BlendEdge[i].point_data[j].radius[string.GetLength()] = '\0';
			}

			UF_CALL(UF_EVAL_free(Eval));
		}

		BlendData.edge_data = BlendEdge;
		UF_CALL(UF_MODL_create_edge_blend(&BlendData, &Blend));

		// 
		for(int i=0;i<iEdgesNum;i++)
			delete [] BlendEdge[i].point_data;
		delete [] BlendEdge;
	}

	//" "  
	PeelBody(Cylinder, dPartThickness);

	//     ,   
	//    
/*	{
		tag_t Cyl2;
		UF_STRING_t Section;
		double dTol[3]={.01, .01, 1};
		uf_list_p_t ExtrObjects, ExtrFeats;
		char *cTaper="0", *cLimit[2]={"0", "0"};
		CString string1, string2;

		UF_MODL_create_string_list(iEdgesNum, iEdgesNum, &Section);
		Section.num = 1;
		Section.string[0] = iEdgesNum;

		for(int i=0;i<iEdgesNum;i++)
		{
			Section.id[i] = SplinesTop[i];
			Section.dir[i] = UF_MODL_CURVE_START_FROM_BEGIN;
		}

		UF_CALL(UF_MODL_create_bplane(&Section, dTol, &Cyl2));
		UF_CALL(UF_OBJ_set_blank_status(Cyl2, UF_OBJ_BLANKED));

		UF_CALL(UF_MODL_create_list(&ExtrObjects));
		UF_CALL(UF_MODL_put_list_item(ExtrObjects, Cyl2));

		string1.Format("%f", -dPartThickness);
		cLimit[0] = string1.GetBuffer();
		string2.Format("%f", dG_BushInPressThickness + dPartThickness);
		cLimit[1] = string2.GetBuffer();

		UF_CALL(UF_MODL_create_extruded(ExtrObjects, cTaper, cLimit, dPointOrigin, dBaseNormal, UF_NULLSIGN, &ExtrFeats));

		UF_CALL(UF_MODL_ask_list_item(ExtrFeats, 0, &Cyl2));

		UF_CALL(UF_MODL_delete_list(&ExtrObjects));
		UF_CALL(UF_MODL_delete_list(&ExtrFeats));

		{
			tag_t Temp;
			UF_CALL(UF_MODL_ask_feat_body(Cyl2, &Temp));
			Cyl2 = Temp;
		}
		PeelBody(Cyl2, dPartThickness);
		UF_CALL(UF_MODL_operations(Cylinder, Cyl2, UF_POSITIVE));
	}*/

	//  
	for(int i=0;i<iEdgesNum;i++)
	{
		UF_CALL(UF_OBJ_set_blank_status(SplinesTop[i], UF_OBJ_BLANKED));
		UF_CALL(UF_OBJ_set_blank_status(SplinesBottom[i], UF_OBJ_BLANKED));
	}

	delete [] SplinesTop;
	delete [] SplinesBottom;

	return Cylinder;
}

int CPressInEdge::CreateInnerPlanes(tag_t Edge, int iIntervalsNum)
{
int iNumber, iIntervalsNumOld = iIntervalsNum;	//    
double dLength, dEvalLimits[2];
double dPoint[3], dTangent[3], dNormal[3], dBiNormal[3];
UF_EVAL_p_t Evaluator;

	//,   
	UF_CALL(UF_CURVE_ask_parameterization(Edge, dEvalLimits, &iNumber));
	if(iNumber==UF_CURVE_CLOSED_PERIODIC_CURVE || iNumber==UF_CURVE_CLOSED_NON_PERIODIC_CURVE)
		iIntervalsNum--;

	Sections = new CPressInSection[iIntervalsNum];
	iSectionsNum = iIntervalsNum;

	// 
	UF_CALL(UF_CURVE_ask_arc_length(Edge, 0, 1, UF_MODL_MMETER, &dLength));
	UF_CALL(UF_EVAL_initialize(Edge, &Evaluator));
	UF_CALL(UF_EVAL_ask_limits(Evaluator, dEvalLimits));


	//      
	for(int i=0;i<iIntervalsNum;i++)
	{
		//   
		UF_CALL(UF_EVAL_evaluate_unit_vectors(
			Evaluator,
			dEvalLimits[0] + (double(i)/double(iIntervalsNumOld-1))*(dEvalLimits[1]-dEvalLimits[0]),
			dPoint,
			dTangent,
			dNormal,
			dBiNormal));

		//    
		Sections[i].PointRealBase = dPoint;

		//       
		UF_CALL(UF_MODL_create_plane(
			dPoint,
			dTangent,
			&Sections[i].PlaneSection));
	}

	//  
	Parameters.ScanEdge(Edge);

	return 0;
}

//    :
//PointRealBottom
int CPressInEdge::ScanRealPoints(uf_list_p_t FacesPressInBevel)
{
int iNumber;
double dPoint1[3], dPoint2[3];

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointRealBase.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPressInBevel, dPoint1, dPoint2)==1)
			MessageBox(NULL, "         ", " ", MB_OK);
		else
			Sections[i].PointRealBottom = dPoint2;
	}

	return 0;
}

//    :
//PointBase
//PointRounding1
//PointBevel
//PointBottom
// :
//1 -       
//2 -     ,  ,   
int CPressInEdge::ScanPoints(uf_list_p_t FacesPressInBevel, uf_list_p_t FacesPressInRounding1, uf_list_p_t FacesPressInRounding2, uf_list_p_t FacesPressInBottom)
{
int iNumber;
double dPoint1[3], dPoint2[3];

//,     
UF_MODL_intersect_info_p_t **Intersections;
tag_t Face;
int *iIntNum, iMinI=-1, iMinJ=-1, iFlag=UF_MODL_LOC;
double dDist, daTemp[3], dPoint[3], dMinDist=1e100, dP0=0, dP1=1, dPoint0[3], dDist0, dDist1, dSectionPoint[3];

	//    -             PointBase
	for(int k=0;k<iSectionsNum;k++)
	{
		Sections[k].PointRealBase.Fill(dPoint); //???

		//  
		UF_CALL(UF_MODL_ask_list_count(FacesPressInBottom, &iNumber));
		Intersections = new UF_MODL_intersect_info_p_t*[iNumber];
		iIntNum = new int[iNumber];
		for(int i=0;i<iNumber;i++)
		{
			UF_CALL(UF_MODL_ask_list_item(FacesPressInBottom, i, &Face));
			UF_CALL(UF_MODL_intersect_objects(Sections[k].PlaneSection, Face, 1e-5, &iIntNum[i], &Intersections[i]));
		}

		//  
		for(int i=0;i<iNumber;i++)
		for(int j=0;j<iIntNum[i];j++)
		{
			if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_CURVE)
			{
				UF_CALL(UF_MODL_ask_minimum_dist(
					Intersections[i][j]->intersect.curve.identifier,
					NULL_TAG,
					0,
					daTemp,
					1,
					dPoint,
					&dDist,
					daTemp,
					daTemp));
			}
			else if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_POINT)
			{
				UF_CALL(UF_MODL_ask_minimum_dist(
					NULL_TAG,
					NULL_TAG,
					1,
					Intersections[i][j]->intersect.point.coords,
					1,
					dPoint,
					&dDist,
					daTemp,
					daTemp));
			}
			if(dDist<dMinDist)
			{
				dMinDist = dDist;
				iMinI = i;
				iMinJ = j;
			}
		}

		if(iMinI==-1)
		{
			//  
			for(int i=0;i<iNumber;i++)
			for(int j=0;j<iIntNum[i];j++)
			{
				if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_CURVE)
					UF_CALL(UF_ASSEM_remove_instance(Intersections[i][j]->intersect.curve.identifier));
				else if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_COINCIDE)
					UF_CALL(UF_ASSEM_remove_instance(Intersections[i][j]->intersect.coincide.identifier));
				UF_free(Intersections[i][j]);
			}
			delete [] Intersections;

			return 1;
		}

		//          ,     dPoint
		if(Intersections[iMinI][iMinJ]->intersect_type==UF_MODL_INTERSECT_CURVE)
		{
			UF_CALL(UF_MODL_evaluate_curve(Intersections[iMinI][iMinJ]->intersect.curve.identifier, &dP0, &iFlag, dPoint0));
			UF_CALL(UF_MODL_ask_minimum_dist(
				NULL_TAG,
				NULL_TAG,
				1,
				dPoint0,
				1,
				dPoint,
				&dDist0,
				daTemp,
				daTemp));

			UF_CALL(UF_MODL_evaluate_curve(Intersections[iMinI][iMinJ]->intersect.curve.identifier, &dP1, &iFlag, dPoint1));
			UF_CALL(UF_MODL_ask_minimum_dist(
				NULL_TAG,
				NULL_TAG,
				1,
				dPoint1,
				1,
				dPoint,
				&dDist1,
				daTemp,
				daTemp));
			if(dDist0>dDist1)
				Sections[k].PointBase = dPoint0;
			else
				Sections[k].PointBase = dPoint1;
		}
		else
		{
			//  
			for(int i=0;i<iNumber;i++)
			for(int j=0;j<iIntNum[i];j++)
			{
				if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_CURVE)
					UF_CALL(UF_ASSEM_remove_instance(Intersections[i][j]->intersect.curve.identifier));
				else if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_COINCIDE)
					UF_CALL(UF_ASSEM_remove_instance(Intersections[i][j]->intersect.coincide.identifier));
				UF_free(Intersections[i][j]);
			}
			delete [] Intersections;

			return 2;
		}

		//  
		for(int i=0;i<iNumber;i++)
		for(int j=0;j<iIntNum[i];j++)
		{
			if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_CURVE)
				UF_CALL(UF_ASSEM_remove_instance(Intersections[i][j]->intersect.curve.identifier));
			else if(Intersections[i][j]->intersect_type==UF_MODL_INTERSECT_COINCIDE)
				UF_CALL(UF_ASSEM_remove_instance(Intersections[i][j]->intersect.coincide.identifier));
			UF_free(Intersections[i][j]);
		}
		delete [] Intersections;
	}

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointBase.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPressInBottom, dPoint1, dPoint2)==1)
			MessageBox(NULL, "         ", " ", MB_OK);
		else
			Sections[i].PointBottom = dPoint2;
	}

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointBottom.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPressInRounding2, dPoint1, dPoint2, &Sections[i].r2)==1)
			MessageBox(NULL, "          ", " ", MB_OK);
		else
			Sections[i].PointBevel = dPoint2;
	}

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointBevel.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPressInBevel, dPoint1, dPoint2)==1)
			MessageBox(NULL, "         ", " ", MB_OK);
		else
			Sections[i].PointRounding1 = dPoint2;
	}

	for(int i=0;i<iSectionsNum;i++)
	{
		Sections[i].PointRounding1.Fill(dPoint1);
		if(ScanSectionParms(Sections[i].PlaneSection, FacesPressInRounding1, dPoint1, dPoint2, &Sections[i].r1)==1)
			MessageBox(NULL, "          ", " ", MB_OK);
		else
			Sections[i].PointBase = dPoint2;
	}

	return 0;
}

//      
//   :
//{0} -   iSectionsNum   
void CPressInEdge::WriteEdge(CFile *File, CString sFormatEdge, CString sFormatSections)
{
CString str;

	str.Format("%d", iSectionsNum);
	sFormatEdge.Replace("{0}", str);

	File->Write(sFormatEdge, sFormatEdge.GetLength());

	for(int i=0;i<iSectionsNum;i++)
		Sections[i].WriteSection(File, sFormatSections);
}

//        
void CPressInEdge::Reverse(void)
{
char cMemBlock[sizeof(CPressInSection)];

	for(int i=0;i<iSectionsNum/2;i++)
	{
		memcpy(cMemBlock, &Sections[iSectionsNum-i-1], sizeof(CPressInSection));
		memcpy(&Sections[iSectionsNum-i-1], &Sections[i], sizeof(CPressInSection));
		memcpy(&Sections[i], cMemBlock, sizeof(CPressInSection));
	}
}

//      
//   :
	//{0x},{0y},{0z} -   PointRealBase
	//{1x},{1y},{1z} -   PointRealBottom
	//{2x},{2y},{2z} -   PointBase
	//{3x},{3y},{3z} -   PointRounding1
	//{4x},{4y},{4z} -   PointBevel
	//{5x},{5y},{5z} -   PointBottom
void CPressInSection::WriteSection(CFile *File, CString sFormatSections)
{
CString str;

	str.Format("%f", PointRealBase.X);
	sFormatSections.Replace("{0x}", str);
	str.Format("%f", PointRealBase.Y);
	sFormatSections.Replace("{0y}", str);
	str.Format("%f", PointRealBase.Z);
	sFormatSections.Replace("{0z}", str);

	str.Format("%f", PointRealBottom.X);
	sFormatSections.Replace("{1x}", str);
	str.Format("%f", PointRealBottom.Y);
	sFormatSections.Replace("{1y}", str);
	str.Format("%f", PointRealBottom.Z);
	sFormatSections.Replace("{1z}", str);

	str.Format("%f", PointBase.X);
	sFormatSections.Replace("{2x}", str);
	str.Format("%f", PointBase.Y);
	sFormatSections.Replace("{2y}", str);
	str.Format("%f", PointBase.Z);
	sFormatSections.Replace("{2z}", str);

	str.Format("%f", PointRounding1.X);
	sFormatSections.Replace("{3x}", str);
	str.Format("%f", PointRounding1.Y);
	sFormatSections.Replace("{3y}", str);
	str.Format("%f", PointRounding1.Z);
	sFormatSections.Replace("{3z}", str);

	str.Format("%f", PointBevel.X);
	sFormatSections.Replace("{4x}", str);
	str.Format("%f", PointBevel.Y);
	sFormatSections.Replace("{4y}", str);
	str.Format("%f", PointBevel.Z);
	sFormatSections.Replace("{4z}", str);

	str.Format("%f", PointBottom.X);
	sFormatSections.Replace("{5x}", str);
	str.Format("%f", PointBottom.Y);
	sFormatSections.Replace("{5y}", str);
	str.Format("%f", PointBottom.Z);
	sFormatSections.Replace("{5z}", str);

	File->Write(sFormatSections, sFormatSections.GetLength());
}

//####################################################################################################################################################
//####################################################################################################################################################
//####################################################################################################################################################

void CEdgeParameters::ScanEdge(tag_t Edge)
{
UF_CURVE_struct_s *CurveData;
UF_CURVE_arc_s ArcData;
double dOrientation[3]={0,0,1};

	//  
	UF_CALL(UF_CURVE_ask_curve_struct(Edge, &CurveData));
	switch(CurveData->crv_type)
	{
	case UF_line_type:
		Type = Line;
		break;
	case UF_circle_type:
		Type = Arc;
		break;
	default:
		Type = Other;
		break;
	}

	//  
	UF_CALL(UF_CURVE_ask_arc_length(Edge, 0, 1, UF_MODL_MMETER, &dLength));
	//UF_CALL(UF_CURVE_ask_arc_length(Edge, CurveData->crv_t0, CurveData->crv_t0+CurveData->crv_tscale, UF_MODL_MMETER, &dLength));

	//   
/*	if(Type != Line)
		UF_CALL(UF_CURVE_ask_curve_turn_angle(Edge,	dOrientation, &dAngle));
	else
		dAngle = 0;*/

	//    ( )
	if(Type == Arc)
		dAngle = CurveData->crv_tscale;
	else
		dAngle = 0;

	//   ( )
	if(Type == Arc)
	{
		UF_CALL(UF_CURVE_ask_arc_data(Edge, &ArcData));
		dRadius = ArcData.radius;
	}
	else
		dRadius = 0;

	UF_CALL(UF_CURVE_free_curve_struct(CurveData));
}

void CEdgeParameters::WriteParameters(CFile *File)
{
CString str;
	
	str.Format("\r\n#EdgeLength = %f\r\n#EdgeAngle = %f\r\n#EdgeRadius = %f\r\n", dLength, dAngle, dRadius);

	switch(Type)
	{
	case Line:
		str = "#EdgeType = Line" + str;
		break;
	case Arc:
		str = "#EdgeType = Arc" + str;
		break;
	case Other:
		str = "#EdgeType = Other" + str;
		break;
	}

	File->Write(str, str.GetLength());
}
